const polylabel = require("polylabel");
const hull = require('hull.js');
const turf = require('@turf/turf');
// const offset = require('polygon-offset');
const UuidEncoder = require('uuid-encoder');
const THREE = require('three');

var ifc = (function () {
    var ifcTypes = {
        'IFCPROJECT': new IfcProject(),
        'IFCSITE': new IfcSite(),
        'IFCBUILDING': new IfcBuilding(),
        'IFCPOSTALADDRESS': new IfcPostalAddress(),
        'IFCBUILDINGSTOREY': new IfcBuildingStorey(),
        'IFCSPACE': new IfcSpace(),
        'IFCDOOR': new IfcDoor(),
        'IFCWALL': new IfcWall(),
        'IFCWALLSTANDARDCASE': new IfcWallStandardCase(),
        'IFCSLAB': new IfcSlab(),
        'IFCFLOWTERMINAL': new IfcFlowTerminal(),
        'IFCBUILDINGELEMENTPROXY': new IfcBuildingElementProxy(),
        'IFCCOLUMN': new IfcColumn(),
        'IFCSTAIR': new IfcStair(),
        'IFCSTAIRFLIGHT': new IfcStairFlight(),
        'IFCTRANSPORTELEMENT': new IfcTransportElement(),
        'IFCRAMP': new IfcRamp(),
        'IFCRELDEFINESBYTYPE': new IfcRelDefinesByType(),
        'IFCWALLTYPE': new IfcWallType(),
        'IFCLOCALPLACEMENT': new IfcLocalPlacement(),
        'IFCAXIS2PLACEMENT2D': new IfcAxis2Placement2D(),
        'IFCAXIS2PLACEMENT3D': new IfcAxis2Placement3D(),
        'IFCCARTESIANPOINT': new IfcCartesianPoint(),
        'IFCDIRECTION': new IfcDirection(),
        'IFCRELAGGREGATES': new IfcRelAggregates(),
        'IFCRELDEFINESBYPROPERTIES': new IfcRelDefinesByProperties(),
        'IFCPROPERTYSET': new IfcPropertySet(),
        'IFCPROPERTYSINGLEVALUE': new IfcPropertySingleValue(),
        'IFCRELCONTAINEDINSPATIALSTRUCTURE': new IfcRelContainedInSpatialStructure(),
        'IFCRELREFERENCEDINSPATIALSTRUCTURE': new IfcRelReferencedInSpatialStructure(),
        'IFCPRODUCTDEFINITIONSHAPE': new IfcProductDefinitionShape(),
        'IFCSHAPEREPRESENTATION': new IfcShapeRepresentation(),
        'IFCEXTRUDEDAREASOLID': new IfcExtrudedAreaSolid(),
        'IFCRECTANGLEPROFILEDEF': new IfcRectangleProfileDef(),
        'IFCARBITRARYCLOSEDPROFILEDEF': new IfcArbitraryClosedProfileDef(),
        'IFCARBITRARYPROFILEDEFWITHVOIDS': new IfcArbitraryProfileDefWithVoids(),
        'IFCPOLYLINE': new IfcPolyline(),
        'IFCCOMPOSITECURVE': new IfcCompositeCurve()
    };

// Relevant IFC Classes

    function IfcProject() {
    }

// #246133= IFCSITE('2OMlvMbj97ih$8FxLimvkN',#41,'Default',$,'',#246132,$,$,.ELEMENT.,(42,21,31,181945),(-71,-3,-24,-263305),35200.,$,$);

    IfcProject.prototype.create = function (name) {
      this.type = 'IFCPROJECT';
      this.map = {};

      this.oid = 1;
      this._maxOid = 1;
      this.globalId = getUuid();
      this.name = name;
      this.map[this.oid] = this;

      this.__relAggregates = {};
      this.__relContainedInSpatialStructure = {};
      this.__relReferencedInSpatialStructure = {};
      this.__relDefinesByProperties = {};

      this.entitlements = {};
      this.accessPathDoors = {};
      this.ifcElementAttributes = {};

      return this;
    };

    IfcProject.prototype.load = function (errors) {
        this.globalId = this.__attributes[0];
        this.name = this.__attributes[2];
        this.description = this.__attributes[3];
        this.longName = this.__attributes[5];
    };

    IfcProject.prototype.bind = function (project) {
      this.__sites = [];
      this.__relAggregates = {};
      this.__relContainedInSpatialStructure = {};
      this.__relReferencedInSpatialStructure = {};
      this.__relDefinesByProperties = {};

      if (!this.entitlements) {
        this.entitlements = {};
      }

      if (!this.accessPathDoors) {
        this.accessPathDoors = {};
      }

      if (!this.ifcElementAttributes) {
        this.ifcElementAttributes = {};
      }

      const elementOids = Object.keys(this.map)
        .map((key) => Number(key))
        .filter((oid) => !Number.isNaN(oid));
      this._maxOid = Math.max([...elementOids, this.oid]);
    };

    IfcProject.prototype.register = function (object) {
        ++this._maxOid;

        object.oid = this._maxOid;
        this.map[object.oid] = object;
    };

    IfcProject.prototype.unregister = function (object) {
        delete this.map[object.oid];
    };

    IfcProject.prototype.getFullId = function () {
        return this.globalId;
    };

    IfcProject.prototype.getSites = function () {
        return this.__sites;
    };

    IfcProject.prototype.getSite = function (id) {
        for (var n in this.__sites) {
            if (this.__sites[n]) {
                this.__sites[n].id == id;

                return this.__sites[n];
            }
        }

        return null;
    };

    IfcProject.prototype.getPropertyValueRangeForSpaces = function (name) {
        var valueMap = {};
        var values = [];

        for (var n in this.__sites) {
            var site = this.__sites[n];

            for (var m in site.__buildings) {
                var building = site.__buildings[m];

                for (var k in building.__floors) {
                    var floor = building.__floors[k];

                    for (var i in floor.__spaces) {
                        var space = floor.__spaces[i];

                        for (var l in space.__propertySets) {
                            var propertySet = space.__propertySets[l];

                            for (var j in propertySet.__propertyValues) {
                                var propertyValue = propertySet.__propertyValues[j];

                                if (propertyValue.name == name && !valueMap[propertyValue.value]) {
                                    valueMap[propertyValue.value] = propertyValue.value;
                                    values.push(propertyValue.value);
                                }
                            }
                        }
                    }
                }
            }
        }

        return values;
    };

    IfcProject.prototype.getSpacesWithProperty = function (name, value) {
        var spaces = [];

        for (var n in this.__sites) {
            var site = this.__sites[n];

            for (var m in site.__buildings) {
                var building = site.__buildings[m];

                for (var k in building.__floors) {
                    var floor = building.__floors[k];

                    for (var i in floor.__spaces) {
                        var space = floor.__spaces[i];

                        for (var l in space.__propertySets) {
                            var propertySet = space.__propertySets[l];

                            for (var j in propertySet.__propertyValues) {
                                var propertyValue = propertySet.__propertyValues[j];

                                if (propertyValue.name == name && propertyValue.value == value) {
                                    spaces.push(space);
                                }
                            }

                        }
                    }
                }
            }
        }

        return spaces;
    };

    IfcProject.prototype.getSpaceByFullId = function (fullId) {
        var ids = fullId.split('/');

        return this.getSite(ids[1]).getBuilding(ids[2]).getFloor(ids[3]).getSpace(ids[4]);
    };

    IfcProject.prototype.getRelAggregate = function (relatingObject) {
        var relAggregate = this.__relAggregates[relatingObject.oid];

        if (!relAggregate) {
            relAggregate = this.__relAggregates[relatingObject.oid] = new IfcRelAggregates().create(this, relatingObject);
        }

        return relAggregate;
    };

    IfcProject.prototype.getRelContainedInSpatialStructure = function (relatingStructure) {
        let relContainedInSpatialStructure = this.__relContainedInSpatialStructure[relatingStructure.oid];

        if (!relContainedInSpatialStructure) {
          relContainedInSpatialStructure = this.__relContainedInSpatialStructure[relatingStructure.oid] = new IfcRelContainedInSpatialStructure().create(this, relatingStructure, '');
        }

        return relContainedInSpatialStructure;
    };

    IfcProject.prototype.getRelReferencedInSpatialStructure = function (relatingStructure) {
        let relReferencedInSpatialStructure = this.__relReferencedInSpatialStructure[relatingStructure.oid];

        if (!relReferencedInSpatialStructure) {
          relReferencedInSpatialStructure = this.__relReferencedInSpatialStructure[relatingStructure.oid] = new IfcRelReferencedInSpatialStructure().create(this, relatingStructure, '');
        }

        return relReferencedInSpatialStructure;
    };

    /**
     *
     * @param name
     * @param element first element to be linked with the property set, optional
     */
    IfcProject.prototype.addPropertySet = function (name, element) {
        var propertySet = new IfcPropertySet().create(this, name);
        var relDefinesByProperties = this.__relDefinesByProperties[propertySet.oid] = new IfcRelDefinesByProperties().create(this, propertySet, element);

        return propertySet;
    };

    IfcProject.prototype.addSite = function (name) {
        var site = new IfcSite().create(this, name);

        if (!this.__sites) {
            this.__sites = [];
        }

        this.__sites.push(site);
        site.__project = this;

        this.getRelAggregate(this).addRelatedObject(site);

        return site;
    };

    /**
     *
     * @param project
     */
    IfcProject.prototype.bindDoorsToSpaces = function (tolerance) {
      for (const site of this.getSites()) {
        for (const building of site.getBuildings()) {
          for (const floor of building.getFloors()) {
            floor.bindDoorsToSpaces(tolerance);
          }
        }
      }
    };

    IfcProject.prototype.list = function () {
        console.log('PROJECT: ' + this.name + '(' + this.globalId + ')');
        console.log('-----------------------------------------------------------------\n');

        for (var n in this.getSites()) {
            var site = this.getSites()[n];

            console.log(site.type + ' ' + site.name);

            for (var m in site.getBuildings()) {
                var building = site.getBuildings()[m];
                console.log('\t' + building.type + ' ' + building.name);

                for (var l in building.getFloors()) {
                    var floor = building.getFloors()[l];
                    console.log('\t\t' + floor.type + ' ' + floor.name);

                    for (var i in floor.getSpaces()) {
                        var space = floor.getSpaces()[i];
                        console.log('\t\t\t' + space.type + ' ' + space.name);

                        for (var k in space.__properties) {
                            console.log('\t\t\t\t' + k + ': ', space.__properties[k]);
                        }

                        for (var k in space.__doors) {
                            var door = space.__doors[k];
                            console.log('\t\t\t\t' + door.type + ' ' + door.name);

                        }

                    }

                    for (var i in floor.__doors) {
                        var door = floor.__doors[i];
                        console.log('\t\t\t' + door.type + ' ' + door.name, door.getPlacement());

                        for (var k in door.__properties) {
                            console.log('\t\t\t\t' + k + ': ', door.__properties[k]);
                        }
                    }

                    for (var i in floor.__walls) {
                        var wall = floor.__walls[i];
                        console.log('\t\t\t' + wall.type + ' ' + wall.name);

                        for (var k in wall.__properties) {
                            console.log('\t\t\t\t' + k + ': ', wall.__properties[k]);
                        }
                    }
                }
            }
        }
    };

    IfcProject.prototype.stringify = function () {
        delete this.map[this.oid];

        var string = JSON.stringify(cloneFiltered(this, /\_\_|\$\$/), null, 3);

        this.map[this.oid] = this;

        return string;
    };

    IfcProject.prototype.dump = function (stream) {
        stream.write('#' + this.oid);
        stream.write('= ');
        stream.write('IFCPROJECT(');
        dumpStringAttribute(stream, this.globalId, true);
        dumpOidAttribute(stream, null);
        dumpStringAttribute(stream, this.name);
        stream.write(');\n');

        for (var n in this.__sites) {
            this.__sites[n].dump(stream);
        }

        for (var n in this.__relAggregates) {
            this.__relAggregates[n].dump(stream);
        }

        for (var n in this.__relContainedInSpatialStructure) {
            this.__relContainedInSpatialStructure[n].dump(stream);
        }

        for (const relation of this.__relReferencedInSpatialStructure) {
            relation.dump(stream);
        }

        for (var n in this.__relDefinesByProperties) {
            this.__relDefinesByProperties[n].dump(stream);
        }
    };

    /**
     * Return a specific IFC element's attribute object.
     *
     * Lazily creates an attributes object, if needed, thus the IFC project may get modified.
     *
     * @param {any} ifcElement the IFC element an attribute should be retrieved for
     * @return {Object<string, any>} the attribute object, never undefined
     */
    IfcProject.prototype.getIfcElementAttributes = function (ifcElement) {
        if (!this.__ifcElementAttributes) {
            this.__ifcElementAttributes = {};
        }

        if (!this.__ifcElementAttributes[ifcElement.id]) {
            this.__ifcElementAttributes[ifcElement.id] = {};
        }

        return this.__ifcElementAttributes[ifcElement.id];
    }

    /**
     * Return a specific IFC element's attribute value. If no element attributes exist, return undefined.
     *
     * This is an immutable operation, does not lazily create any attribute values.
     *
     * @param {any} ifcElement the IFC element an attribute should be retrieved for
     * @param {string} name the attribute name
     * @return {any} the attribute value, or {undefined}
     */
    IfcProject.prototype.getIfcElementAttribute = function (ifcElement, name) {
        if (!this.__ifcElementAttributes) {
            return undefined;
        }

        if (!this.__ifcElementAttributes[ifcElement.id]) {
            return undefined;
        }

        return this.__ifcElementAttributes[ifcElement.id][name];
    }

    function IfcSite() {
    }

// #246133= IFCSITE('2OMlvMbj97ih$8FxLimvkN',#41,'Default',$,'',#246132,$,$,.ELEMENT.,(42,21,31,181945),(-71,-3,-24,-263305),35200.,$,$);

    IfcSite.prototype.create = function (project, name) {
        this.type = 'IFCSITE';
        this.id = getUuid();
        this.name = name;
        this.__buildings = [];

        project.register(this);

        return this;
    };

    IfcSite.prototype.getFullId = function () {
      return this.__project.getFullId() + '/' + this.id + '/';
    };

    IfcSite.prototype.area = function () {
      let res = 0;
      for(let building of this.__buildings){
        res += building.area();
      }
      return res;
    };

    IfcSite.prototype.remove = function () {
        for (var n in this.__buildings) {
            this.__buildings[n].remove();
        }

        this.__project.getRelAggregate(this.__project).removeRelatedObject(this);
        this.__project.getRelAggregate(this).remove();
        this.__project.getRelContainedInSpatialStructure(this).remove();
        this.__project.unregister(this);
    };

    IfcSite.prototype.initialize = function () {
        this.__buildings = [];

        return this;
    };

    IfcSite.prototype.load = function (errors) {
        this.id = this.__attributes[0];
        this.name = this.__attributes[2];
        this.description = this.__attributes[3];
        this.longName = this.__attributes[7];
    };

    IfcSite.prototype.bind = function (project) {
        this.__project = project;
    };

    IfcSite.prototype.getProject = function () {
        return this.__project;
    };

    IfcSite.prototype.getBuildings = function () {
        return this.__buildings;
    };

    IfcSite.prototype.getBuilding = function (id) {
        for (var n in this.__buildings) {
            if (this.__buildings[n].id == id) {

                return this.__buildings[n];
            }
        }

        return null;
    };

    IfcSite.prototype.getIfcElementAttributes = function() {
        return this.__project.getIfcElementAttributes(this);
    }

    IfcSite.prototype.getIfcElementAttribute = function(name) {
        return this.__project.getIfcElementAttribute(this, name);
    }

    IfcSite.prototype.addBuilding = function (project, name) {
        var building = new IfcBuilding().create(project, name);

        this.__buildings.push(building);
        building.__site = this;

        project.getRelAggregate(this).addRelatedObject(building);

        return building;
    };

    // IFCSITE('2OMlvMbj97ih$8FxLimvkN',#41,'Default',$,'',#246132,$,$,.ELEMENT.,(42,21,31,181945),(-71,-3,-24,-263305),35200.,$,$);

    IfcSite.prototype.dump = function (stream) {
        stream.write('#' + this.oid);
        stream.write('= ');
        stream.write('IFCSITE(');
        dumpStringAttribute(stream, this.id, true);
        dumpOidAttribute(stream, null);
        dumpStringAttribute(stream, this.name);
        dumpOidAttribute(stream, null);
        dumpStringAttribute(stream, '');
        dumpOidAttribute(stream, null);
        dumpOidAttribute(stream, null);
        dumpOidAttribute(stream, null);
        dumpEnumAttribute(stream, '.ELEMENT.');
        stream.write(');\n');

        for (var n in this.__buildings) {
            this.__buildings[n].dump(stream);
        }
    };

    function IfcBuilding() {
    }

// #111= IFCBUILDING('2OMlvMbj97ih$8FxLimvkK',#41,'',$,$,#32,$,'',.ELEMENT.,$,$,#107);

    IfcBuilding.prototype.create = function (project, name) {
        this.type = 'IFCBUILDING';
        this.id = getUuid();
        this.name = name;
        this.__project = project;

        this.__floors = [];
        this.__stairEnsembles = [];
        this.__escalatorEnsembles = [];
        this.__uncontElevators = [];

        project.register(this);

        return this;
    };

    IfcBuilding.prototype.getFullId = function () {
      return this.__site.getFullId() + this.id + '/';
    };

    IfcBuilding.prototype.isAssetOfSpatialInterest = function (asset) {
      let accTypes = [
        'desk/desk',
        'locker-station/lockerStation', 
        'vecos/locker', 
        'loqit/locker', 
        'locker-station/locker'
      ];
      for(let type of accTypes){
        if(type == asset.type){
          return true;
        }
      }
      return false;
    }
    
    IfcBuilding.prototype.getMultiAssiDoors = function () {
      const res = [];
      for (const floor of this.__floors) {
        res.push(...floor.getMultiAssiDoors());
      }
      return res;
    }

    IfcBuilding.prototype.isIn2DPolygonConvFoo = function (spaceShape, beg, end) {
      return isIn2DPolygon(spaceShape, pPCenter2D(beg, end));
    };

    IfcBuilding.prototype.getEntryExitDoors = function () {
      const res = [];
      for (const floor of this.__floors) {
        for (const door of floor.getDoors()) {
          const adjSpaces = door.getAdjacentSpaces();
          if (adjSpaces.length == 1 && !door.__elevator && !adjSpaces[0].isSpaceInSpace()) {
            res.push(door);
          }
        }
      }
      return res;
    };

    IfcBuilding.prototype.getEntryExitNeighbors = function () {
      const res = [];
      const baseFloor = this.getBaseFloor();
      const topFloor = this.getTopFloor();
      for (const floor of this.__floors) {
        for (const neighbor of floor.getCustomNeighbors()) {
          for (const node of neighbor.nodes) {
            if(node){
              if (node.spaceIds.length == 1 && (!node.stairIds || !node.stairIds.length)) {
                const spaces = floor.__spaces.filter(s => {
                  if (s.id == node.spaceIds[0]) {
                    return true;
                  }
                  return false;
                });
                if (spaces.length && !spaces[0].isSpaceInSpace()) {
                  res.push(neighbor);
                }
              } else if ((!node.spaceIds || !node.spaceIds.length) && node.stairIds && node.stairIds.length == 1 && (floor.id == baseFloor.id || floor.id == topFloor.id)) {
                res.push(neighbor);
              }
            }
          }
        }
      }
      return res;
    };

    IfcBuilding.prototype.getBaseFloor = function () {
      let height = Number.MAX_SAFE_INTEGER;
      let res = null;
      for (const floor of this.__floors) {
        if (floor.elevation !== undefined && floor.elevation !== null && floor.elevation < height) {
          height = floor.elevation;
          res = floor;
        }
      }
      return res;
    };

    IfcBuilding.prototype.getTopFloor = function () {
      let height = -Number.MAX_SAFE_INTEGER;
      let res = null;
      for (const floor of this.__floors) {
        if (floor.elevation !== undefined && floor.elevation !== null && floor.elevation > height) {
          height = floor.elevation;
          res = floor;
        }
      }
      return res;
    };

    IfcBuilding.prototype.getStairShapesFloor = function (floor_) {
      let res = [];
      for(let floor of this.__floors){
        for(let stair of floor.__stairs){
          stair.transGeoToFloor(floor_)
          let ps = stair.shapePoints()
          if(stair.isPureRectangleRepresentation()){
            ps.push(ps[0])
          }
          stair.transGeoToFloor(stair.__floor, floor_)
          let fBelow = floor_.getFloorBelow()
          if(floor === floor_ || fBelow !== undefined && fBelow === floor){
            res.push(ps)
          }
        }
      }
      return res;
  };

    IfcBuilding.prototype.addRampEnsemble = function (name) {
      let newRampEnsemble = {
        ramps: [],
        name: name,
        id: getUuid(),
        building: this
      };
      this.__rampEnsembles.push(newRampEnsemble);
      this.setRampEnsembles();
      return newRampEnsemble;
    };

    IfcBuilding.prototype.removeRampEnsemble = function (ensemble) {
      for(let iii=0;iii<this.__rampEnsembles.length;++iii){
        if(this.__rampEnsembles[iii].id === ensemble.id){
          for(let ensembleRamp of ensemble.ramps){
            ensembleRamp.remove();
          }
          this.__rampEnsembles.splice(iii, 1);
          this.setRampEnsembles();
          break;
        }
      }
    };

    IfcBuilding.prototype.getNotInEnsembleRamps = function () {
      let notInEnsembleRamps = [];
      for(const ramp of this.getAllRampsInBuilding()){
        let match = false;
        for(const ensemble of this.__rampEnsembles){
          for(const rampOfEnsemble of ensemble.ramps){
            if(rampOfEnsemble.id === ramp.id){
              match = true;
              break;
            }
          }
        }
        if(!match){
          notInEnsembleRamps.push(ramp);
        }
      }
      return notInEnsembleRamps;
    };

    IfcBuilding.prototype.area = function () {
      let res = 0;
      for(let floor of this.__floors){
        res += floor.area()
      }
      return res;
    };

    IfcBuilding.prototype.getAllRampsInBuilding = function () {
      let allRampsInBuilding = [];
      for(const floor of this.__floors){
        for(const ramp of floor.__ramps){
          allRampsInBuilding.push(ramp);
        }
      }
      return allRampsInBuilding;
    };

    IfcBuilding.prototype.getRampEnsembles = function () {
      let ensembles = [];
      const attributeEntry = this.getIfcElementAttributes().rampEnsembles;
      if(attributeEntry){
        for(let entry of JSON.parse(JSON.stringify(attributeEntry))){
          ensembles.push(entry);
        }
        if(ensembles){
          const allRampsInBuilding = this.getAllRampsInBuilding();
          for(let ensemble of ensembles){
            ensemble.building = this;
            for(let iii=ensemble.ramps.length - 1;iii>=0;--iii){
              for(let ramp of allRampsInBuilding){
                if(ensemble.ramps[iii] === ramp.id){
                  ensemble.ramps[iii] = ramp;
                  break;
                }
              }
              if(typeof(ensemble.ramps[iii]) !== 'object'){
                console.log("didn't found ramp object with id ", ensemble.ramps[iii]);
                ensemble.ramps.splice(iii, 1);
              }
            }
          }
          return ensembles;
        }
      }
      return [];
    };

    IfcBuilding.prototype.setRampEnsembles = function () {
      let sRampEnsembles = [];
      for(let rampEnsemble of this.__rampEnsembles){
        let sRampEnsemble = {ramps: []};
        for(let key of Object.keys(rampEnsemble)){
          if (typeof (rampEnsemble[key]) === 'string') {
            Object.assign(sRampEnsemble, {[key]: rampEnsemble[key]});
          }
        }
        for(let ramp of rampEnsemble.ramps){
          sRampEnsemble.ramps.push(ramp.id);
        }
        sRampEnsembles.push(sRampEnsemble);
      }
      this.getIfcElementAttributes().rampEnsembles = sRampEnsembles;
    };

    IfcBuilding.prototype.lIntersectsWIFC = function (p1, p2, compensateP1 = false, compensateP2 = false, spaces = [], walls = [], assetsShapes = [], parkingBoxes = [], noGoAreas = [], allowSpaceEdgeCross = true) {
      if (p1.stairId && p2.stairId) {
        return true;
      }
      const getFloor = (p) => {
        for (const floor of this.__floors) {
          if (p.z == floor.elevation * 1000) {
            return floor;
          }
        }
        return null;
      };
      const floorP1 = getFloor(p1), floorP2 = getFloor(p2);
      if (floorP1 !== null && floorP2 !== null && floorP1 === floorP2) {
        const compensateAssignment = (p1, p2) => {
          if (p1.x == p2.x && p1.y == p2.y) {
            return p2;
          }
          const p2New = {x: null, y: null};
          Object.assign(p2New, p2);
          const lamb = 500 / Math.sqrt(Math.pow(p1.x - p2.x, 2) + Math.pow(p1.y - p2.y, 2));
          p2New.x = p2New.x + lamb * (p1.x - p2.x);
          p2New.y = p2New.y + lamb * (p1.y - p2.y);
          return p2New;
        };
        const p1_ = compensateP1 === true ? compensateAssignment(p2, p1) : p1;
        const p2_ = compensateP2 === true ? compensateAssignment(p1, p2) : p2;
        const intersects = (objs) => {
          for (const obj of objs) {
            obj.transGeoToFloor(floorP1);
            if (obj.intersectWith(p1_, p2_)) {
              obj.transGeoToFloor(obj.__floor, floorP1);
              return true;
            }
            obj.transGeoToFloor(obj.__floor, floorP1);
          }
          return false;
        };
        const noGoAreaIntersect = (p1, p2, points) => {
          for (let iii = 0; iii < points.length - 1; ++iii) {
            if (psIntersect(p1, p2, points[iii], points[iii + 1])) {
              return true;
            }
          }
          return false;
        };
        // spaces
        for (const space of spaces) {
          if (space.intersectWith(p1_, p2_) > (allowSpaceEdgeCross ? 1 : 0)) {
              return true;
          }
        }
        // walls
        for (const wall of walls) {
          if (wall.intersectWith(p1_, p2_)) {
            return true;
          }
        }
        // no go areas
        for (const noGoArea of noGoAreas) {
          if (noGoAreaIntersect(p1, p2, noGoArea.points)) {
            return true;
          }
        }
        // parking boxes
        for (const space of parkingBoxes) {
          if (space.intersectWith(p1_, p2_) > 1) {
              return true;
          }
        }
        // assets
        for (const assetShape of assetsShapes) {
          for (let iii = 0; iii < assetShape.length - 1; ++iii) {
            if (psIntersect(p1, p2, assetShape[iii], assetShape[iii + 1])) {
              return true;
            }
          }
        }
        // transport elements
        if (intersects(floorP1.getAllRefAndContTransportElems().filter(elem => {
          if (elem.isElevator()) {
            return true;
          }
          return false;
        }))) {
          return true;
        }
        // stairs and ramps
        const floorP1Below = floorP1.getFloorBelow(), stairsBelow = [], rampsBelow = [];
        if (floorP1Below) {
          stairsBelow.push(...floorP1Below.__stairs);
          rampsBelow.push(...floorP1Below.__ramps);
        }
        if (intersects([...floorP1.__stairs, ...stairsBelow]) || intersects([...floorP1.__ramps, ...rampsBelow])) {
          return true;
        }
      }
      return false;
  };

    IfcBuilding.prototype.getMDoors = function (change) {
      let changedDoors = [];
      if(Array.isArray(change.doors)){
        changedDoors = change.doors;
      } else {
        changedDoors.push(change.doors);
      }
      let mDoors = this.multiAssiDoors();
      if(changedDoors.length){
        mDoors = mDoors.filter(door => {
          for(let changedDoor of changedDoors){
            if(door == changedDoor){
              return true;
            }
          }
          return false;
        });
      }
      return mDoors;
    };

    IfcBuilding.prototype.remove = function () {

        this.__stairEnsembles.splice(0, this.__stairEnsembles.length);
        this.setStairEnsembles();

        this.__escalatorEnsembles.splice(0, this.__escalatorEnsembles.length);
        this.setEscalatorEnsembles();

        this.__uncontElevators.splice(0, this.__uncontElevators.length);
        this.setUncontElevators();

        // We cannot use a regular for ... of loop since the element is being removed from the array during iteration
        for (let n = this.__floors.length - 1; n >= 0; n--) {
          this.__floors[n].remove();
        }

        this.__project.getRelAggregate(this.__site).removeRelatedObject(this);
        this.__project.getRelAggregate(this).remove();
        this.__project.getRelContainedInSpatialStructure(this).remove();
        this.__project.unregister(this);
    };

    IfcBuilding.prototype.initialize = function () {
        this.__floors = [];

        return this;
    };

    IfcBuilding.prototype.getSpaceById = function (id) {
      for(let floor of this.__floors){
        for(let space of floor.__spaces){
          if(space.id == id){
            return space;
          }
        }
      }
      return null;
    };

    IfcBuilding.prototype.load = function (errors) {
        this.id = this.__attributes[0];
        this.name = this.__attributes[2];
        this.description = this.__attributes[3];
        this.objectPlacement = this.__attributes[5];
        this.representation = this.__attributes[6];
        this.longName = this.__attributes[7];
        this.compositionType = this.__attributes[8];
        this.elevationOfRefHeight = this.__attributes[9];
        this.elevationOfTerrain = this.__attributes[10];
        this.buildingAddress = this.__attributes[11];
    };

    IfcBuilding.prototype.bind = function (project) {
        this.__project = project;

        if (this.objectPlacement) {
            this.__objectPlacement = project.map[this.objectPlacement];
        }

        if (this.representation) {
            this.__representation = project.map[this.representation];
        }

        if (this.buildingAddress) {
            this.__buildingAddress = project.map[this.buildingAddress];
        }
    };

    IfcBuilding.prototype.getProject = function () {
        return this.__site.__project;
    };

    IfcBuilding.prototype.getFloors = function () {
        return this.__floors;
    };

    IfcBuilding.prototype.getFloor = function (id) {
        for (var n in this.__floors) {
            if (this.__floors[n].id == id) {
                return this.__floors[n];
            }
        }

        return null;
    };

    IfcBuilding.prototype.addFloor = function (project, name) {
        var floor = new IfcBuildingStorey().create(project, name);

        this.__floors.push(floor);
        floor.__building = this;

        project.getRelAggregate(this).addRelatedObject(floor);

        return floor;
    };

    IfcBuilding.prototype.addEscalatorEnsemble = function (name) {
        let newEscalatorEnsemble = {
            escalators: [],
            name: name,
            id: getUuid(),
            building: this
          };
          this.__escalatorEnsembles.push(newEscalatorEnsemble);
          this.setEscalatorEnsembles();
          return newEscalatorEnsemble;
    };

    IfcBuilding.prototype.removeEscalatorEnsemble = function (ensemble) {
        for(let iii=0;iii<this.__escalatorEnsembles.length;++iii){
            if(this.__escalatorEnsembles[iii].id === ensemble.id){
              for(let ensembleEscalator of ensemble.escalators){
                ensembleEscalator.remove();
              }
              this.__escalatorEnsembles.splice(iii, 1);
              this.setEscalatorEnsembles();
              break;
            }
          }
    };

    IfcBuilding.prototype.addStairEnsemble = function (name) {
        let newStairEnsemble = {
            stairs: [],
            name: name,
            id: getUuid(),
            building: this
          };
          this.__stairEnsembles.push(newStairEnsemble);
          this.setStairEnsembles();
          return newStairEnsemble;
    };

    IfcBuilding.prototype.removeStairEnsemble = function (ensemble) {
        for(let iii=0;iii<this.__stairEnsembles.length;++iii){
            if(this.__stairEnsembles[iii].id === ensemble.id){
              for(let ensembleStair of ensemble.stairs){
                ensembleStair.remove();
              }
              this.__stairEnsembles.splice(iii, 1);
              this.setStairEnsembles();
              break;
            }
          }
    };

    IfcBuilding.prototype.removeStairsFromEnsembles = function (stairs) {
        for(let ensemble of this.__stairEnsembles){
            for(let iii=ensemble.stairs.length-1;iii>=0;--iii){
              for(let stair of stairs){
                if(stair === ensemble.stairs[iii]){
                  ensemble.stairs.splice(iii, 1);
                  stair.remove();
                  break;
                }
              }
            }
          }
          this.setStairEnsembles();
    };

    IfcBuilding.prototype.removeRampsFromEnsembles = function (ramps) {
      for(let ensemble of this.__rampEnsembles){
        for(let iii=ensemble.ramps.length-1;iii>=0;--iii){
          for(let ramp of ramps){
            if(ramp === ensemble.ramps[iii]){
              ensemble.ramps.splice(iii, 1);
              ramp.remove();
              break;
            }
          }
        }
      }
      this.setRampEnsembles();
    };

    IfcBuilding.prototype.removeEscalatorsFromEnsembles = function (escalators) {
        for(let ensemble of this.__escalatorEnsembles){
            for(let iii=ensemble.escalators.length-1;iii>=0;--iii){
              for(let escalator of escalators){
                if(escalator === ensemble.escalators[iii]){
                  ensemble.escalators.splice(iii, 1);
                  break;
                }
              }
            }
          }
          this.setEscalatorEnsembles();
    };

    IfcBuilding.prototype.getNotInEnsembleStairs = function () {
      let notInEnsembleStairs = [];
      for(const stair of this.getAllStairsInBuilding()){
        let match = false;
        for(const ensemble of this.__stairEnsembles){
          for(const stairOfEnsemble of ensemble.stairs){
            if(stairOfEnsemble.id === stair.id){
              match = true;
              break;
            }
          }
          if(match){
            break;
          }
        }
        if(!match){
          notInEnsembleStairs.push(stair);
        }
      }
      return notInEnsembleStairs;
    };

    IfcBuilding.prototype.getNotInEnsembleEscalators = function () {
        let notInEnsembleEscalators = [];
        for(const escalator of this.getAllEscalatorsInBuilding()){
          let match = false;
          for(const ensemble of this.__escalatorEnsembles){
            for(const escalatorOfEnsemble of ensemble.escalators){
              if(escalatorOfEnsemble.id === escalator.id){
                match = true;
                break;
              }
            }
          }
          if(!match){
            notInEnsembleEscalators.push(escalator);
          }
        }
        return notInEnsembleEscalators;
    };

    IfcBuilding.prototype.ensembleOfStair = function (stair) {
        for(const ensemble of this.__stairEnsembles){
            for(const ensembleStair of ensemble.stairs){
              if(ensembleStair.id === stair.id){
                return ensemble;
              }
            }
          }
          return null;
    };

    IfcBuilding.prototype.getAllEscalatorsInBuilding = function () {
        let allEscalatorsInBuilding = [];
        for(const floor of this.__floors){
          allEscalatorsInBuilding.push(...floor.getAllEscalators());
        }
        return allEscalatorsInBuilding;
    };

    IfcBuilding.prototype.getAllStairsInBuilding = function () {
        let allStairsInBuilding = [];
        for(const floor of this.__floors){
          for(const stair of floor.__stairs){
            allStairsInBuilding.push(stair);
          }
        }
        return allStairsInBuilding;
    };

    IfcBuilding.prototype.getAllSpacesInBuilding = function () {
        let allSpacesInBuilding = [];
        for(const floor of this.__floors){
          for(const space of floor.__spaces){
            allSpacesInBuilding.push(space);
          }
        }
        return allSpacesInBuilding;
    };

    IfcBuilding.prototype.getBasement = function () {
        let basement;
        for(const floor of this.__floors){
          if(floor.elevation !== undefined && floor.elevation !== null && (basement === undefined || basement.elevation > floor.elevation)){
            basement = floor;
          } else {
            console.log(floor, 'does not has a defined elevation; determination of basement might not work');
          }
        }
        return basement;
    };

    IfcBuilding.prototype.getAllTransportElementsInBuilding = function () {
        let allTransportElementsInBuilding = [];
        for(const floor of this.__floors){
          allTransportElementsInBuilding.push(...floor.__transportElements);
        }
        return allTransportElementsInBuilding;
    };

    IfcBuilding.prototype.dump = function (stream) {
        stream.write('#' + this.oid);
        stream.write('= ');
        stream.write('IFCBUILDING(');
        dumpStringAttribute(stream, this.id, true);
        dumpOidAttribute(stream, null);
        dumpStringAttribute(stream, this.name);
        stream.write(');\n');

        for (var n in this.__floors) {
            this.__floors[n].dump(stream);
        }
    };

    IfcBuilding.prototype.getIfcElementAttributes = function() {
        return this.__project.getIfcElementAttributes(this);
    }

    IfcBuilding.prototype.getIfcElementAttribute = function(name) {
        return this.__project.getIfcElementAttribute(this, name);
    }

    IfcBuilding.prototype.getNavVisWgs84Center = function () {
        return this.getIfcElementAttributes().navVisWgs84Center;
    }

    IfcBuilding.prototype.setNavVisWgs84Center = function (lat, lon, z) {
        this.getIfcElementAttributes().navVisWgs84Center = {lat: lat, lon: lon, z: z};
    }

    IfcBuilding.prototype.getDijkstraCache = function () {
        return this.getIfcElementAttributes().dijkstraCache;
    }

    IfcBuilding.prototype.getEscalatorEnsembles = function () {
        let ensembles = this.getIfcElementAttributes().escalatorEnsembles;
        if(ensembles){
          const allEscalatorsInBuilding = this.getAllEscalatorsInBuilding();
          for(let ensemble of ensembles){
            ensemble.building = this;
            for(let iii=ensemble.escalators.length - 1;iii>=0;--iii){
              for(let escalator of allEscalatorsInBuilding){
                if(ensemble.escalators[iii] === escalator.id){
                  ensemble.escalators[iii] = escalator;
                  break;
                }
              }
              if(typeof(ensemble.escalators[iii]) !== 'object'){
                console.log("didn't found escalator object with id ", ensemble.escalators[iii]);
                ensemble.escalators.splice(iii, 1);
              }
            }
          }
          return ensembles;
        }
        return [];
    }

    IfcBuilding.prototype.getStairEnsembles = function () {
      let stairEnsembles = this.getIfcElementAttributes().stairEnsembles;
      if(stairEnsembles){
        let ensembles = [];
        for(let entry of JSON.parse(JSON.stringify(stairEnsembles))){
          ensembles.push(entry);
        }
        if(ensembles){
          const allStairsInBuilding = this.getAllStairsInBuilding();
          for(let ensemble of ensembles){
            ensemble.building = this;
            for(let iii=ensemble.stairs.length - 1;iii>=0;--iii){
              for(let stair of allStairsInBuilding){
                if(ensemble.stairs[iii] === stair.id){
                  ensemble.stairs[iii] = stair;
                  break;
                }
              }
              if(typeof(ensemble.stairs[iii]) !== 'object'){
                console.log("didn't found stair object with id ", ensemble.stairs[iii]);
                ensemble.stairs.splice(iii, 1);
              }
            }
          }
          return ensembles;
        }
      }
      return [];
    }

    IfcBuilding.prototype.setEscalatorEnsembles = function () {
        let sEscalatorEnsembles = [];
        for(let escalatorEnsemble of this.__escalatorEnsembles){
          let sEscalatorEnsemble = {escalators: []};
          for(let key of Object.keys(escalatorEnsemble)){
            if (typeof (escalatorEnsemble[key]) === 'string') {
              Object.assign(sEscalatorEnsemble, {[key]: escalatorEnsemble[key]});
            }
          }
          for(let escalator of escalatorEnsemble.escalators){
            sEscalatorEnsemble.escalators.push(escalator.id);
          }
          sEscalatorEnsembles.push(sEscalatorEnsemble);
        }
        this.getIfcElementAttributes().escalatorEnsembles = sEscalatorEnsembles;
    }

    IfcBuilding.prototype.setStairEnsembles = function () {
        let sStairEnsembles = [];
        for(let stairEnsemble of this.__stairEnsembles){
          let sStairEnsemble = {stairs: []};
          for(let key of Object.keys(stairEnsemble)){
            if (typeof (stairEnsemble[key]) === 'string') {
              Object.assign(sStairEnsemble, {[key]: stairEnsemble[key]});
            }
          }
          for(let stair of stairEnsemble.stairs){
            sStairEnsemble.stairs.push(stair.id);
          }
          sStairEnsembles.push(sStairEnsemble);
        }
        this.getIfcElementAttributes().stairEnsembles = sStairEnsembles;
    }

    IfcBuilding.prototype.setUncontElevators = function () {
        let uncontElevators = [];
        for(const undcontElevator of this.__uncontElevators){
          uncontElevators.push({name: undcontElevator.name, id: undcontElevator.id});
        }
        this.getIfcElementAttributes().uncontElevators = uncontElevators;
    }

    IfcBuilding.prototype.getUncontElevators = function () {
        let uncontElevators = this.getIfcElementAttributes().uncontElevators;
        if(uncontElevators){
          return uncontElevators;
        } else {
          return [];
        }
    }

    IfcBuilding.prototype.getGroundFloor = function () {
      //simplest version of ground floor determination; see https://standards.buildingsmart.org/IFC/RELEASE/IFC2x3/FINAL/HTML/ for improvement
      for(const floor of this.__floors){
        if(floor.elevation !== undefined && floor.elevation !== null && floor.elevation == 0){
          return floor;
        }
      }
      return null;
    }

    IfcBuilding.prototype.getGroundFloorEntryExitDoors = function () {
      let res = [];
      const groundFloor = this.getGroundFloor();
      if(!groundFloor){
        return res;
      }
      for(let door of groundFloor.getDoors()){
        const adjSpaces = door.getAdjacentSpaces();
        if(adjSpaces.length == 1 && !adjSpaces[0].isSpaceInSpace() && !door.__elevator){
          res.push(door);
        }
      }
      return res;
    }

    IfcBuilding.prototype.getGroundFloorEntryExitNeighbors = function () {
      let res = [];
      const groundFloor = this.getGroundFloor();
      if(!groundFloor){
        return res;
      }
      for(let neighbor of groundFloor.getCustomNeighbors()){
        for(let node of neighbor.nodes){
          if(node.spaceIds.length == 1){
            const spaces = groundFloor.__spaces.filter(s => {
              if(s.id == node.spaceIds[0]){
                return true;
              }
              return false;
            })
            if(spaces.length && !spaces[0].isSpaceInSpace()){
              res.push(neighbor)
            }
          }
        }
      }
      return res;
    }

    IfcBuilding.prototype.getGroundFloorEntryExitPointsFromRamps = function () {
      let res = [];
      const groundFloor = this.getGroundFloor();
      if(!groundFloor){
        return res;
      }
      const floorBelowGroundFloor = groundFloor.getFloorBelow();
      const ramps = this.getAllRampsInBuilding();
      let pointsPool = [];
      for(let space of groundFloor.__spaces){
        pointsPool.push(...space.shapePoints().map(p => [p.x, p.y]));
      }
      const buildingHull = hull(pointsPool.map(pair => pair.map(p => p *= 1/1000)), 100).map(pair => pair.map(p => p *= 1000));
      const buildingHullXY = [];
      for(let p of buildingHull){
        buildingHullXY.push({x: p[0], y: p[1]});
      }
      for(let ramp of ramps){
        for(let p of ramp.getEntryExitPs()){
          Object.assign(p, {ramp: ramp})
          if(p.z == groundFloor.elevation * 1000){
            if(!isIn2DPolygon(buildingHullXY, p)){
              res.push(p);
            } else {
              for(let iii=0;iii<buildingHullXY.length - 1;++iii){
                if(distToSegment2D(p, buildingHullXY[iii], buildingHullXY[iii + 1]) < 500){
                  res.push(p);
                }
              }
            }
          }
        }
      }
      return res;
    }

    IfcBuilding.prototype.createTransportElement = function (project, name) {
        const transportElement = new IfcTransportElement().create(project, name);
        transportElement.__building = this;
        return transportElement;
    }

    IfcBuilding.prototype.checkUncontElevators = function (elevator) {
        let isCont = false;
        for(const floor of this.__floors){
          for(const transportElement of floor.__transportElements){
            if(elevator === transportElement){
              isCont = true;
              break;
            }
          }
        }
        let isInUncont = false;
        for(const elevator_ of this.__uncontElevators){
          if(elevator === elevator_){
            isInUncont = true;
            break;
          }
        }
        if(!isInUncont && !isCont){
          this.__uncontElevators.push(elevator);
          this.setUncontElevators();
        } else if(isInUncont && isCont){
          for(let iii=0;iii<this.__uncontElevators.length; ++iii){
            if(this.__uncontElevators[iii] === elevator){
              this.__uncontElevators.splice(iii, 1);
              this.setUncontElevators();
              break;
            }
          }
        }
    }

    IfcBuilding.prototype.getAllElevators = function () {
        let elevators = [];
        for(const floor of this.__floors){
          elevators.push(...floor.getAllElevators());
        }
        for(const elevator of this.__uncontElevators){
          elevators.push(elevator);
        }
        return elevators;
    }

// #107= IFCPOSTALADDRESS($,$,$,$,('Enter address here'),$,'','Boston','','MA');

    function IfcPostalAddress() {
    }

    IfcPostalAddress.prototype.load = function (project) {
    };

    IfcPostalAddress.prototype.bind = function (project) {
    };

    function IfcBuildingStorey() {
    }

// #144= IFCBUILDINGSTOREY('2OMlvMbj97ih$8FxMJEPuF',#41,'2.OG',$,$,#143,$,'2.OG',.ELEMENT.,8625.);

    IfcBuildingStorey.prototype.create = function (project, name) {
        this.type = 'IFCBUILDINGSTOREY';
        this.id = getUuid();
        this.name = name;

        project.register(this);

        this.__project = project;
        this.__spaces = [];
        this.__doors = [];
        this.__stairs = [];
        this.__ramps = [];
        this.__transportElements = [];
        this.__refTransportElements = [];
        // this.__uncreatedWalls = [];
        this.__walls = [];

        return this;
    };

    IfcBuildingStorey.prototype.getFullId = function () {
      return this.__building.getFullId() + this.id + '/';
    };

    IfcBuildingStorey.prototype.getAssets = function (buildingAssets = null) {
      let res = [];
      if(buildingAssets){
        for(let asset of buildingAssets){
          if(asset.floor === this.id){
            res.push(asset);
          }
        }
      }
      return res;
    };

    IfcBuildingStorey.prototype.getDoors = function () {
      return this.__doors;
    };

    IfcBuildingStorey.prototype.getMultiAssiDoors = function () {
      const res = [];
      for (const door of this.__doors) {
        if (door.__spaces.length > 2 && !door.__elevator || door.__spaces.length > 1 && door.__elevator) {
          res.push(door);
        }
      }
      return res;
    };

    IfcBuildingStorey.prototype.getAssetsShapes = function (buildingAssets = null) {
      let res = [];
      for(let asset of this.getAssets(buildingAssets)){
        let w, h, x = asset.xPosition, y = asset.yPosition;
        //spatial expansion of asset
        if(asset.model3d && asset.model3d.boundingBox){
          w = asset.model3d.boundingBox.width;
          h = asset.model3d.boundingBox.height;
        } else {  //if asset has no spatial expansion: don't take it into account
          continue;
        }
        let assetBox = [
          {x: x - w / 2, y: y - h / 2},
          {x: x + w / 2, y: y - h / 2},
          {x: x + w / 2, y: y + h / 2},
          {x: x - w / 2, y: y + h / 2}];
        assetBox.push(assetBox[0]);
        res.push(assetBox);
      }
      return res;
    };

    IfcBuildingStorey.prototype.makeCustomNeighborFromP = function (p, spaceIds) {
      return {type: 'CUSTOMEDGENEIGHBOR', id: getUuid(), nodes: [{id: getUuid(), x: p.x, y: p.y, z: p.z, type: 'CUSTOMEDGENODE', spaceIds: spaceIds}]};
    };

    IfcBuildingStorey.prototype.setCustomNeighbors = function (neighbors) {
      this.getIfcElementAttributes().neighbors = neighbors;
    };

    IfcBuildingStorey.prototype.getCustomNeighbors = function () {
      let res = this.getIfcElementAttributes().neighbors;
      if(res){
        return res
      } else {
        return []
      }
    };

    IfcBuildingStorey.prototype.area = function () {
      let res = 0;
      const arrOfArrsToXY = (arrOfArrs) => {
        let res = [];
        for(let arr of arrOfArrs){
          res.push({x: arr[0], y: arr[1]});
        }
        return res;
      }
      const arrOfXYToArrs = (arrOfXY) => {
        let res = [];
        for(let arr of arrOfXY){
          res.push([arr.x, arr.y]);
        }
        return res;
      }
      for(let space of this.__spaces){
        res += space.area();
      }
      const atLeastOnePInsidePoly = (shapePs1, shapePs2) => {
        for(let p of shapePs1){
          if(isIn2DPolygon(shapePs2, p)){
            return true;
          }
        }
        for(let p of shapePs2){
          if(isIn2DPolygon(shapePs1, p)){
            return true;
          }
        }
        return false;
      }
      for(let iii=0;iii<this.__spaces.length;++iii){
        for(let jjj=iii+1;jjj<this.__spaces.length;++jjj){
          const shapePsiii = this.__spaces[iii].shapePoints();
          const shapePsjjj = this.__spaces[jjj].shapePoints();
          if(shapePsiii.length >= 4 && shapePsjjj.length >= 4 && atLeastOnePInsidePoly(shapePsiii, shapePsjjj)){
            let psiii = arrOfXYToArrs(shapePsiii);
            let psjjj = arrOfXYToArrs(shapePsjjj);
            psiii.push(psiii[0]);
            psjjj.push(psjjj[0]);
            psiii = [psiii];
            psjjj = [psjjj];
            let polygoniii = turf.polygon(psiii);
            let polygonjjj = turf.polygon(psjjj);
            try {
              let inter = turf.intersect(polygoniii, polygonjjj);
              if(inter && inter.geometry && inter.geometry.coordinates){
                for(let interArea of inter.geometry.coordinates){
                  if(interArea.length >= 3){
                    res -= calcPolygonArea(arrOfArrsToXY(interArea)) / 1000000;
                  }
                }     
              }
            } catch(error){
              //console.log('turf error for intersecting polygons:', error);
            }
          }
        }
      }
      return res;
    };

    IfcBuildingStorey.prototype.bindCustomEntryExitPoints = function (pos, tolerance) {
      const getSpacesCloseToPoint = (x, y, tolerance) => {
        if (tolerance === undefined) {
          tolerance = 100;
        }
        const spaces = [];
        for (const __space of this.__spaces) {
          const geo = __space.getGeo();
          let result;
          if(__space.isPurePolylineRepresentation()){
            result = getClosestPointOnLines({x: x, y: y}, geo.points);
          } else {
            const box_ = geo.getBoundingBox();
            result = getClosestPointOnRectangle(new Vector(x, y, 0), box_.x, box_.y, box_.width, box_.height);
          }
          if (result.distance <= tolerance) {
            spaces.push(__space);
          }
        }
        return spaces;
      };
      const spaces = getSpacesCloseToPoint(pos.x, pos.y, tolerance);
      if(spaces.length){
        for (const space of spaces) {
          console.log('\t<<< Bind custom entry-/exit-point to space ' + space.name);
        }
      }
      return spaces;
    };

    IfcBuildingStorey.prototype.remove = function () {
    // We cannot use a regular for ... of loop since the element is being removed from the array during iteration
    for (let n = this.__spaces.length - 1; n >= 0; n--) {
        this.__spaces[n].remove();
      }

      // We cannot use a regular for ... of loop since the element is being removed from the array during iteration
      for (let n = this.__doors.length - 1; n >= 0; n--) {
        this.__doors[n].remove();
      }

      for (let n = this.__stairs.length - 1; n >= 0; n--) {
        this.__stairs[n].remove();
      }

      for (let n = this.__ramps.length - 1; n >= 0; n--) {
        this.__ramps[n].remove();
      }

      for (let n = this.__walls.length - 1; n >= 0; n--) {
        this.__walls[n].remove();
      }

      for (let n = this.__refTransportElements.length - 1; n >= 0; n--) {
        this.unrefTransportElement(this.__refTransportElements[n]);
      }

    //   this.__uncreatedWalls.splice(0, this.__uncreatedWalls.length);
    //   this.setUncreatedWalls();

      for (let n = this.__transportElements.length - 1; n >= 0; n--) {
        if(this.__transportElements[n].isElevator()){
          let elevator = this.__transportElements[n];
          this.uncontTransportElement(elevator);
          if(elevator.__refFloors.length){
            elevator.__refFloors[0].contTransportElement(elevator);
            elevator.__refFloors[0].unrefTransportElement(elevator);
          }
          this.__building.checkUncontElevators(elevator);
          if(!elevator.__floor && elevator.__representation){
            elevator.__representation.remove(this.__project)
            elevator.__representation = null
            elevator.representation = null
          }
        } else {
          this.__transportElements[n].remove();
        }
      }

      this.__project.getRelAggregate(this.__building).removeRelatedObject(this);
      this.__project.getRelAggregate(this).remove();
      this.__project.getRelContainedInSpatialStructure(this).remove();
      this.__project.getRelReferencedInSpatialStructure(this).remove();
      this.__project.unregister(this);
    };

    IfcBuildingStorey.prototype.initialize = function () {
        this.__spaces = [];
        this.__doors = [];
        this.__stairs = [];
        this.__walls = [];
        this.__ramps = [];

        return this;
    };

    IfcBuildingStorey.prototype.addWall = function (project) {
        const wall = new IfcWall().create(project);
        if (!this.__walls) {
        this.__walls = [];
        }
        wall.__floor = this;
        project.getRelContainedInSpatialStructure(this).addRelatedElement(wall);
        return wall;
    };

    IfcBuildingStorey.prototype.doorsToSpAEl = function () {
      let reassignedDoors = [];
      for(let floor of this.__building.__floors){
        for (let door of floor.__doors) {
          if(door.hasRepresentation()){
            let b = door.getGeo().getBoundingBox();
            const matchResult = floor.getElevatorOrSpaceClosestToPos({x: b.xCenter, y: b.yCenter});
            const center = door.getCenter(), dir = door.getDirection(), assSpaceIds = door.__spaces.slice(0, door.__spaces.length).map(data => data.id), assElevatorId = door.__elevator ? door.__elevator.id : undefined;
            if(matchResult != null && (door.centerOrDirDiff(center, dir) || door.assObjChange(assSpaceIds, assElevatorId))){
              const doorBaseLine = new Vector(matchResult.segment[1].x - matchResult.segment[0].x,
                matchResult.segment[1].y - matchResult.segment[0].y, 0).unit().multiply(door.overallWidth);
              const direction = doorBaseLine.unit();
              door.setPointRepresentation(matchResult.x, matchResult.y, direction.x, direction.y);
              floor.bindDoorsToElASp(door, 500) 
              reassignedDoors.push(door);
            }
          }
        }
      }
      if(this.__component && reassignedDoors.length){
        this.__component.doorChange({doors: reassignedDoors, propagateChange: false})
      }
    };

    IfcBuildingStorey.prototype.bindDoorsToElASp = function (door, tolerance) {
      this.bindDoorToSpaces(door, tolerance)
      this.bindDoorToElevators(door, tolerance)
      let elev = door.__elevator, spaces = door.__spaces
      if(elev && spaces.length == 2){
        let mid
        const b = door.getGeo().getBoundingBox()
        if(door.__objectPlacement && door.__objectPlacement.__relativePlacement && door.__objectPlacement.__relativePlacement.__direction){
          const dir = door.__objectPlacement.__relativePlacement.__direction.coordinates
          mid = {x: b.x + 0.5 * dir.x * door.overallWidth, y: b.y + 0.5 * dir.y * door.overallWidth}
        } else {
          mid =  {x: b.xCenter, y: b.yCenter}
        }
        elev.transGeoToFloor(this)
        const elevShapePs = elev.shapePoints()
        elev.transGeoToFloor(elev.__floor, this)
        if(elev.isPureRectangleRepresentation()){
          elevShapePs.push(elevShapePs[0])
        }
        let seg, dist = Number.MAX_SAFE_INTEGER
        for(let iii=0;iii<elevShapePs.length - 1; ++iii){
          let _dist = distToSegment2D(mid, elevShapePs[iii], elevShapePs[iii + 1])
          if(_dist < dist){
            dist = _dist
            seg = [elevShapePs[iii], elevShapePs[iii + 1]]
          }
        }
        const maxDistToSegment = 1000
        const norm = Math.sqrt(Math.pow(seg[1].y - seg[0].y, 2) + Math.pow(seg[1].x - seg[0].x, 2))
        const perp = {x: -(seg[1].y - seg[0].y) / norm, y: (seg[1].x - seg[0].x) / norm}
        const perp1 = {x: mid.x + perp.x * maxDistToSegment, y: mid.y + perp.y * maxDistToSegment, isInSpace1: false, isInSpace2: false, isInElev: false}
        const perp2 = {x: mid.x - perp.x * maxDistToSegment, y: mid.y - perp.y * maxDistToSegment, isInSpace1: false, isInSpace2: false, isInElev: false}
        const spacePs1 = spaces[0].shapePoints(), spacePs2 = spaces[1].shapePoints()
        perp1.isInSpace1 = isIn2DPolygon(spacePs1, perp1)
        perp2.isInSpace1 = isIn2DPolygon(spacePs1, perp2)
        perp1.isInSpace2 = isIn2DPolygon(spacePs2, perp1)
        perp2.isInSpace2 = isIn2DPolygon(spacePs2, perp2)
        perp1.isInElev = isIn2DPolygon(elevShapePs, perp1)
        perp2.isInElev = isIn2DPolygon(elevShapePs, perp2)
        const remDoorRef = (space, door) => {
          for(let iii = 0; iii<space.__doors.length; ++iii){
            if(space.__doors[iii] === door){
              space.__doors.splice(iii, 1)
              this.getProject().getRelReferencedInSpatialStructure(space).removeRelatedElement(door);
              break
            }
          }
        }
        if(perp1.isInSpace1 && perp1.isInElev && !perp1.isInSpace2 || perp2.isInSpace1 && perp2.isInElev && !perp2.isInSpace2){
          console.log('<<< Unbind ', door.name, ' from ', spaces[0].name)
          remDoorRef(spaces[0], door)
          spaces.splice(0, 1)
        } else if(perp1.isInSpace2 && perp1.isInElev && !perp1.isInSpace1 || perp2.isInSpace2 && perp2.isInElev && !perp2.isInSpace1){
          console.log('<<< Unbind ', door.name, ' from ', spaces[1].name)
          remDoorRef(spaces[1], door)
          spaces.splice(1, 1)
        } else {
          console.log('the space/elevator assignment for ', door.name, ' will cause problems')
        }
      } else if(elev && spaces.length > 2){
        console.log('the space/elevator assignment for ', door.name, ' will cause problems')
      }
    };

    IfcBuildingStorey.prototype.getFloorBelow = function () {
        let floorBelow;
        if(this.elevation !== undefined && this.elevation !== null){
          for(const floor of this.__building.__floors){
            if(this !== floor && floor.elevation !== undefined && floor.elevation !== null && floor.elevation < this.elevation &&
              (!floorBelow || floorBelow.elevation < floor.elevation)){
                floorBelow = floor;
            }
          }
        }
        return floorBelow;
    }

    IfcBuildingStorey.prototype.load = function (errors) {
        this.id = this.__attributes[0];
        this.name = this.__attributes[2];
        this.description = this.__attributes[3];
        this.objectPlacement = this.__attributes[5];
        this.representation = this.__attributes[6];
        this.longName = this.__attributes[7];
        this.compositionType = this.__attributes[8];
        this.elevation = this.__attributes[9];
    };

    IfcBuildingStorey.prototype.bind = function (project) {
        this.__project = project;
        this.__spaces = [];
        this.__doors = [];
        this.__windows = [];
        this.__walls = [];
        this.__stairs = [];
        this.__ramps = [];
        this.__transportElements = [];
        this.__refTransportElements = [];

        if (this.objectPlacement) {
          this.__objectPlacement = project.map[this.objectPlacement];
        }

        if (this.representation) {
          this.__representation = project.map[this.representation];
        }
    };

    /**
     * @deprecated
     *
     * @returns {*}
     */
    IfcBuildingStorey.prototype.getProject = function () {
        return this.__building.__site.__project;
    };

    IfcBuildingStorey.prototype.getPlacement = function () {
        if (!this.__objectPlacement) {
            return null;
        }

        return this.__objectPlacement.getPlacement();
    };

    IfcBuildingStorey.prototype.hasPredefinedFloorPlan = function () {
        return this.floorPlanSvgFileUri != null;
    };

    IfcBuildingStorey.prototype.hasFloorplan = function () {
        if (this.representation) {
            return true;
          }

          for (const n in this.__spaces) {
            if (this.__spaces[n].representation) {
              return true;
            }
          }

          for (const n in this.__stairs) {
            if (this.__stairs[n].representation) {
              return true;
            }
          }

          for (const n in this.__walls) {
            if (this.__walls[n].representation) {
              return true;
            }
          }

          for (const n in this.__transportElements) {
            if (this.__transportElements[n].representation) {
              return true;
            }
          }

          for (const n in this.__refTransportElements) {
            if (this.__refTransportElements[n].representation) {
              return true;
            }
          }

          return false;
    };

    IfcBuildingStorey.prototype.getSpaces = function (id) {
        return this.__spaces;
    };

    IfcBuildingStorey.prototype.getSpace = function (id) {
        for (var n in this.__spaces) {
            if (this.__spaces[n].id == id) {
                return this.__spaces[n];
            }
        }

        return null;
    };

    /**
     *
     * @param position
     * @returns {*}
     */
    IfcBuildingStorey.prototype.getSpaceClosestToPosition = function (position) {
      let matchSpace, matchInSpaces = [];
      let matchResult, matchInSpaceResults = [];
      let distance = Number.MAX_SAFE_INTEGER;
      for (const space of this.__spaces) {
        try {
          const geo = space.getGeo();
          let result;
          if(space.isPurePolylineRepresentation()){
            result = getClosestPointOnLines(position, geo.points);
          } else{  //else -> because it was implemented like this by now
            const boundingBox = geo.getBoundingBox();
            result = getClosestPointOnRectangle(position, boundingBox.x, boundingBox.y, boundingBox.width, boundingBox.height);
          }
          if (result.distance < distance) {
            distance = result.distance;
            matchSpace = space;
            matchResult = result;
          }
          if(isIn2DPolygon(space.shapePoints(), position)){
            matchInSpaces.push(space);
            matchInSpaceResults.push(result);
          }
        } catch (error) {
          console.error(error);
        }
      }
      if(matchResult){
        let setSpace = true;
        if(matchInSpaces.length == 1 && matchInSpaceResults.length == 1 && Math.abs(matchInSpaceResults[0].distance - matchResult.distance) <= 250){
            matchResult = Object.assign({}, matchInSpaceResults[0]);
            matchResult.space = matchInSpaces[0];
            setSpace = false;
        } else if (matchInSpaces.length >= 1 && matchInSpaceResults.length >= 1) {
          console.log('position is located in more than one space; not possible to determine a space closest to this position which includes this information');
        }
        if(setSpace){
          matchResult.space = matchSpace;
        }
      }
      return matchResult;
    };

    IfcBuildingStorey.prototype.getClosestPositionOnStair = function (position, stair) {
      stair.transGeoToFloor(this);
      if(stair.isPurePolylineRepresentation()){
        let res = getClosestPointOnLines(position, stair.shapePoints());
        stair.transGeoToFloor(stair.__floor, this);
        return res
      } else {
        const geo = stair.getGeo();
        const box = geo.getBoundingBox();
        let res = getClosestPointOnRectangle(position, box.x, box.y, box.width, box.height);
        stair.transGeoToFloor(stair.__floor, this);
        return res
      }
    };

    IfcBuildingStorey.prototype.getClosestPositionOnRamp = function (position, stair) {
      ramp.transGeoToFloor(this);
      let res = getClosestPointOnLines(position, ramp.shapePoints());
      ramp.transGeoToFloor(ramp.__floor, this);
      return res
    };

    IfcBuildingStorey.prototype.getElevatorClosestToPosition = function (position) {
        let matchElevator;
        let matchResult;
        let distance = Number.MAX_SAFE_INTEGER;
        for (const elevator of this.getAllRefAndContTransportElems().filter(elem => {
          if(elem.isElevator()){
            return true
          }
          return false
        })){
          try{
            elevator.transGeoToFloor(this)
            const geo = elevator.getGeo()
            let result;
            if(elevator.isPurePolylineRepresentation()){
              result = getClosestPointOnLines(position, geo.points);
            } else {
              const box = geo.getBoundingBox();
              result = getClosestPointOnRectangle(position, box.x, box.y, box.width, box.height);
            }
            if (result.distance < distance) {
              distance = result.distance;
              matchElevator = elevator;
              matchResult = result;
            }
            elevator.transGeoToFloor(elevator.__floor, this)
          }catch (error) {
            console.error(error);
          }
        }
        if(matchResult){
          matchResult.elevator = matchElevator;
        }
        return matchResult;
    };


    IfcBuildingStorey.prototype.getElevatorOrSpaceClosestToPos = function (position) {
        let spaceRes = this.getSpaceClosestToPosition(position);
        let elevatorRes = this.getElevatorClosestToPosition(position);
        if(spaceRes && !elevatorRes || spaceRes && elevatorRes && spaceRes.distance < elevatorRes.distance){
          return spaceRes
        } else if(!spaceRes && elevatorRes || spaceRes && elevatorRes && spaceRes.distance > elevatorRes.distance){
          return elevatorRes
        }
        return null
      }


    IfcBuildingStorey.prototype.getFloorAbove = function (subAmountFloors = undefined) {
        let floorAbove;
        if(this.elevation !== undefined && this.elevation !== null){
          for(const floor of subAmountFloors ? subAmountFloors : this.__building.__floors){
            if(floor.id !== this.id && floor.elevation !== undefined && floor.elevation !== null && floor.elevation > this.elevation &&
              (!floorAbove || floor.elevation < floorAbove.elevation)){
                floorAbove = floor;
            }
          }
        }
        return floorAbove;
    };

    /**
     *
     * @param x
     * @param y
     * @param tolerance
     * @returns {Array}
     */
    IfcBuildingStorey.prototype.getSpaceWithWallsCloseToPoint = function (x, y, tolerance) {
        if (tolerance == undefined) {
            tolerance = 100; // 10 cm by default
        }

        var spaces = [];

        for (var n in this.__spaces) {
            var geo = this.__spaces[n].getGeo();
            var box = geo.getBoundingBox();

            // TODO Also consider polylines

            var result = getClosestPointOnRectangle(new Vector(x, y), box.x, box.y, box.width, box.height);

            console.log(this.__spaces[n].name + ': ' + result.distance);

            if (result.distance <= tolerance) {
                spaces.push(this.__spaces[n]);
            }
        }

        return spaces;
    };

    IfcBuildingStorey.prototype.addSpace = function (project, name) {
        var space = new IfcSpace().create(project, name);

        if (!this.__spaces) {
            this.__spaces = [];
        }

        this.__spaces.push(space);

        space.__project = project;
        space.__floor = this;

        project.getRelAggregate(this).addRelatedObject(space);

        return space;
    };

    IfcBuildingStorey.prototype.getNextName = function (list, baseName, index) {
      for (const n in list) {
        if (list[n].name === baseName + ' ' + index) {
          return this.getNextName(list, baseName, ++index);
        }
      }
      return baseName + ' ' + index;
    };

    IfcBuildingStorey.prototype.addDoor = function (project, name = null) {

      if(!name){
        name = this.getNextName(this.__doors, "Door", 1);
      }
  
      const door = new IfcDoor().create(project, name);
  
      if (!this.__doors) {
        this.__doors = [];
      }
  
      door.__project = project;
      door.__floor = this;
  
      // this.__doors.push(door); // Not required since below statement seems to do this too!
      project.getRelContainedInSpatialStructure(this).addRelatedElement(door);
  
      return door;
    };

    IfcBuildingStorey.prototype.addStair = function (name) {
      const stair = new IfcStair().create(this.__project, name);
      this.__project.getRelContainedInSpatialStructure(this).addRelatedElement(stair);
      return stair;
    };

    IfcBuildingStorey.prototype.addStair = function (name) {
      const ramp = new IfcRamp().create(this.__project, name);
      this.__project.getRelContainedInSpatialStructure(this).addRelatedElement(ramp);
      return ramp;
    };

    IfcBuildingStorey.prototype.getAllElevators = function () {
        let elevators = [];
        for(const transportElement of this.__transportElements){
          if(transportElement.isElevator()){
            elevators.push(transportElement);
          }
        }
        return elevators;
    };

    IfcBuildingStorey.prototype.getAllMovingWalkways = function () {
        let movingWalkways = [];
        for(const transportElement of this.__transportElements){
          if(transportElement.isMovingWalkway()){
            movingWalkways.push(transportElement);
          }
        }
        return movingWalkways;
    };

    IfcBuildingStorey.prototype.getAllEscalators = function () {
        let escalators = [];
        for(const transportElement of this.__transportElements){
          if(transportElement.isEscalator()){
            escalators.push(transportElement);
          }
        }
        return escalators;
    };

    IfcBuildingStorey.prototype.getAllRefAndContTransportElems = function () {
        let tElems = [...this.__transportElements];
        for(const refTElem of this.__refTransportElements){
          let alreadyIn = false;
          for(const tElem of this.__transportElements){
            if(refTElem === tElem){
              alreadyIn = true;
              break;
            }
          }
          if(!alreadyIn){
            tElems.push(refTElem);
          }
        }
        return tElems;
    };

    IfcBuildingStorey.prototype.contTransportElement = function (transportElement) {
        this.__project.getRelContainedInSpatialStructure(this).addRelatedElement(transportElement);
        transportElement.__floor = this;
        return transportElement;
    };

    IfcBuildingStorey.prototype.uncontTransportElement = function (transportElement) {
        this.__project.getRelContainedInSpatialStructure(this).removeRelatedElement(transportElement);
        for(let iii = 0; iii<this.__transportElements.length; ++iii){
          if(this.__transportElements[iii] === transportElement){
            this.__transportElements.splice(iii, 1);
            transportElement.__floor = undefined;
            break;
          }
        }
    };

    IfcBuildingStorey.prototype.refTransportElement = function (transportElement) {
        this.__project.getRelReferencedInSpatialStructure(this).addRelatedElement(transportElement);
        let alreadyRef = false;
        for(const floor of transportElement.__refFloors){
          if(floor === this){
            alreadyRef = true;
            break;
          }
        }
        if(!alreadyRef){
          transportElement.__refFloors.push(this);
        }
        return transportElement;
    };

    IfcBuildingStorey.prototype.unrefTransportElement = function (transportElement) {
        this.__project.getRelReferencedInSpatialStructure(this).removeRelatedElement(transportElement);
        for(let iii = 0; iii<this.__refTransportElements.length; ++iii){
          if(this.__refTransportElements[iii] === transportElement){
            this.__refTransportElements.splice(iii, 1);
            break;
          }
        }
        for(let iii = 0; iii<transportElement.__refFloors.length; ++iii){
          if(transportElement.__refFloors[iii] === this){
            transportElement.__refFloors.splice(iii, 1);
            break;
          }
        }
    };

    IfcBuildingStorey.prototype.stairOfEnsemble = function (ensemble) {
        for(let stair of this.__stairs){
            for(let ensembleStair of ensemble.stairs){
              if(stair.id === ensembleStair.id){
                return stair;
              }
            }
          }
          return null;
    };

    IfcBuildingStorey.prototype.rampOfEnsemble = function (ensemble) {
      for(let ramp of this.__ramps){
        for(let ensembleRamp of ensemble.ramps){
          if(ramp.id === ensembleRamp.id){
            return ramp;
          }
        }
      }
      return null;
    };

    IfcBuildingStorey.prototype.escalatorOfEnsemble = function (ensemble) {
        for(let transportElement of this.__transportElements){
            if(transportElement.isEscalator()){
              for(let ensembleEscalator of ensemble.escalators){
                if(transportElement.id === ensembleEscalator.id){
                  return transportElement;
                }
              }
            }
          }
          return null;
    };

    IfcBuildingStorey.prototype.dump = function (stream) {
        stream.write('#' + this.oid);
        stream.write('= ');
        stream.write('IFCBUILDINGSTOREY(');
        dumpStringAttribute(stream, this.id, true);
        dumpOidAttribute(stream, null, false);
        dumpStringAttribute(stream, this.name, false);
        stream.write(');\n');

        for (const n in this.__spaces) {
          this.__spaces[n].dump(stream);
        }

        for (const n in this.__doors) {
          this.__doors[n].dump(stream);
        }

        for (const n in this.__walls) {
          this.__walls[n].dump(stream);
        }

        for (const n in this.__stairs) {
          this.__stairs[n].dump(stream);
        }
    };

    IfcBuildingStorey.prototype.getBoundingBox = function () {
        var viewBox = {
            left: Number.MAX_SAFE_INTEGER,
            top: Number.MAX_SAFE_INTEGER,
            right: Number.MIN_SAFE_INTEGER,
            bottom: Number.MIN_SAFE_INTEGER
        };
        var geo;
        var boundingBox;

        for (var i in this.__walls) {
            var wall = this.__walls[i];

            geo = wall.getGeo();
            boundingBox = geo.getBoundingBox();

            viewBox.left = Math.min(viewBox.left, boundingBox.x);
            viewBox.top = Math.min(viewBox.top, boundingBox.y);
            viewBox.right = Math.max(viewBox.right, boundingBox.x + boundingBox.width);
            viewBox.bottom = Math.max(viewBox.bottom, boundingBox.y + boundingBox.height);
        }

        // TODO Can probably omit

        for (var i in this.__doors) {
            var door = this.__doors[i];

            geo = door.getGeo();
            boundingBox = geo.getBoundingBox();

            viewBox.left = Math.min(viewBox.left, boundingBox.x);
            viewBox.top = Math.min(viewBox.top, boundingBox.y);
            viewBox.right = Math.max(viewBox.right, boundingBox.x + boundingBox.width);
            viewBox.bottom = Math.max(viewBox.bottom, boundingBox.y + boundingBox.height);
        }

        for (var i in this.__spaces) {
            var space = this.__spaces[i];

            geo = space.getGeo();
            boundingBox = geo.getBoundingBox();

            viewBox.left = Math.min(viewBox.left, boundingBox.x);
            viewBox.top = Math.min(viewBox.top, boundingBox.y);
            viewBox.right = Math.max(viewBox.right, boundingBox.x + boundingBox.width);
            viewBox.bottom = Math.max(viewBox.bottom, boundingBox.y + boundingBox.height);
        }

        return viewBox;
    };

    IfcBuildingStorey.prototype.renderFloorPlanSvg = function (svg, options) {
            // Scale is Pixel/Meter

            for (var i in this.__walls) {
                this.__walls[i].renderFloorPlanSvg(svg, options);
            }

            for (const __door of this.__doors) {
                __door.renderFloorPlanSvg(svg, options);
              }

              for (const __space of this.__spaces) {
                __space.renderFloorPlanSvg(svg, options);
              }

              let floorBelow = this.getFloorBelow(), stairsBelow = []
              if(floorBelow){
                stairsBelow.push(...floorBelow.__stairs)
              }
              for (const __stair of [...this.__stairs, ...stairsBelow]) {
                __stair.renderFloorPlanSvg(svg, options, this);
              }

              let rampsBelow = [];
              if(floorBelow){
                rampsBelow.push(...floorBelow.__ramps)
              }
              for (const __ramp of [...this.__ramps, ...rampsBelow]) {
                __ramp.renderFloorPlanSvg(svg, options, this);
              }

              for (const __transportElement of this.__transportElements) {
                __transportElement.renderFloorPlanSvg(svg, options, this);
              }

              for (const __refTransportElement of this.__refTransportElements) {
                __refTransportElement.renderFloorPlanSvg(svg, options, this);
              }
        };

    /**
     *
     * @param project
     */
    IfcBuildingStorey.prototype.bindDoorsToSpaces = function (tolerance) {
        for (const __door of this.__doors) {
            this.bindDoorToSpaces(__door, tolerance);
          }
    };

    /**
     *
     * @param project
     */
    IfcBuildingStorey.prototype.bindDoorToSpaces = function (door, tolerance) {
        for(let iii=0;iii<this.__spaces.length;++iii){
            this.getProject().getRelReferencedInSpatialStructure(this.__spaces[iii]).removeRelatedElement(door);
          }
          door.__spaces = []
          const getSpacesWithWallsCloseToPoint = (x, y, tolerance) => {
            if (tolerance === undefined) {
              tolerance = 100; // 10 cm by default
            }
            const spaces = [];
            for (const __space of this.__spaces) {
              const geo = __space.getGeo();
              let result;
              if(__space.isPurePolylineRepresentation()){
                result = getClosestPointOnLines({x: x, y: y}, geo.points);
              }else{  //else -> because it was implemented like this by now
                const box_ = geo.getBoundingBox();
                result = getClosestPointOnRectangle(new Vector(x, y, 0), box_.x, box_.y, box_.width, box_.height);
              }
              if (result.distance <= tolerance) {
                spaces.push(__space);
              }
            }
            return spaces;
          }
          const box = door.getGeo().getBoundingBox();
          const spaces = getSpacesWithWallsCloseToPoint(box.x, box.y, tolerance);
          if(spaces.length){
            for (const space of spaces) {
              console.log('\t<<< Bind door ' + door.name + ' to space ' + space.name);
              this.getProject().getRelReferencedInSpatialStructure(space).addRelatedElement(door);
            }
          }
    };

    IfcBuildingStorey.prototype.bindDoorToElevators = function (door, tolerance) {
        let elevatorsOnFloor = this.getAllRefAndContTransportElems().filter(elem => {
          if(elem.isElevator()){
            return true
          }
          return false
        })
        for(let iii=0;iii<elevatorsOnFloor.length;++iii){
          this.getProject().getRelReferencedInSpatialStructure(elevatorsOnFloor[iii]).removeRelatedElement(door);
        }
        door.__elevator = undefined
        const getElevatorsWithShapeLinesCloseToPoint = (x, y, tol) => {
          if (tol === undefined) {
            tol = 100;
          }
          const elevators = [];
          for (const elevator of elevatorsOnFloor) {
            elevator.transGeoToFloor(this)
            const geo = elevator.getGeo()
            let result;
            if(elevator.isPurePolylineRepresentation()){
              result = getClosestPointOnLines({x: x, y: y}, geo.points);
            } else {
              const box_ = geo.getBoundingBox();
              result = getClosestPointOnRectangle(new Vector(x, y, 0), box_.x, box_.y, box_.width, box_.height);
            }
            if (result.distance <= tolerance) {
              elevators.push(elevator);
            }
            elevator.transGeoToFloor(elevator.__floor, this)
          }
          return elevators;
        }
        const box = door.getGeo().getBoundingBox()
        const elevators = getElevatorsWithShapeLinesCloseToPoint(box.x, box.y, tolerance)
        if(elevators.length == 1){
          console.log('\t<<< Bind door ' + door.name + ' to elevator ' + elevators[0].name);
          this.getProject().getRelReferencedInSpatialStructure(elevators[0]).addRelatedElement(door);
        } else if(elevators.length > 1){
          console.log("it's not intended to bind one door to more than one elevator - door will not get bound to any elevator")
        }
    };

    IfcBuildingStorey.prototype.getIfcElementAttributes = function() {
        return this.__project.getIfcElementAttributes(this);
    }

    IfcBuildingStorey.prototype.getIfcElementAttribute = function(name) {
        return this.__project.getIfcElementAttribute(this, name);
    }

    IfcBuildingStorey.prototype.getOrigin = function () {
        return this.getIfcElementAttributes().origin;
    }

    IfcBuildingStorey.prototype.setOrigin = function (x, y) {
        this.getIfcElementAttributes().origin = {x: x, y: y};
    }

    IfcBuildingStorey.prototype.getNavVisOrigin = function () {
        return this.getIfcElementAttributes().navVisOrigin;
    }

    IfcBuildingStorey.prototype.setNavVisOrigin = function (x, y) {
        this.getIfcElementAttributes().navVisOrigin = {x: x, y: y};
    }

    IfcBuildingStorey.prototype.getArchilogicOrigin = function () {
        return this.getIfcElementAttributes().archilogicOrigin;
    }

    IfcBuildingStorey.prototype.setArchilogicOrigin = function (x, y) {
        this.getIfcElementAttributes().archilogicOrigin = {x: x, y: y};
    }

    function IfcSpace() {
    }

// #218= IFCSPACE('3OrU2VYXn9rPL0hrsHQNl7',#41,'02.A.01',$,$,#201,#214,'ELT',.ELEMENT.,.INTERNAL.,$);

    IfcSpace.prototype.getClosestPositionOnSpace = function (position) {
      const geo = this.getGeo();
      if(this.isPurePolylineRepresentation()){
        let res = getClosestPointOnLines(position, this.shapePoints());
        return res
      } else {
        const box = geo.getBoundingBox();
        let res = getClosestPointOnRectangle(position, box.x, box.y, box.width, box.height);
        return res
      }
    };

    IfcSpace.prototype.removeNoGoAreaByClick = function (clickPos) {
      let noGoAreas = this.getNoGoAreas();
      let index = null, shortestDist = Number.MAX_SAFE_INTEGER;
      for(let [iii, noGoArea] of noGoAreas.entries()){
        for(let jjj=0;jjj<noGoArea.points.length - 1;++jjj){
          let newDist = distToSegment2D(clickPos, noGoArea.points[jjj], noGoArea.points[jjj + 1]);
          if(newDist < shortestDist){
            index = iii;
            shortestDist = newDist;
          }
        }
      }
      if(index !== null){
        noGoAreas.splice(index, 1);
        this.setNoGoAreas(noGoAreas);
      }
    };

    /**
     * Encodes a location within a Space as a 6 character alphanumeric string, which is as follows:
     * First 4 characters: Base32 encoded Space UUID
     * Last 2 characters: Base32 encoded (x, y) as represented within a 32x32 grid relative to the Space
     *
     * Base32 encoding uses the characters 0-9 and A-Z (excluding I, L, O and U)
     * See https://www.crockford.com/base32.html for additional information on Base32 encoding
     *
     * private
     * param {*} space
     * param {number} x
     * param {number} y
     * returns {string}
     */
    IfcSpace.prototype.getLocationCode = function (x, y) {
      const center = this.getShapeCenter();
      if (center) {
        x = x == null ? center.x : x;
        y = y == null ? center.y : y;
      }
  
      // Provided point is not in the shape
      if (!this.isInSpaceShape({x, y})) {
        console.error(`getAssetLocationCode(): Point (${x}, ${y}) is not in the Space ${this.name}`);
        return;
      }
  
      const bBox = this.getGeo().getBoundingBox();
      if (!bBox) {
        return;
      }
  
      // See https://www.crockford.com/base32.html
      const encoder = new UuidEncoder('base32');
      const encodedUuid = encoder.encode(this.id);
  
      const encodedSpace = encodedUuid.substring(0, 4).toUpperCase();
  
      const xAdjusted = x - bBox.left;
      const yAdjusted = y - bBox.top;
  
      const NUM_GRID_COLUMNS = 32;
      const NUM_GRID_ROWS = 32;
      const unitWidth = bBox.width / NUM_GRID_COLUMNS;
      const unitHeight = bBox.height / NUM_GRID_ROWS;
  
      let cellX = Math.floor(xAdjusted / unitWidth);
      let cellY = Math.floor(yAdjusted / unitHeight);
  
      cellX = cellX < 0 ? 0 : cellX;
      cellX = cellX >= NUM_GRID_COLUMNS ? NUM_GRID_COLUMNS - 1 : cellX;
  
      cellY = cellY < 0 ? 0 : cellY;
      cellY = cellY >= NUM_GRID_ROWS ? NUM_GRID_ROWS - 1 : cellY;
  
      // See https://www.crockford.com/base32.html
      const symbols = '0123456789ABCDEFGHJKMNPQRSTVWXYZ';
  
      const encodedX = symbols[cellX];
      const encodedY = symbols[cellY];
  
      const locationCode = `${encodedSpace}${encodedX}${encodedY}`.toUpperCase();
  
      return locationCode;
    };

    IfcSpace.prototype.getNoGoAreas = function () {
      let noGoAreas = this.getIfcElementAttributes().noGoAreas;
      if(!noGoAreas){
        noGoAreas = []
      }
      return noGoAreas;
    };

    IfcSpace.prototype.setNoGoAreas = function (noGoAreas) {
      this.getIfcElementAttributes().noGoAreas = noGoAreas
    };

    IfcSpace.prototype.addNoGoArea = function (points) {
      let existing = this.getNoGoAreas();
      existing.push({
        points: points
      });
      this.setNoGoAreas(existing);
    };

    IfcSpace.prototype.isSpaceInSpace = function () {
      for(let space of this.__floor.__spaces){
        if(space !== this){
          const spacesInSpace = space.getSpacesInSpace();
          for(let spaceInSpace of spacesInSpace){
            if(spaceInSpace === this){
              return true;
            }
          }
        }
      }
      return false;
    };

    IfcSpace.prototype.getNeighbors = function (knownNodes, allowSpaceEdgeCross = true, apprPartsTANCBWCache = null, isClientRun = false, stairShapesBuilding = {}){
      let neighbors = [];
      neighbors.push(...this.getAdjacentDoors().filter(door => {
        if(door.__elevator){
          return false
        }
        return true
      }));
      neighbors.push(...this.getAdjacentStairs());
      neighbors.push(...this.getAdjacentRamps());
      neighbors.push(...this.getAdjacentTransportElements());
      if(allowSpaceEdgeCross){  //for a simplified scenario without walls the navigation can't cross space edges
      neighbors.push(...this.getArtificialEdgeNeighbors(knownNodes, apprPartsTANCBWCache, isClientRun, stairShapesBuilding[this.__floor.id]));
      }
      neighbors.push(...this.getCustomEdgeNeighbors());
      return neighbors;
    }

    IfcSpace.prototype.genFreeSegmentsPsArrs = function (freeSegments){
      let psArrs = [];
      const everyXmm = 75;
      for(const seg of freeSegments){
        const dirX = seg[1].x - seg[0].x;
        const dirY = seg[1].y - seg[0].y;
        const numSegs = Math.floor(Math.sqrt(Math.pow(dirX, 2) + Math.pow(dirY, 2)) / everyXmm);
        let psArr = [];
        for(let iii=0;iii<=numSegs;++iii){
          psArr.push({x: seg[0].x + dirX * iii / numSegs, y: seg[0].y + dirY * iii / numSegs})
        }
        psArrs.push(psArr)
      }
      return psArrs
    }

    IfcSpace.prototype.getFreeSegsPsArrPairs = function (freeSegsPsArr, otherFreeSegsPsArr){
      let pPairs = []
      for(const thisP of freeSegsPsArr){
        let closestP = null, minDist = Number.MAX_SAFE_INTEGER;
        for(const oneOtherP of otherFreeSegsPsArr){
          const dist = dist2D(thisP, oneOtherP)
          if(dist < minDist){
            minDist = dist;
            closestP = oneOtherP
          }
        }
        pPairs.push([thisP, closestP])
      }
      return pPairs;
    }

    IfcSpace.prototype.getCustomEdgeNeighbors = function (){
      let res = [];
      const sISs = this.getSpacesInSpace();
      for(let neighbor of this.__floor.getCustomNeighbors()){
        for(let node of neighbor.nodes){
          let found = false;
          if(node){
            for(let id of node.spaceIds){
              if(id == this.id){
                res.push(neighbor);
                found = true;
                break;
              }
              if(node.spaceIds.length == 1){
                for(let sIS of sISs){
                  if(id == sIS.id){
                    res.push(neighbor);
                    found = true;
                    break;
                  }
                }
              }
            }
          }
          if(found){
            break;
          }
        }
      }
      return res;
    }

    IfcSpace.prototype.getAdjacentDoors = function () {
      let res = []
      for(let door of this.__floor.__doors){
        for(let space of door.getAdjacentSpaces()){
          if(space === this){
            res.push(door)
          }
        }
      }
      return res
    };

    IfcSpace.prototype.scale = function (scaleFactor) {
      if(this.isPurePolylineRepresentation()){
        let shapePs = this.shapePoints()
        for(let p of shapePs){
          p.x *= scaleFactor
          p.y *= scaleFactor
        }
        this.setPolylineRepresentation(shapePs)
      } else {
        let geo = this.getGeo()
        geo.x *= scaleFactor
        geo.y *= scaleFactor
        geo.width *= scaleFactor
        geo.height *= scaleFactor
        this.setRectangleRepresentation(geo.x, geo.y, geo.width, geo.height)
      }
      let newNGAs = [];
      for(let nGA of this.getNoGoAreas()){
        for(let p of nGA.points){
          p.x *= scaleFactor
          p.y *= scaleFactor
        }
        newNGAs.push(nGA);
      }
      this.setNoGoAreas(newNGAs);
  };

  IfcSpace.prototype.getFreeSegments = function (apprPartsTANCBWCache = null, isClientRun = false, spaceNeighborsSegs_ = null) {
    const apprPartsTANCBW = this.getApprPartsTANCBWCache(apprPartsTANCBWCache);
    const spaceNeighborsSegs = spaceNeighborsSegs_ !== null ? spaceNeighborsSegs_ : this.spaceNeighborsSegments(apprPartsTANCBWCache);
    let res = [];
    const minFreeSegDist = 750;
    for(let apprPartTINCBW of apprPartsTANCBW){
      const coherenceMaxThres = 50;
      let freeSegPsCoherNeighs = [];
      for(let neighSeg of spaceNeighborsSegs){
        if(distToSegment2D(neighSeg.seg[0], apprPartTINCBW[0], apprPartTINCBW[1]) < coherenceMaxThres && distToSegment2D(neighSeg.seg[1], apprPartTINCBW[0], apprPartTINCBW[1]) < coherenceMaxThres){
          freeSegPsCoherNeighs.push(...this.genFreeSegmentsPsArrs([neighSeg.seg]));          
        }
      }
      const freeSegPsApprPartTINCBW = this.genFreeSegmentsPsArrs([apprPartTINCBW]);
      if(freeSegPsCoherNeighs.length){
        const pairs = this.getFreeSegsPsArrPairs(freeSegPsApprPartTINCBW[0], freeSegPsCoherNeighs[0]);
        let lastiii = 0;
        for(let iii=0;iii<pairs.length - 2;++iii){
          if(pairs[iii][1] != pairs[iii + 1][1] && pairs[iii + 1][1] != pairs[iii + 2][1]){
            if(iii - lastiii > 1 && dist2D(pairs[lastiii][0], pairs[iii][0]) >= minFreeSegDist){
              res.push([pairs[lastiii][0], pairs[iii][0]]);
            }
            lastiii = iii;
          } else if(iii == pairs.length - 2 && dist2D(pairs[lastiii][0], pairs[iii][0]) >= minFreeSegDist){
            res.push([pairs[lastiii][0], pairs[iii][0]]);
          }
        }
      } else {
        res.push(apprPartTINCBW);
      }
    }
    return res;
  };

    IfcSpace.prototype.getArtificialEdgeNeighbors = function (knownNodes, apprPartsTANCBWCache = null, isClientRun = false, stairShapesFloor_ = undefined) {
      let neighbors = [], idsOfSpacesToExclude = [];
      //check if ARTIFICIALEDGENEIGHBOR where already calculated
      Object.keys(knownNodes).forEach(key => {
        for(const node of knownNodes[key]){
          if(node.type && node.type == 'ARTIFICIALEDGENODE'){
            let otherSpaceId = node.spaceIds.filter(id => {
              if(id != this.id){
                return true
              } else {
                return false
              }
            });
            if(otherSpaceId.length == 1){
              idsOfSpacesToExclude.push(...otherSpaceId);
              neighbors.push({type: 'ARTIFICIALEDGENEIGHBOR', id: key, nodes: [node]})
            }
          }
        }
      });
      //is this space a space in a space or has this space spaces in itself
      const thisNeighSegs = this.spaceNeighborsSegments(apprPartsTANCBWCache);
      let freeSegs = [];
      for(let space of this.__floor.__spaces){
        if(space !== this && space.getSpacesInSpace().filter(sIS => { //case: this space is space in other space
          if(this === sIS){
            return true;
          }
          return false;
        }).length){
          for(let seg of this.getFreeSegments(apprPartsTANCBWCache, isClientRun, thisNeighSegs)){
            freeSegs.push({seg: seg, thisSpace: this, otherSpace: space});
          }
          break;
        } else if(space === this){  //case: other spaces are spaces in this space
          for(let sIS of this.getSpacesInSpace()){
            for(let seg of sIS.getFreeSegments(apprPartsTANCBWCache, isClientRun)){
              freeSegs.push({seg: seg, thisSpace: this, otherSpace: sIS});
            }
          }
        }
      }
      //determine stair shapes to avoid
      let stairShapesFloor = stairShapesFloor_ !== undefined ? stairShapesFloor_ : this.__floor.__building.getStairShapesFloor(this.__floor);
      //determine transport element shapes to avoid
      let TEShapes = []
      for(let TE of this.__floor.getAllRefAndContTransportElems()){
        TE.transGeoToFloor(this.__floor)
        let ps = TE.shapePoints()
        TE.transGeoToFloor(TE.__floor, this.__floor)
        if(TE.isPureRectangleRepresentation()){
          ps.push(ps[0])
        }
        TEShapes.push(ps)
      }
      //define function that checks shape conflicts
      const conflictsShape = (p, shapeV, minDistToSeg = null) => {
        if(!shapeV.length){
          return false;
        }
        for(const shape of shapeV){
          if(isIn2DPolygon(shape, p)){
            return true;
          }
          for(let iii=0;iii<shape.length - 1;++iii){
            if(minDistToSeg !== null && distToSegment2D(p, shape[iii], shape[iii + 1]) < minDistToSeg){
              return true
            }
          }
        }
        return false;
      }
      //generate neighbors
      const thisSpaceIsParkingBox = this.isParkingBox();
      for(const segWSpaces of [...thisNeighSegs, ...freeSegs]){
        const makeNodes = (numNodes = null, run = 0) => {
          let exclude = false;
          for(const id of idsOfSpacesToExclude){
            if(segWSpaces.otherSpace.id == id){
              exclude = true;
              break
            }
          }
          if(exclude){
            return
          }
          let p1 = segWSpaces.seg[0], p2 = segWSpaces.seg[1];
          const segLen = Math.sqrt(Math.pow(p2.x - p1.x, 2) + Math.pow(p2.y - p1.y, 2));
          if(numNodes === null){
            if(segLen < 8000){
              numNodes = 1
            } else if(segLen < 17000){
              numNodes = 2
            } else if(segLen < 27000){
              numNodes = 3
            } else if(segLen < 38000){
              numNodes = 4
            } else if(segLen < 50000){
              numNodes = 5
            } else {
              numNodes = 6
            }
          }
          const dirX = p2.x - p1.x;
          const dirY = p2.y - p1.y;
          let _neighbors = [];
          const otherSpaceIsParkingBox = segWSpaces.otherSpace.isParkingBox();
          for(let iii=0;iii<numNodes;++iii){
            let potP = {x: p1.x + dirX * (iii + 1) / (numNodes + 1), y: p1.y + dirY * (iii + 1) / (numNodes + 1), z: this.__floor.elevation * 1000}
            if(!(thisSpaceIsParkingBox && otherSpaceIsParkingBox) && !conflictsShape(potP, stairShapesFloor, 250) && !conflictsShape(potP, TEShapes, 250)){
              let newNeigh = {type: 'ARTIFICIALEDGENEIGHBOR', id: getUuid(), nodes: [{id: getUuid(), x: potP.x, y: potP.y, z: potP.z, type: 'ARTIFICIALEDGENODE', spaceIds: [segWSpaces.thisSpace.id, segWSpaces.otherSpace.id]}]};
              _neighbors.push(newNeigh);
            }
          }
          if(!_neighbors.length && run++ < 5){
            makeNodes(numNodes * 2, run)
          } else {
            neighbors.push(..._neighbors)
          }
        }
        makeNodes()
      }
      return neighbors
    };

    IfcSpace.prototype.spaceNeighborsSegments = function (apprPartsTANCBWCache = null) {
        let res = [];
        const thisFreeSegments = this.getApprPartsTANCBWCache(apprPartsTANCBWCache), allOtherFreeSegmentsPsArrsAndSpace = [];
        for(const space of this.__floor.__spaces){
          if(space !== this){
            allOtherFreeSegmentsPsArrsAndSpace.push({segs: space.getApprPartsTANCBWCache(apprPartsTANCBWCache), space: space})
          }
        }
        let thisFreeSegmentsPsArrs = this.genFreeSegmentsPsArrs(thisFreeSegments);
        for(let elem of allOtherFreeSegmentsPsArrsAndSpace){
          Object.assign(elem, {psArrs: this.genFreeSegmentsPsArrs(elem.segs)})
        }
        const minFreeSegDist = 750;
        const maxSegEndsDists = 250;
        for(const thisFreeSegmentsPsArr of thisFreeSegmentsPsArrs){
          for(const oneOtherFreeSegmentsPsArrsAndSpace of allOtherFreeSegmentsPsArrsAndSpace){
            for(const oneOtherFreeSegmentsPsArr of oneOtherFreeSegmentsPsArrsAndSpace.psArrs){
            const p1Start_ = thisFreeSegmentsPsArr[0];
            const p1End_ = thisFreeSegmentsPsArr[thisFreeSegmentsPsArr.length - 1];
            const p2Start_ = oneOtherFreeSegmentsPsArr[0];
            const p2End_ = oneOtherFreeSegmentsPsArr[oneOtherFreeSegmentsPsArr.length - 1];
            const p1Start = new THREE.Vector3(p1Start_.x, p1Start_.y, 0);
            const p1End = new THREE.Vector3(p1End_.x, p1End_.y, 0);
            const p2Start = new THREE.Vector3(p2Start_.x, p2Start_.y, 0);
            const p2End = new THREE.Vector3(p2End_.x, p2End_.y, 0);
            if(distanceBetweenFeatureLines(p1Start, p1End, p2Start, p2End) >= maxSegEndsDists){
              continue;
            }
            let pPairs = this.getFreeSegsPsArrPairs(thisFreeSegmentsPsArr, oneOtherFreeSegmentsPsArr);
            let leftStart = 0, rightStart = pPairs.length - 1;
            for(let iii=0;iii<pPairs.length - 1;++iii){
              if(pPairs[iii][1] != pPairs[iii + 1][1]){
                leftStart = iii;
                break
              }
            }
            for(let iii=pPairs.length - 1;iii>=1;--iii){
              if(pPairs[iii][1] != pPairs[iii - 1][1]){
                rightStart = iii;
                break
              }
            }                              
            if(dist2D(pPairs[rightStart][0], pPairs[rightStart][1]) < maxSegEndsDists && 
            dist2D(pPairs[leftStart][0], pPairs[leftStart][1]) < maxSegEndsDists && 
              dist2D(pPairs[leftStart][1], pPairs[rightStart][1]) >= minFreeSegDist){
                res.push({seg: [pPairs[leftStart][0], pPairs[rightStart][0]], thisSpace: this,
                  otherSpace: oneOtherFreeSegmentsPsArrsAndSpace.space})
            }
          }
        }
      }
      return res
    };

    IfcSpace.prototype.approxPartsThatAreNotCoveredByWalls = function () {
      const eqRectAndPoly = (ps) => {
        if(this.isPureRectangleRepresentation()){
          ps.push(ps[0])
        }
        return ps
      };
      let res = [], thisPs = eqRectAndPoly(this.shapePoints()), wShapes = [];
      for (const wall of this.__floor.__walls) {
        wShapes.push(wall.shapePoints());
      }
      const minFreeSegDist = 750, minDistToWall = 100, everyXmm = 75;
      for(let iii=0;iii<thisPs.length - 1;++iii){
        const dirX = thisPs[iii + 1].x - thisPs[iii].x;
        const dirY = thisPs[iii + 1].y - thisPs[iii].y;
        const numSegs = Math.floor(Math.sqrt(Math.pow(dirX, 2) + Math.pow(dirY, 2)) / everyXmm);
        let segment = [];
        for(let jjj=0;jjj<=numSegs;++jjj){
          const p = {x: thisPs[iii].x + dirX * jjj / numSegs, y: thisPs[iii].y + dirY * jjj / numSegs};
          let isInPolygon = false;
          for(const wShape of wShapes){
            let distToSegment = false;
            for(let kkk=0;kkk<wShape.length - 1;++kkk){
              if(distToSegment2D(p, wShape[kkk], wShape[kkk + 1]) < minDistToWall){
                distToSegment = true;
                break
              }
            }
            if(distToSegment || isIn2DPolygon(wShape, p)){
              isInPolygon = true;
              break
            }
          }
          if(!isInPolygon && !segment.length){
            segment.push(p)
          } else if(isInPolygon && segment.length){
            let pBefore = {x: thisPs[iii].x + dirX * (jjj - 1) / numSegs, y: thisPs[iii].y + dirY * (jjj - 1) / numSegs};
            if(Math.sqrt(Math.pow(segment[0].x - pBefore.x, 2) + Math.pow(segment[0].y - pBefore.y, 2)) >= minFreeSegDist){
              segment.push(pBefore);
              res.push(segment)
            }
            segment = []
          } else if(jjj == numSegs && !isInPolygon){
            if(Math.sqrt(Math.pow(segment[0].x - p.x, 2) + Math.pow(segment[0].y - p.y, 2)) >= minFreeSegDist){
              segment.push(p);
              res.push(segment)
            }
            segment = []
          }
        }
      }
      return res
    };

    IfcSpace.prototype.getApprPartsTANCBWCache = function (apprPartsTANCBWCache = null) {
      if(apprPartsTANCBWCache){
        for(let entry of apprPartsTANCBWCache){
          if(entry.space.id === this.id){
            return entry.apprPartsTANCBW;
          }
        }
      }
      const parts = this.approxPartsThatAreNotCoveredByWalls();
      if(apprPartsTANCBWCache){
        apprPartsTANCBWCache.push({space: this, apprPartsTANCBW: parts});
      }
      return parts;
    };

    IfcSpace.prototype.getClosestPositionOnRelWall = function (position) {
      let ress = [], shortestDist = Number.MAX_SAFE_INTEGER, ressIII = 0, run = 0;
      for(const wall of this.getRelWalls()){
        ress.push(Object.assign(getClosestPointOnLines(position, wall.shapePoints()), {wall: wall}));
        if(ress[ress.length - 1].distance < shortestDist){
          shortestDist = ress[ress.length - 1].distance;
          ressIII = run
        }
        ++run
      }
      if(ress.length > ressIII){
        return ress[ressIII]
      } else {
        return null
      }
  };

    IfcSpace.prototype.create = function (project, name) {
      this.type = 'IFCSPACE';
      this.id = getUuid();
      this.name = name;
      this.__propertySets = [];

      project.register(this);

      return this;
    };

    IfcSpace.prototype.remove = function () {
      this.removeRelWalls();

      const thisCustomNeigh = this.getCustomEdgeNeighbors();
      let floorCustomNeigh = this.__floor.getCustomNeighbors();
      for(let iii = floorCustomNeigh.length - 1; iii>= 0; --iii){
        for(let thisNeigh of thisCustomNeigh){
          if(floorCustomNeigh[iii].id == thisNeigh.id){
            floorCustomNeigh.splice(iii, 1);
            break;
          }
        }
      }
      this.__floor.setCustomNeighbors(floorCustomNeigh);

      if(this.__doors && this.__doors.length){
        for(let door of this.__doors){
          if(door.__spaces && door.__spaces.length){
            for(let iii=door.__spaces.length-1;iii>=0;--iii){
              if(door.__spaces[iii]===this){
                door.__spaces.splice(iii,1)
                break
              }
            }
          }
        }
      }

      this.__project.getRelAggregate(this.__floor).removeRelatedObject(this);
      this.__project.getRelContainedInSpatialStructure(this).remove();
      this.__project.getRelReferencedInSpatialStructure(this).remove();

      // TODO Put in Relationship Class

      for (const __propertySet of this.__propertySets) {
        _.pull(this.__project.__relDefinesByProperties[__propertySet.oid].relatedObjects, this.oid);
        _.pull(this.__project.__relDefinesByProperties[__propertySet.oid].__relatedObjects, this);

        if (this.__project.__relDefinesByProperties[__propertySet.oid].relatedObjects.length === 0) {
          this.__project.__relDefinesByProperties[__propertySet.oid].remove();
        }
      }

      if (this.__objectPlacement) {
        this.__objectPlacement.remove(this.__project);
      }

      if (this.__representation) {
        this.__representation.remove(this.__project);
      }

      this.__project.unregister(this);

      for(const key of Object.keys(this.__project.__ifcElementAttributes)){
        if(key == this.id){
          delete this.__project.__ifcElementAttributes[key];
          break
        }
      }
    };

    IfcSpace.prototype.removeRelWalls = function (){
      for(let wallId of this.getRelatedWallsIds()){
        for(let wall of this.__floor.__walls){
          if(wall.id == wallId){
            wall.remove();
            break
          }
        }
      }
      this.setRelWallsIds([])
    }

    IfcSpace.prototype.getRelWalls = function (){
      let res = [];
      for(let wallId of this.getRelatedWallsIds()){
        for(let wall of this.__floor.__walls){
          if(wall.id == wallId){
            res.push(wall);
            break
          }
        }
      }
      return res
    }

    IfcSpace.prototype.getRelatedWallsIds = function (){
      let relWalls = this.getIfcElementAttributes().relWalls;
      if(!relWalls){
        relWalls = []
      }
      return relWalls
    }

    IfcSpace.prototype.setRelWallsIds = function (relWalls){
      this.getIfcElementAttributes().relWalls = relWalls
    }

    IfcSpace.prototype.initialize = function () {
        return this;
    };

    IfcSpace.prototype.hasRepresentation = function () {
        return this.representation != null;
    };

    IfcSpace.prototype.load = function (errors) {
        this.id = this.__attributes[0];
        this.name = this.__attributes[2];
        this.objectPlacement = this.__attributes[5];
        this.representation = this.__attributes[6];
    };

    IfcSpace.prototype.bind = function (project) {
      this.__project = project;

      if (this.objectPlacement) {
        this.__objectPlacement = this.__project.map[this.objectPlacement];
      }

      if (this.representation) {
        this.__representation = this.__project.map[this.representation];
      }

      this.__doors = [];
      this.__windows = [];
      // this.__walls = [];
      this.__propertySets = [];
      this.__category = this.getIfcElementAttributes().category;
      this.__capacity = this.getIfcElementAttributes().capacity
    };

    IfcSpace.prototype.getFullId = function () {
        return this.__floor.getFullId() + this.id + '/';
    };

    IfcSpace.prototype.getProject = function () {
        return this.__floor.__building.__site.__project;
    };

    IfcSpace.prototype.getPlacement = function () {
        return this.__objectPlacement.getPlacement();
    };

    IfcSpace.prototype.get2DBoundingBox = function () {
        return this.__representation.get2DBoundingBox();
    };

    IfcSpace.prototype.getGeo = function () {
        if (!this.__representation) {
            return new GeoNullProfile();
        }

        var geo = this.__representation.getGeo();

        if (this.__objectPlacement) {
            geo.applyPlacement(this.__objectPlacement.getPlacement());
        }

        return geo;
    };

    IfcSpace.prototype.getIfcElementAttributes = function() {
        return this.__project.getIfcElementAttributes(this);
    }

    IfcSpace.prototype.getIfcElementAttribute = function(name) {
        return this.__project.getIfcElementAttribute(this, name);
    }

    IfcSpace.prototype.setRectangleRepresentation = function (x, y, width, height) {
        if (!this.__representation) {
            this.__representation = new IfcProductDefinitionShape().create(this.getProject());
            this.representation = this.__representation.oid;
        }

        this.__representation.setRectangleRepresentation(this.getProject(), x + 0.5 * width, y + 0.5 * height, width, height);
    };

    IfcSpace.prototype.setPolylineRepresentation = function (points) {
        if (!this.__representation) {
            this.__representation = new IfcProductDefinitionShape().create(this.getProject());
            this.representation = this.__representation.oid;
        }

        this.__representation.setPolylineRepresentation(this.getProject(), points);
    };

    IfcSpace.prototype.dump = function (stream) {
        stream.write('#' + this.oid);
        stream.write('= ');
        stream.write('IFCSPACE(');
        dumpStringAttribute(stream, this.id, true);
        dumpOidAttribute(stream, null);
        dumpStringAttribute(stream, this.name);
        stream.write(');\n');
    };

    IfcSpace.prototype.renderFloorPlanSvg = function (svg, options) {
      if (!this.__representation) {
        return;
      }
  
      const group = this.getGeo().renderSvg(svg).attr('id', this.id).addClass('ifcSpace').attr('id', this.id).attr('stroke', 'green')
        .attr('stroke-width', 50).attr('fill', 'none');
  
      const boundingBox = this.getGeo().getBoundingBox();
  
      const thisRef = this;
      this.__spaceLabel = group.text(this.name).font({
        size: 400,
        anchor: 'middle'
      }).attr({
        stroke: 'none',
        fill: 'green',
        class: 'space-label counter-rotate',
        x: boundingBox.x + 0.5 * boundingBox.width,
        y: boundingBox.y + 0.5 * boundingBox.height - 400
      }).on('click', function(){
        event.preventDefault();
        if(!(options && options.editor && options.editor && (options.editor.buttonActionStarted || options.editor.space == thisRef))){
          event.stopPropagation();
          floorplanElemSelectChange.next({obj: thisRef, floor: null});
        }
      }).mouseover(function (event) {
        event.preventDefault();
        if(!(options && options.editor && options.editor && options.editor.buttonActionStarted)){
          this.style('cursor', 'hand');
        } else {
          this.style('cursor', 'crosshair');
        }
      }).mouseout(function (event) {
        event.preventDefault();
        this.style('cursor', 'default');
      });
  
      if(!(options && options.editor && options.editor.space && options.editor.space.id == this.id)){
        for(let nGA of this.getNoGoAreas()){
          group.polyline(nGA.points).attr('stroke', 'green').attr('stroke-width', 75).attr({ fill: 'grey', opacity: 0.4 });
        }
      }
    };

    IfcSpace.prototype.addToScene = function (scene) {
    };

    // TODO Added for Dijkstra Caching

    IfcSpace.prototype.isPurePolylineRepresentation = function () {
        return this.isPureXRepresentation('IFCARBITRARYCLOSEDPROFILEDEF');
    }

    IfcSpace.prototype.isPureRectangleRepresentation = function () {
        return this.isPureXRepresentation('IFCRECTANGLEPROFILEDEF');
    }

    IfcSpace.prototype.isPureXRepresentation = function (repType) {
        if (this.hasRepresentation()) {
            if (this.__representation.__representations) {
                for (const rep of this.__representation.__representations) {
                    if (rep.__items) {
                        for (const item of rep.__items) {
                            if (item.__sweptArea && item.__sweptArea.type !== repType) {
                                return false;
                            }
                        }
                    }
                }
            }
            return true;
        }
        return false;
    }

    IfcSpace.prototype.isInSpaceShape = function (point) {
        let outerSpacePoints = this.shapePoints();
        if (isIn2DPolygon(outerSpacePoints, point)) {
            return true;
        } else {
            return false;
        }
    }

    IfcSpace.prototype.shapePoints = function () {
      let g = this.getGeo();
      let shapePoints = [];
      if(this.isPurePolylineRepresentation()){
        shapePoints = g.points;
      } else if(this.isPureRectangleRepresentation()){
        shapePoints.push({x: g.x, y: g.y});
        shapePoints.push({x: g.x + g.width, y: g.y});
        shapePoints.push({x: g.x + g.width, y: g.y + g.height});
        shapePoints.push({x: g.x, y: g.y + g.height});
      } else {
        //console.log('unknown space representation for', this.name, 'on', this.__floor.name);
        return []
      }
      if(this.__floor.elevation !== undefined && this.__floor.elevation !== null){
        for(let point of shapePoints){
          Object.assign(point, {z: this.__floor.elevation * 1000})
        }
      }
      return shapePoints;
    }

    IfcSpace.prototype.area = function () {
      let ps = this.shapePoints()
      if(this.isPureRectangleRepresentation()){
        ps.push(ps[0])
      }
      return calcPolygonArea(ps) / 1000000
    }

    IfcSpace.prototype.getShapeCenter = function () {
      let center;
      if(this.isPureRectangleRepresentation()){
        center = centerOfMass(this.shapePoints())
      } else if(this.isPurePolylineRepresentation()){
        // //TODO
        let points = this.shapePoints();
        center = centerOfMass(points.slice(0, points.length - 1)) //only for testing purposes
        // center = visualCenter(this.shapePoints());
      } else {
        //console.log('representation type unknwon')
      }
      return center
    }

    IfcSpace.prototype.giveArea = function (points) {
        let total = 0;
        for (let i = 0, l = points.length; i < l; i++) {
            let addX = points[i].x;
            let addY = points[i == points.length - 1 ? 0 : i + 1].y;
            let subX = points[i == points.length - 1 ? 0 : i + 1].x;
            let subY = points[i].y;
            total += (addX * addY * 0.5);
            total -= (subX * subY * 0.5);
        }
        return Math.abs(total);
    }

    IfcSpace.prototype.getSpacesInSpace = function () {
      let outerSpacePoints = this.shapePoints();
      if(this.isPureRectangleRepresentation()){
        outerSpacePoints.push(outerSpacePoints[0])
      }
      return this.__floor.__spaces.filter(space => {
        if(space !== this){
          let allIn = true;
          for(const p of space.shapePoints()){
            let dist = Number.MAX_SAFE_INTEGER
            for(let iii=0;iii<outerSpacePoints.length - 1;++iii){
              let newDist = distToSegment2D(p, outerSpacePoints[iii], outerSpacePoints[iii + 1])
              if(newDist < dist){
                dist = newDist
              }
            }
            if(!isIn2DPolygon(outerSpacePoints, p) && dist > 250){
              allIn = false;
              break
            }
          }
          const spaceCenter = space.getShapeCenter();
          if(allIn && spaceCenter && isIn2DPolygon(outerSpacePoints, spaceCenter)){
            return true
          }
        }
        return false
      });
  }

      IfcSpace.prototype.getPointDensXY = function (area, spaceAssets = null, isClientRun = false, assetDefSpacer = 0) {
        if(spaceAssets){
          let assetWMin = Number.MAX_SAFE_INTEGER, assetHMin = Number.MAX_SAFE_INTEGER;
          const shapePs = this.shapePoints();
          let containsAsset = false;
          for(let asset of spaceAssets){
            if(asset.floor === this.__floor.id && asset.zPosition <= 2000){ //also see: conflictsAsset
              let w = assetDefSpacer, h = assetDefSpacer;
    
              //spatial expansion of asset
              if(asset && asset.model3d && asset.model3d.boundingBox){
                w += asset.model3d.boundingBox.width;
                h += asset.model3d.boundingBox.height;
              } else  { //if asset spatial expansion smaller or equal to 500 x 500: don't take it into account
                continue;
              }
    
              containsAsset = true;
              if(w < assetWMin){
                assetWMin = w;
              }
              if(h < assetHMin){
                assetHMin = h;
              }
            }
          }
          if(containsAsset){
            let minWH = Math.min(assetWMin, assetHMin) / 2;
            let minDens = 250;
            let maxArea = 200;
            const notClientRunFactor = 2;
            if(!isClientRun){
              //minDens /= notClientRunFactor;
              maxArea *= notClientRunFactor;
            }
            if(minWH >= minDens && area <= maxArea){
              return Math.round(minWH);
            } else {
              console.log('ignore asset influence on density for space', this.name);
            }
          }
        }
        const minDens = 300;
        const minArea = 20;
        //return Math.round(minDens + (area <= minArea ? 0 : area - minArea) * 2);
        // let resDens = Math.round(minDens + 20 * Math.sqrt(area <= minArea ? 0 : area - minArea));
        let resDens = Math.round(minDens + 12 * Math.pow(area <= minArea ? 0 : area - minArea, 0.7));
        if(isClientRun){
          resDens *= 2;
        }
        return resDens;
    }
    

      IfcSpace.prototype.getDijGridAdaptiveDens = function (spaceAssets = null, isClientRun = false, recRun = false, givenPointDens = null) {
        let geo = this.getGeo();
        let b = geo.getBoundingBox();
        const area = this.area();
        const defSpacer = 250;
        const assetSpacer = recRun ? 100 : defSpacer;
        //set density; the higher dens is, the coarser the grid; density has to take minimum asset dimension into account; this can lead to very low pointDensXY!
        let pointDens = givenPointDens === null ? this.getPointDensXY(area, spaceAssets, isClientRun, assetSpacer) : givenPointDens;
        //generate regular grid
        const maxDim = Math.max(b.width, b.height);
        const numPoints = Math.round(maxDim / pointDens);
        let grid2D = [], zPos;
        if(this.__floor.elevation !== undefined && this.__floor.elevation !== null){
          zPos = this.__floor.elevation * 1000;
        }
        let spaceShapePs = this.shapePoints();
        if(this.isPureRectangleRepresentation()){
          spaceShapePs.push(spaceShapePs[0])
        }
    
        // let vOfVSpaceShapePs = [];
        // for(let spaceShapeP of spaceShapePs){
        //   vOfVSpaceShapePs.push([spaceShapeP.x, spaceShapeP.y]);
        // }
        // //to avoid multiple padding area results the last p of a polyline space should not be very close but equal to the first p
        // if(this.isPurePolylineRepresentation()){
        //   vOfVSpaceShapePs.splice(vOfVSpaceShapePs.length - 1, 1);
        //   vOfVSpaceShapePs.push(vOfVSpaceShapePs[0]);
        // }
        // //determine padding polygon which covers about x% of the actual polygon
        // const paddingAreaFactor = 1.5;
        // let paddingArea = area, padding = area * paddingAreaFactor, paddingShapePs = [];
        //does not work properly for complex spaces
        // while(paddingArea / area > 0.8){
        //   const spaceShapeOffset = new offset();
        //   const spaceShapePaddingPs = spaceShapeOffset.data(vOfVSpaceShapePs).padding(padding);
        //   let shapePaddingPs = [];
        //   if(spaceShapePaddingPs.length == 1){
        //     for(let iii=0;iii<spaceShapePaddingPs[0].length;++iii){
        //       shapePaddingPs.push({x: spaceShapePaddingPs[0][iii][0], y: spaceShapePaddingPs[0][iii][1]});
        //     }
        //     paddingShapePs = JSON.parse(JSON.stringify(shapePaddingPs));
        //     paddingArea = calcPolygonArea(shapePaddingPs.splice(0, shapePaddingPs.length - 1)) / 1000000;
        //   } else {
        //     console.log('padding area for', this.name, 'does not consist of a single area')
        //   }
        //   padding += area * paddingAreaFactor;
        // }
    
        // const getMinDistToWall = (p) => {
        //   let res = Number.MAX_SAFE_INTEGER;
        //   for(let iii=0;iii<spaceShapePs.length - 1;++iii){
        //     const dist = distToSegment2D(p, spaceShapePs[iii], spaceShapePs[iii + 1]);
        //     if(dist < res){
        //       res = dist;
        //     }
        //   }
        //   return res;
        // }
        // const minDistToWall = 2000;
    
        const minNumGridPs = 10;
        let gridPCounter = 0;
        for(let iii=0;iii<numPoints;++iii){
          let x = b.x + maxDim / numPoints * iii;
          let grid1D = [];
          for(let jjj=0;jjj<numPoints;++jjj){
            let y = b.y + maxDim / numPoints * jjj;
            let p = {x: x, y: y, z: zPos};
    
            // if(paddingShapePs.length && (iii % 2 == 0 || jjj % 2 == 0) && isIn2DPolygon(paddingShapePs, p)){
            //   Object.assign(p, {conflictsSomething: true})
            // }
    
            // if((iii % 2 == 0 || jjj % 2 == 0) && isIn2DPolygon(spaceShapePs, p) && getMinDistToWall(p) > minDistToWall){
            //   Object.assign(p, {conflictsSomething: true})
            // }

            if(!isIn2DPolygon(spaceShapePs, p, defSpacer)){
              Object.assign(p, {conflictsSomething: true})
            } else {
              gridPCounter++;
            } 
    
            grid1D.push(p);
          }
          grid2D.push(grid1D);
        }
        //very small rooms need a minimum amount of points
        if(!recRun && gridPCounter < minNumGridPs){
          return this.getDijGridAdaptiveDens(spaceAssets, isClientRun, true, pointDens / 1.5);
        }
        //determine shapes where navigation can't occur
        //get all relevant elements
        const floorBelow = this.__floor.getFloorBelow();
        //stairs
        const stairsOfFloorBelow = floorBelow ? floorBelow.__stairs : [];
        const stairs = [...this.__floor.__stairs, ...stairsOfFloorBelow];
        //ramps
        const rampsOfFloorBelow = floorBelow ? floorBelow.__ramps : [];
        const ramps = [...this.__floor.__ramps, ...rampsOfFloorBelow];
        //spaces in space
        const spacesInSpace = this.getSpacesInSpace().filter(s => !s.isExcludedFromNav());
        //transport elements
        const transportElements = this.__floor.getAllRefAndContTransportElems()
        //no go areas
        const noGoAreas = this.getNoGoAreas();
        //walls
        const walls = this.getRelWalls();
        //get shapes
        let stairsShapes = [], TEsShapes = [], sISShapes = [], wallsShapes = [], rampsShapes = [], noGoAreaShapes = [];
        //get spaces in space shapes
        for(let elem of spacesInSpace){
          sISShapes.push({ps: elem.shapePoints(), isPoly: elem.isPurePolylineRepresentation(), geo: elem.getGeo()})
        }
        //get stair, ramp, transport element and no go area shapes
        const genShapes = (objs, shapes, isPoly = undefined, transGeoToOtherFloor = true) => {
          for(let obj of objs){
            if(transGeoToOtherFloor){
              obj.transGeoToFloor(this.__floor);
            }
            shapes.push({ps: obj.shapePoints(), isPoly: isPoly === undefined ? obj.isPurePolylineRepresentation() : isPoly, geo: isPoly === undefined ? obj.getGeo() : null});
            if(transGeoToOtherFloor){
              obj.transGeoToFloor(obj.__floor, this.__floor);
            }
          }
        };
        genShapes(stairs, stairsShapes);
        genShapes(transportElements, TEsShapes);
        genShapes(ramps, rampsShapes, true);
        genShapes(walls, wallsShapes, true, false);
        for(let noGoArea of noGoAreas){
          noGoAreaShapes.push({ps: noGoArea.points, isPoly: true, geo: null});
        }
        //check if point conflicts something
        const conflictsShapes = (p, shapes, minDistToShape = null) => {
          if(!shapes.length){
            return false;
          }
          for(const shape of shapes){
            if(isIn2DPolygon(shape.ps, p)){
              return true;
            }
            if(minDistToShape !== null){
              if(shape.isPoly){
                if(getClosestPointOnLines(p, shape.ps).distance < minDistToShape){
                  return true
                }
              } else {
                const boundingBox = shape.geo.getBoundingBox();
                if(getClosestPointOnRectangle(p, boundingBox.x, boundingBox.y, boundingBox.width, boundingBox.height).distance < minDistToShape){
                  return true
                }
              }
            }
          }
          return false;
        };
        const conflictsSomething = (p) => {
          // if(!isIn2DPolygon(spaceShapePs, p, defSpacer)){
          //   return true;
          // }
          if(this.conflictsAsset(spaceAssets, p, assetSpacer)){
            p.conflictsAsset = true;
            return true;
          } else {
            p.conflictsAsset = false;
          }
          if(conflictsShapes(p, stairsShapes, defSpacer)){
            return true;
          }
          if(conflictsShapes(p, TEsShapes, defSpacer)){
            return true;
          }
          if(conflictsShapes(p, sISShapes, defSpacer)){
            return true;
          }
          if(conflictsShapes(p, wallsShapes, defSpacer)){
            return true;
          }
          if(conflictsShapes(p, rampsShapes, defSpacer)){
            return true;
          }
          if(conflictsShapes(p, noGoAreaShapes, defSpacer)){
            return true;
          }
          return false;
        }
        const areaAroundAssetsToKeepPs = assetSpacer + 1500;
        //let numsAroundAssets = 0;
        for(let iii=0;iii<grid2D.length;++iii){
          for(let jjj=grid2D[iii].length - 1;jjj>=0;--jjj){
            let p = grid2D[iii][jjj];
            p.spaceId = this.id;
            // let loweredDensP = false;
            // if(p.conflictsSomething != undefined){
            //   loweredDensP = true;
            // }
            if(p.conflictsSomething || conflictsSomething(p)){
              p.conflictsSomething = true;
            } else {
              p.conflictsSomething = false;
            }
            //keep all available grid points around an asset
            // if(loweredDensP){ //p.conflictsSomething was true then
            //   if(!p.conflictsSomething && this.conflictsAsset(assets, p, areaAroundAssetsToKeepPs)){
            //     numsAroundAssets++;
            //     p.conflictsSomething = false;
            //   } else {
            //     p.conflictsSomething = true;
            //   }
            // }
          }
        }
        //gains about 0.7 for numsAroundAssets / this.getNumOfAssetsInSpace(assets); temporarily deactivated because: time demanding
        // if(!recRun && numsAroundAssets / this.getNumOfAssetsInSpace(assets) <= 10){
        //   return this.getDijGridAdaptiveDens(assets, isClientRun, true);
        // }
        return grid2D;
    }

    IfcSpace.prototype.getAssets = function (buildingAssets = null) {
      let shapePs = this.shapePoints(), res = [];
      if(this.isPureRectangleRepresentation()){
        shapePs.push(shapePs[0]);
      }
      if(buildingAssets){
        for(let asset of buildingAssets){
          if(asset.floor === this.__floor.id && isIn2DPolygon(shapePs, {x: asset.xPosition, y: asset.yPosition})){
            res.push(asset);
          }
        }
      }
      return res;
    }

    IfcSpace.prototype.getAssetsShapes = function (buildingAssets = null) {
      let res = [];
      for(let asset of this.getAssets(buildingAssets)){
        let w, h, x = asset.xPosition, y = asset.yPosition;
        //spatial expansion of asset
        if(asset.model3d && asset.model3d.boundingBox){
          w = asset.model3d.boundingBox.width;
          h = asset.model3d.boundingBox.height;
        } else {  //if asset has no spatial expansion: don't take it into account
          continue;
        }
        let assetBox = [
          {x: x - w / 2, y: y - h / 2},
          {x: x + w / 2, y: y - h / 2},
          {x: x + w / 2, y: y + h / 2},
          {x: x - w / 2, y: y + h / 2}];
        assetBox.push(assetBox[0]);
        res.push(assetBox);
      }
      return res;
    }

    IfcSpace.prototype.getDijGrid = function (spaceAssets = null, isClientRun = false, pointDens = null, recRun = 0, excludeWalls = [], excludeSpaceEdges = [], defSpacer = 250) {

      const maxRecRuns = 2  //maximum number of recursive runs; if necessary: in each call the grid parameter will be adjustet to get more grid points; more grid points will increase the dijkstra calculation time

      let geo = this.getGeo();
      let b = geo.getBoundingBox();
      const area = this.area();
      if(pointDens === null){
        pointDens = this.getPointDensXY(area, spaceAssets, isClientRun);
      }
      const points = Math.round(b.width / pointDens);
      let grid2D = [], zPos;
      if(this.__floor.elevation !== undefined && this.__floor.elevation !== null){
        zPos = this.__floor.elevation * 1000;
      }
      for(let iii=0;iii<points;++iii){
        let x = b.x + b.width / points * iii;
        let grid1D = [];
        for(let jjj=0;jjj<points;++jjj){
          let y = b.y + b.height / points * jjj;
          grid1D.push({x: x, y: y, z: zPos});
        }
        grid2D.push(grid1D);
      }
      let outerSpacePoints = this.shapePoints();
      if(this.isPureRectangleRepresentation()){
        outerSpacePoints.push(outerSpacePoints[0])
      }
      //let shortestSegDs = []; //un-/comment for weighted/unweighted beg
      const floorBelow = this.__floor.getFloorBelow()
      const stairsOfFloorBelow = floorBelow ? floorBelow.__stairs : []
      const rampsOfFloorBelow = floorBelow ? floorBelow.__ramps : []
      const stairs = [...this.__floor.__stairs, ...stairsOfFloorBelow],
        transportElements = this.__floor.getAllRefAndContTransportElems(),
        spacesInSpace = this.getSpacesInSpace(), 
        ramps = [...this.__floor.__ramps, ...rampsOfFloorBelow]
      let stairsShapePoints = [], TEsShapePoints = [], sISShapePoints = [], wallsShapePoints = [], rampsShapePoints = [];
      for(let elem of spacesInSpace){
        sISShapePoints.push({ps: elem.shapePoints(), isPoly: elem.isPurePolylineRepresentation(), geo: elem.getGeo()})
      }
      const genShapePoints = (objs, shapePointsV) => {
        for(let obj of objs){
          obj.transGeoToFloor(this.__floor);
            shapePointsV.push({ps: obj.shapePoints(), isPoly: obj.isPurePolylineRepresentation(), geo: obj.getGeo()});
          obj.transGeoToFloor(obj.__floor, this.__floor)
        }
      };
      genShapePoints(stairs, stairsShapePoints);
      genShapePoints(transportElements, TEsShapePoints);
      for(let ramp of ramps){
        ramp.transGeoToFloor(this.__floor);
        rampsShapePoints.push({ps: ramp.shapePoints(), isPoly: true, geo: null});
        ramp.transGeoToFloor(ramp.__floor, this.__floor)
      }
      const customWallSpacer = defSpacer / (recRun + 1);
      const customSpaceEdgeSpacer = defSpacer / (recRun + 1);
      const relWalls = this.getRelWalls()
      for(let wall of relWalls){
        let shapeObj = {ps: wall.shapePoints(), isPoly: true, geo: null}
        for(let exWall of excludeWalls){
          if(wall === exWall){
            Object.assign(shapeObj, {customSpacer: customWallSpacer})
          }
        }
        wallsShapePoints.push(shapeObj)
      }
      const conflictsShapes = (p, shapePointsV, minDistToSegment_ = null) => {
        if(!shapePointsV.length){
          return false;
        }
        for(const shapePoints of shapePointsV){
          let minDistToSegment = minDistToSegment_
          if(shapePoints.customSpacer){
            minDistToSegment = shapePoints.customSpacer
          }
          if(isIn2DPolygon(shapePoints.ps, p)){
            return true;
          }
          if(shapePoints.isPoly){
            if(minDistToSegment !== null && getClosestPointOnLines(p, shapePoints.ps).distance < minDistToSegment){
              return true
            }
          } else {
            const boundingBox = shapePoints.geo.getBoundingBox();
            if(getClosestPointOnRectangle(p, boundingBox.x, boundingBox.y, boundingBox.width, boundingBox.height).distance < minDistToSegment){
              return true
            }
          }
        }
        return false;
      };
      //let numPsWithNoConflicts = 0
      let noGoAreasV = [];
      for(let noGoArea of this.getNoGoAreas()){
        noGoAreasV.push({ps: noGoArea.points, isPoly: true, geo: null});
      }
      const conflictsSomething = (p) => {
        if(!isIn2DPolygon(outerSpacePoints, p, defSpacer, excludeSpaceEdges, customSpaceEdgeSpacer)){
          return true;
        }
        if(this.conflictsAsset(spaceAssets, p, defSpacer)){
          return true;
        }
        if(conflictsShapes(p, stairsShapePoints, defSpacer)){
          return true;
        }
        if(conflictsShapes(p, TEsShapePoints, defSpacer)){
          return true;
        }
        if(conflictsShapes(p, sISShapePoints, defSpacer)){
          return true;
        }
        if(conflictsShapes(p, wallsShapePoints, defSpacer)){
          return true;
        }
        if(conflictsShapes(p, rampsShapePoints, defSpacer)){
          return true;
        }
        if(conflictsShapes(p, noGoAreasV, defSpacer)){
          return true;
        }
        return false;
      }
      for(let iii=0;iii<grid2D.length;++iii){
        for(let jjj=grid2D[iii].length - 1;jjj>=0;--jjj){
          let p = grid2D[iii][jjj];
          p.spaceId = this.id;
          // if(this.isCloseToEdge(outerSpacePoints, p)){
          //   p.isCloseToEdge = true;
          // } else {
          //   p.isCloseToEdge = false;
          // }
          if(this.conflictsAsset(spaceAssets, p, defSpacer)){
            p.conflictsAsset = true;
          } else {
            p.conflictsAsset = false;
          }
          if(conflictsSomething(p)){
            p.conflictsSomething = true;
          } else {
            p.conflictsSomething = false;
          }
          // if(isIn2DPolygon(outerSpacePoints, p, defSpacer, excludeSpaceEdges, customSpaceEdgeSpacer)){
          //   p.isIn2DPolygon = true;
          // } else {
          //   p.isIn2DPolygon = false;
          // }
          // if(conflictsShapes(p, stairsShapePoints, defSpacer)){
          //   p.conflictsStair = true;
          // } else {
          //   p.conflictsStair = false;
          // }
          // if(conflictsShapes(p, TEsShapePoints, defSpacer)){
          //   p.conflictsTransportElement = true;
          // } else {
          //   p.conflictsTransportElement = false;
          // }
          // if(conflictsShapes(p, sISShapePoints, defSpacer)){
          //   p.conflictsOtherSpace = true;
          // } else {
          //   p.conflictsOtherSpace = false;
          // }
          // if(conflictsShapes(p, wallsShapePoints, defSpacer)){
          //   p.conflictsWall = true;
          // } else {
          //   p.conflictsWall = false;
          // }
          // if(!IfcSpace.pConflictsSomething(p)){
          //   numPsWithNoConflicts++
          // }
          //un-/comment for weighted/unweighted beg
          // if(!IfcSpace.pConflictsSomething(p)){
          //   const shortestSegementDist = (edgePoints, point) => {
          //     let dist = Number.MAX_SAFE_INTEGER;
          //     const isRect = this.isPureRectangleRepresentation();
          //     for(let iii=0;iii<(isRect ? edgePoints.length : edgePoints.length - 1); ++iii){
          //       let newDist = distToSegment2D(point, edgePoints[iii], edgePoints[isRect ? ((iii + 1) % 4) : (iii + 1)]);
          //       if(newDist < dist){
          //         dist = newDist;
          //       }
          //     }
          //     return dist;
          //   };
          //   let dist = shortestSegementDist(outerSpacePoints, p);
          //   p.weight = dist;
          //   shortestSegDs.push(dist);
          // }
          //un-/comment for weighted/unweighted end
        }
      }
      //un-/comment for weighted/unweighted beg
      // let maxShortestSegD = Math.max(...shortestSegDs),
      //   minShortestSegD = Math.min(...shortestSegDs);
      // const weightMinBorder = 0.1;
      //   minShortestSegD /= maxShortestSegD;
      // const adaptWeight = (w) => {
      //   if(w>=1000){
      //     return 1;
      //   }
      //   w /= maxShortestSegD;
      //   return (1 - weightMinBorder) * (w - minShortestSegD) / (1 - minShortestSegD) + weightMinBorder;
      // };
      // for(let iii=0;iii<grid2D.length;++iii){
      //   for(let jjj=grid2D[iii].length - 1;jjj>=0;--jjj){
      //     if(!IfcSpace.pConflictsSomething(grid2D[iii][jjj])){
      //       grid2D[iii][jjj].weight = adaptWeight(grid2D[iii][jjj].weight);
      //     }
      //   }
      // }
      //un-/comment for weighted/unweighted end
      const checkForNarrowRegions = () => {
        const angleDeg = (p1, p2, p3) => {
          var p1p2 = Math.sqrt(Math.pow(p2.x-p1.x,2)+ Math.pow(p2.y-p1.y,2));
          var p2p3 = Math.sqrt(Math.pow(p2.x-p3.x,2)+ Math.pow(p2.y-p3.y,2));
          var p1p3 = Math.sqrt(Math.pow(p3.x-p1.x,2)+ Math.pow(p3.y-p1.y,2));
          return Math.acos((p2p3*p2p3+p1p2*p1p2-p1p3*p1p3)/(2*p2p3*p1p2))/Math.PI*180
        };
        const len = (p1, p2) => {
          return Math.sqrt(Math.pow(p2.y - p1.y, 2) + Math.pow(p2.x - p1.x, 2))
        }
        const intersectP = (p1, p2, p3, p4) => {
          let ua, ub, denom = (p4.y - p3.y)*(p2.x - p1.x) - (p4.x - p3.x)*(p2.y - p1.y);
          if (denom == 0) {
              return null;
          }
          ua = ((p4.x - p3.x)*(p1.y - p3.y) - (p4.y - p3.y)*(p1.x - p3.x))/denom;
          ub = ((p2.x - p1.x)*(p1.y - p3.y) - (p2.y - p1.y)*(p1.x - p3.x))/denom;
          return {
              x: p1.x + ua * (p2.x - p1.x),
              y: p1.y + ua * (p2.y - p1.y),
              seg1: ua >= 0 && ua <= 1,
              seg2: ub >= 0 && ub <= 1
          };
        }
        const oSPs = outerSpacePoints
        const oSPsArea = calcPolygonArea(oSPs) / 1000000
        const angleThres = 7.5, parallDistThres = 2000, prolong = 500, aimDensThresDiv = (recRun + 1) * 0.75, closeToWallThres = 100, closeToSpaceEdgeThres = 100, bottleneckDist = 1500, smallRoomAreaThres = 2, nicheThres = 1  //nicheThres < smallRoomAreaThres
        let wallsForSmallerSpacer = [], spaceEdgesForSmallerSpacer = [], rerun = false
        const isRedundantWall = (newEntry) => {
          for(let w of wallsForSmallerSpacer){
            if(newEntry === w){
              return true
            }
          }
          return false
        }
        const isRedundantSpaceEdge = (newEntry) => {
          for(let s of spaceEdgesForSmallerSpacer){
            if(newEntry[0].x == s[0].x && newEntry[1].x == s[1].x && newEntry[0].y == s[0].y && newEntry[1].y == s[1].y){
              return true
            }
          }
          return false
        }
        for(let iii=0;iii<oSPs.length - 1; ++iii){
          for(let jjj=iii + 1;jjj<oSPs.length - 1; ++jjj){
            //determine edge point-pairs which could be a bottleneck for navigation
            if(jjj > iii + 2 && jjj < oSPs.length - 2 && pToPDist(oSPs[iii], oSPs[jjj]) < bottleneckDist && isIn2DPolygon(oSPs, pPCenter2D(oSPs[iii], oSPs[jjj])) && calcPolygonArea(oSPs.slice(iii, jjj + 1)) > nicheThres){
              let edges = [[oSPs[iii == 0 ? oSPs.length - 1 : iii - 1], oSPs[iii]], [oSPs[iii], oSPs[iii + 1]], [oSPs[jjj - 1], oSPs[jjj]], [oSPs[jjj], oSPs[jjj + 1]]]
              for(let edge of edges){
                if(!isRedundantSpaceEdge(edge)){
                  spaceEdgesForSmallerSpacer.push(edge)
                }
                let middle = pPCenter2D(edge[0], edge[1])
                let closestWall = {wall: null , dist: Number.MAX_SAFE_INTEGER, shapePs: null}
                for(let wall of relWalls){
                  const wallShapePs = wall.shapePoints()
                  for(let kkk=0;kkk<wallShapePs.length - 1;++kkk){
                    let wallSegDist = distToSegment2D(middle, wallShapePs[kkk], wallShapePs[kkk + 1])
                    if(closestWall.dist > wallSegDist){
                      closestWall.wall = wall
                      closestWall.dist = wallSegDist
                      closestWall.shapePs = wallShapePs
                    }
                  }
                }
                if(closestWall.wall !== null && (closeToWallThres > closestWall.dist || isIn2DPolygon(closestWall.shapePs, middle)) && 
                  !isRedundantWall(closestWall.wall)){
                  wallsForSmallerSpacer.push(closestWall.wall)
                }
              }
            }
            //determine (nearly) square areas with low point density
            const iii1 = oSPs[iii]
            const iii2 = oSPs[iii + 1]
            const jjj1 = oSPs[jjj]
            const jjj2 = oSPs[jjj + 1]
            const viiiLen = len(iii2, iii1)
            const vjjjLen = len(jjj2, jjj1)
            const viiiN = {x: (iii2.x - iii1.x)/viiiLen, y: (iii2.y - iii1.y)/viiiLen}
            const vjjjN = {x: (jjj2.x - jjj1.x)/vjjjLen, y: (jjj2.y - jjj1.y)/vjjjLen}
            const angle = angleDeg(viiiN, {x: 0, y: 0}, vjjjN)
            if(angle > -angleThres && angle < angleThres || angle - 180 > -angleThres && angle - 180 < angleThres){
              const viiiNPerp = {x: -viiiN.y, y: viiiN.x}
              const vjjjNPerp = {x: -vjjjN.y, y: vjjjN.x}
              const iii1Long = {x: iii1.x - viiiN.x * prolong, y: iii1.y - viiiN.y * prolong}
              const iii2Long = {x: iii2.x + viiiN.x * prolong, y: iii2.y + viiiN.y * prolong}
              const jjj1Long = {x: jjj1.x - vjjjN.x * prolong, y: jjj1.y - vjjjN.y * prolong}
              const jjj2Long = {x: jjj2.x + vjjjN.x * prolong, y: jjj2.y + vjjjN.y * prolong}
              const viii1TojjjLongDir1 = intersectP(iii1, {x: iii1.x + viiiNPerp.x * parallDistThres, y: iii1.y + viiiNPerp.y * parallDistThres}, jjj1Long, jjj2Long)
              const viii1TojjjLongDir2 = intersectP(iii1, {x: iii1.x - viiiNPerp.x * parallDistThres, y: iii1.y - viiiNPerp.y * parallDistThres}, jjj1Long, jjj2Long)
              const viii2TojjjLongDir1 = intersectP(iii2, {x: iii2.x + viiiNPerp.x * parallDistThres, y: iii2.y + viiiNPerp.y * parallDistThres}, jjj1Long, jjj2Long)
              const viii2TojjjLongDir2 = intersectP(iii2, {x: iii2.x - viiiNPerp.x * parallDistThres, y: iii2.y - viiiNPerp.y * parallDistThres}, jjj1Long, jjj2Long)
              const vjjj1ToiiiLongDir1 = intersectP(jjj1, {x: jjj1.x + vjjjNPerp.x * parallDistThres, y: jjj1.y + vjjjNPerp.y * parallDistThres}, iii1Long, iii2Long)
              const vjjj1ToiiiLongDir2 = intersectP(jjj1, {x: jjj1.x - vjjjNPerp.x * parallDistThres, y: jjj1.y - vjjjNPerp.y * parallDistThres}, iii1Long, iii2Long)
              const vjjj2ToiiiLongDir1 = intersectP(jjj2, {x: jjj2.x + vjjjNPerp.x * parallDistThres, y: jjj2.y + vjjjNPerp.y * parallDistThres}, iii1Long, iii2Long)
              const vjjj2ToiiiLongDir2 = intersectP(jjj2, {x: jjj2.x - vjjjNPerp.x * parallDistThres, y: jjj2.y - vjjjNPerp.y * parallDistThres}, iii1Long, iii2Long)
              let p1, p2, p3, p4, p1Mirr, p2Mirr, p3Mirr, p4Mirr
              if(viii1TojjjLongDir1.seg1 && viii1TojjjLongDir1.seg2){
                p1 = iii1
                p1Mirr = viii1TojjjLongDir1
              } else if(viii1TojjjLongDir2.seg1 && viii1TojjjLongDir2.seg2){
                p1 = iii1
                p1Mirr = viii1TojjjLongDir2
              }
              if(viii2TojjjLongDir1.seg1 && viii2TojjjLongDir1.seg2){
                p2 = iii2
                p2Mirr = viii2TojjjLongDir1
              } else if(viii2TojjjLongDir2.seg1 && viii2TojjjLongDir2.seg2){
                p2 = iii2
                p2Mirr = viii2TojjjLongDir2
              }
              if(vjjj1ToiiiLongDir1.seg1 && vjjj1ToiiiLongDir1.seg2){
                p3 = jjj1
                p3Mirr = vjjj1ToiiiLongDir1
              } else if(vjjj1ToiiiLongDir2.seg1 && vjjj1ToiiiLongDir2.seg2){
                p3 = jjj1
                p3Mirr = vjjj1ToiiiLongDir2
              }
              if(vjjj2ToiiiLongDir1.seg1 && vjjj2ToiiiLongDir1.seg2){
                p4 = jjj2
                p4Mirr = vjjj2ToiiiLongDir1
              } else if(vjjj2ToiiiLongDir2.seg1 && vjjj2ToiiiLongDir2.seg2){
                p4 = jjj2
                p4Mirr = vjjj2ToiiiLongDir2
              }
              let p1_, p2_, p3_, p4_
              if(p1){
                p1_ = p1
              } else if(p4Mirr){
                p1_ = p4Mirr
              }
              if(p2){
                p2_ = p2
              } else if(p3Mirr){
                p2_ = p3Mirr
              }
              if(p3){
                p3_ = p3
              } else if(p2Mirr){
                p3_ = p2Mirr
              }
              if(p4){
                p4_ = p4
              } else if(p1Mirr){
                p4_ = p1Mirr
              }
              if(p1_ && p2_ && p3_ && p4_){
                let noConflicts = 0
                const pArr = [p1_, p2_, p3_, p4_]
                //check if determined area is inside the polygon
                if(!isIn2DPolygon(oSPs, centerOfMass(pArr))){
                  continue
                }
                pArr.push(pArr[0])
                //determine density of points that don't conflict with anything
                for(let kkk=0;kkk<grid2D.length;++kkk){
                  for(let lll=grid2D[kkk].length - 1;lll>=0;--lll){
                    if(grid2D[kkk][lll].conflictsSomething === false && isIn2DPolygon(pArr, grid2D[kkk][lll])){
                      noConflicts++
                    }
                  }
                }
                const pArrArea = calcPolygonArea(pArr) / 1000000
                const pPSM = noConflicts / pArrArea, aimDens = Math.pow(1000, 2) / pointDens / pointDens
                //if dennsity is too low determine walls which belong to the area
                if((pArrArea > nicheThres && oSPsArea > smallRoomAreaThres || oSPsArea <= smallRoomAreaThres) && pPSM < aimDens / aimDensThresDiv){ //pArrArea > 1 to avoid taking niches into account
                  rerun = true
                  //collect walls
                  for(let mmm = 0; mmm < pArr.length - 1; ++mmm){
                    let middle = pPCenter2D(pArr[mmm], pArr[mmm + 1])
                    let closestWall = {wall: null , dist: Number.MAX_SAFE_INTEGER, shapePs: null}
                    for(let wall of relWalls){
                      const wallShapePs = wall.shapePoints()
                      for(let nnn=0;nnn<wallShapePs.length - 1;++nnn){
                        let wallSegDist = distToSegment2D(middle, wallShapePs[nnn], wallShapePs[nnn + 1])
                        if(closestWall.dist > wallSegDist){
                          closestWall.wall = wall
                          closestWall.dist = wallSegDist
                          closestWall.shapePs = wallShapePs
                        }
                      }
                    }
                    if(closestWall.wall !== null && (closeToWallThres > closestWall.dist || isIn2DPolygon(closestWall.shapePs, middle)) && 
                      !isRedundantWall(closestWall.wall)){
                      wallsForSmallerSpacer.push(closestWall.wall)
                    }
                  }
                  //collect space edges
                  for(let mmm = 0; mmm<oSPs.length - 1; ++mmm){
                    for(let nnn = 0; nnn < pArr.length - 1; ++nnn){
                      const middle = pPCenter2D(pArr[nnn], pArr[nnn + 1]), edge = [oSPs[mmm], oSPs[mmm + 1]]
                      if(distToSegment2D(middle, oSPs[mmm], oSPs[mmm + 1]) < closeToSpaceEdgeThres && !isRedundantSpaceEdge(edge)){
                        spaceEdgesForSmallerSpacer.push(edge);
                        break;
                      }
                    }
                  }
                }
              }
            }
          }
        }
        return {walls: wallsForSmallerSpacer, rerun: rerun, spaceEdges: spaceEdgesForSmallerSpacer}
      }
      if(maxRecRuns){
        let narrowRegsRes = checkForNarrowRegions();
        if(narrowRegsRes.rerun && recRun < maxRecRuns){
          if(recRun == 0){
            if(narrowRegsRes.walls.length && narrowRegsRes.spaceEdges.length){
              console.log('reduce spacer for some walls and for some space edges for space', this.name)
            } else if (narrowRegsRes.walls.length){
              console.log('reduce spacer for some walls for space', this.name)
            } else {
              console.log('reduce spacer for some space edges for space', this.name)
            }
            return this.getDijGrid(spaceAssets, isClientRun, pointDens, ++recRun, narrowRegsRes.walls, narrowRegsRes.spaceEdges)
          } else if(isClientRun === false) {
            if(narrowRegsRes.walls.length && narrowRegsRes.spaceEdges.length){
              console.log('increase point density and reduce spacer for some walls and for some space edges for space', this.name)
            } else if (narrowRegsRes.walls.length){
              console.log('increase point density and reduce spacer for some walls for space', this.name)
            } else {
              console.log('increase point density and reduce spacer for some space edges for space', this.name)
            }
            return this.getDijGrid(spaceAssets, isClientRun, Math.round(pointDens/1.25), ++recRun, narrowRegsRes.walls, narrowRegsRes.spaceEdges)
          } else {
            console.log('space', this.name, 'may be lead to problems during navigation due to too less grid points without conflicts');
            return grid2D;
          }
        } else {
          if(recRun >= maxRecRuns){
            console.log('space', this.name, 'may be lead to problems during navigation due to too less grid points without conflicts')
          }
          return grid2D
        }
      } else {
        //grid gets returned without density checks
        return grid2D
      }
      // const maxRecRuns = 3
      // if(numPsWithNoConflicts <= 5 && recRun < maxRecRuns){
      //   return this.getDijGrid(assets, Math.round(pointDensX/2), Math.round(pointDensY/2), defSpacer/2, ++recRun)
      // } else {
      //   if(recRun >= maxRecRuns){
      //     console.log('space ', this.name, ' on floor ', this.__floor.name, ' may be lead to problems during navigation due to too less grid points without conflicts')
      //   } else {
  
      //   }
      //   return grid2D;
      // }
    }

    IfcSpace.prototype.getNumOfAssetsInSpace = function (assets) {
      let shapePs = this.shapePoints(), numAssetsInSpace = 0;
      if(this.isPureRectangleRepresentation()){
        shapePs.push(shapePs[0]);
      }
      for(let asset of assets){
        if(asset.floor === this.__floor.id && isIn2DPolygon(shapePs, {x: asset.xPosition, y: asset.yPosition})){
          numAssetsInSpace++;
        }
      }
      return numAssetsInSpace;
    }

    IfcSpace.prototype.isParkingBox = function () {
      if(this.__category == 'parkingLot'){
        return true;
      }
      return false;
    }

    IfcSpace.prototype.isTechnicalZone = function () {
      if(this.__category == 'technicalZone'){
        return true;
      }
      return false;
    }

    IfcSpace.prototype.isExcludedFromNav = function () {
      if(this.isTechnicalZone()){
        return true;
      }
      return false;
    }

    IfcSpace.prototype.conflictsAsset = function (assets, point, bufferDist) {
      if(assets && assets.length){
        for(const asset of assets){
          if(asset.floor === this.__floor.id && asset.zPosition <= 2000){   //also see: getPointDensXY
            let w = 0, h = 0, x = asset.xPosition, y = asset.yPosition;
  
            //spatial expansion of asset
            if(asset && asset.model3d && asset.model3d.boundingBox){
              w = asset.model3d.boundingBox.width;
              h = asset.model3d.boundingBox.height;
            } else {  //if asset spatial expansion smaller or equal to 500 x 500: don't take it into account
              continue;
            }
  
            let assetBox = [
              {x: x - bufferDist - w / 2, y: y - bufferDist - h / 2},
              {x: x + bufferDist + w / 2, y: y - bufferDist - h / 2},
              {x: x + bufferDist + w / 2, y: y + bufferDist + h / 2},
              {x: x - bufferDist - w / 2, y: y + bufferDist + h / 2}];
            assetBox.push(assetBox[0]);
            if(isIn2DPolygon(assetBox, point)){
              return true;
            }
          }
        }
      }
      return false;
    }

    IfcSpace.prototype.getTransportElementsInSpace = function () {
        let spacePoints = this.shapePoints(),
        tElemRefs = [],
        floorBelow = this.getFloorBelow(this.__floor)
      const addTEs = (relTEs) => {
        for(let TE of relTEs){
          TE.transGeoToFloor(this.__floor)
          if(isIn2DPolygon(spacePoints, TE.getShapeCenter())){
            tElemRefs.push(TE);
          }
          TE.transGeoToFloor(TE.__floor, this.__floor)
        }
      }
      addTEs(this.__floor.getAllRefAndContTransportElems())
      if(floorBelow){
        addTEs(floorBelow.getAllEscalators())
      }
      return tElemRefs;
    }

    IfcSpace.prototype.getAdjacentTransportElements = function () {
      const tEs = this.__floor.getAllRefAndContTransportElems(), spacePoints = this.shapePoints();
      let TERefs = [];
      //const distThreshold = 500;
      for(const tE of tEs){
        if(tE.isElevator()){
          for(let door of tE.__doors){
            if(door.__spaces.length == 1){
              if(door.__spaces[0] == this){
                TERefs.push(tE)
              }
            } else if(door.__floor == this.__floor){
              let doorGeo = door.getGeo()
              if(isIn2DPolygon(spacePoints, doorGeo)){
                TERefs.push(tE)
              }
            }
          }
          // tE.transGeoToFloor(this.__floor);
          // let eWCIP = this.elemsWithCenterInPolygon([tE], spacePoints);
          // TERefs.push(...eWCIP);
          // if(!eWCIP.length){
          //   let TEPoints = tE.shapePoints(), found = false, isRectSpace = this.isPureRectangleRepresentation(), isRectTE = tE.isPureRectangleRepresentation();
          //   const lastP = (pureRect, run) => {
          //     return pureRect ? ((run + 1) % 4) : (run + 1)
          //   };
          //   for(let iii=0; iii<(isRectSpace ? spacePoints.length : spacePoints.length - 1); ++iii){
          //     for(let jjj=0; jjj<(isRectTE ? TEPoints.length : TEPoints.length - 1); ++jjj){
          //       let d1 = distToSegment2D(TEPoints[jjj], spacePoints[iii], spacePoints[lastP(isRectSpace, iii)]),
          //           d2 = distToSegment2D(TEPoints[lastP(isRectTE, jjj)], spacePoints[iii], spacePoints[lastP(isRectSpace, iii)]);
          //       if(d1 <= distThreshold && d2 <= distThreshold){
          //         TERefs.push(tE);
          //         found = true;
          //         break
          //       }
          //     }
          //     if(found){
          //       break
          //     }
          //   }
          // }
          // tE.transGeoToFloor(tE.__floor, this.__floor)
        } else if(tE.isEscalator()){
          //TODO
        } else if(tE.isMovingWalkway()){
          //TODO
        } else {
          console.log('unknwon transport element')
        }
      }
      return TERefs
    }

    IfcSpace.prototype.getFloorBelow = function (floor) {
        let floorBelow;
        if(floor.elevation !== undefined && floor.elevation !== null){
          for(const floor_ of floor.__building.__floors){
            if(floor.id != floor_.id && floor_.elevation !== undefined && floor_.elevation !== null && floor_.elevation < floor.elevation &&
              (!floorBelow || floorBelow.elevation < floor_.elevation)){
                floorBelow = floor_;
            }
          }
        }
        return floorBelow;
    }

    IfcSpace.prototype.elemsWithCenterInPolygon = function (elems, polygon) {
        let res = [];
        for(let elem of elems){
          if(isIn2DPolygon(polygon, elem.getShapeCenter())){
            res.push(elem);
          }
        }
        return res;
    }

    IfcSpace.prototype.intersectWith = function (p1, p2) {
      if(this.__floor.elevation === undefined || this.__floor.elevation === null ||
        p1.z != this.__floor.elevation * 1000 || p2.z != this.__floor.elevation * 1000){
          return false;
      }
      const outerSpacePoints = this.shapePoints(), isRect = this.isPureRectangleRepresentation();
      let intersectCount = 0;
      for(let iii=0;iii<(isRect ? outerSpacePoints.length : outerSpacePoints.length - 1); ++iii){
        if(psIntersect(p1, p2, outerSpacePoints[iii], outerSpacePoints[isRect ? ((iii + 1) % 4) : (iii + 1)])){
          intersectCount++
        }
      }
      return intersectCount
    }

    IfcSpace.prototype.isCloseToEdge = function (edgePoints, point) {
        for (let iii = 0; iii < edgePoints.length; ++iii) {
            if (horiDist(edgePoints[iii], point) <= 800) {
                return true;
            }
        }
        return false;
    }

    IfcSpace.prototype.simpleNonGridBases = function () {
        let geo = this.getGeo();
        let b = geo.getBoundingBox();
        if (this.isPurePolylineRepresentation()) {
            const pointDensX = Math.round(b.width / 1000), pointDensY = Math.round(b.height / 1000);
            let grid = [];
            for (let iii = 0; iii < pointDensX; ++iii) {
                let x = b.x + b.width / pointDensX * iii;
                for (let jjj = 0; jjj < pointDensY; ++jjj) {
                    let y = b.y + b.height / pointDensY * jjj;
                    grid.push({x: x, y: y});
                }
            }
            for (let iii = grid.length - 1; iii >= 0; --iii) {
                if (!isIn2DPolygon(geo.points, grid[iii])) {
                    grid.splice(iii, 1);
                }
            }
            let centers = [];
            const groupSize = Math.round(grid.length / 5);
            // for(let iii=0;iii<grid.length;iii+=groupSize){
            //   let innerGroupSize = iii+groupSize;
            //   if(iii+groupSize > grid.length){
            //     innerGroupSize = grid.length;
            //   }
            //   let center = {x: 0, y:0};
            //   for(let jjj=iii;jjj<innerGroupSize;++jjj){
            //     center.x += grid[jjj].x;
            //     center.y += grid[jjj].y;
            //   }
            //   center.x = center.x / innerGroupSize;
            //   center.y = center.y / innerGroupSize;
            //   centers.push(center);
            // }
            let groups = [];
            while (grid.length) {
                let group = [];
                group.push(grid[0])
                let innerGroupSize = groupSize;
                if (grid.length < innerGroupSize) {
                    innerGroupSize = grid.length;
                }
                for (let iii = 1; iii < innerGroupSize; ++iii) {
                    let shortDIndex = 1, shortD = horiDist(grid[shortDIndex], grid[0]);
                    for (let jjj = 2; jjj < grid.length; ++jjj) {
                        let newD = horiDist(grid[jjj], grid[0]);
                        if (newD < shortD) {
                            shortD = newD;
                            shortDIndex = jjj;
                        }
                    }
                    group.push(grid[shortDIndex]);
                    grid.splice(shortDIndex, 1);
                }
                grid.splice(0, 1);
                groups.push(group);
            }
            for (let iii = 0; iii < groups.length; ++iii) {
                let center = {x: 0, y: 0}, groupLen = groups[iii].length;
                for (let jjj = 0; jjj < groupLen; ++jjj) {
                    center.x += groups[iii][jjj].x / groupLen;
                    center.y += groups[iii][jjj].y / groupLen;
                }
                centers.push(center);
            }
            for (let iii = centers.length - 1; iii >= 0; --iii) {
                if (!isIn2DPolygon(geo.points, centers[iii])) {
                    centers.splice(iii, 1);
                }
            }
            return centers;
        } else {
            return [{x: b.xCenter, y: b.yCenter}];
        }
    }

    IfcSpace.prototype.getAdjacentStairs = function () {
      let spacePoints = this.shapePoints(), result = []
      const maxDistToSegment = 1000
      const isRect = this.isPureRectangleRepresentation();
      for(let floor of this.__floor.__building.__floors){
        for(let stair of floor.__stairs){
          stair.transGeoToFloor(this.__floor);
          for(let node of stair.getDijNodes()){
            if(node.z == this.__floor.elevation * 1000){
              let isClose = false, dist, perp1, perp2
              for(let iii=0; iii<(isRect ? spacePoints.length : spacePoints.length - 1); ++iii){
                const sp1 = spacePoints[iii],
                      sp2 = (isRect ? spacePoints[(iii + 1) % 4] : spacePoints[iii + 1])
                dist = distToSegment2D(node, sp1, sp2)
                if(dist < maxDistToSegment){
                  isClose = true
                  let stairShapePs = stair.shapePoints(), _ssp1, _ssp2, minDist = Number.MAX_SAFE_INTEGER
                  const isStairRect = stair.isPureRectangleRepresentation()
                  for(let jjj=0; jjj<(isStairRect ? stairShapePs.length : stairShapePs.length - 1);++jjj){
                    const ssp1 = stairShapePs[jjj],
                          ssp2 = (isStairRect ? stairShapePs[(jjj + 1) % 4] : stairShapePs[jjj + 1])
                          let stairSegDist = distToSegment2D(node, ssp1, ssp2)
                    if(stairSegDist < minDist){
                      minDist = stairSegDist
                      _ssp1 = ssp1
                      _ssp2 = ssp2
                    }
                  }
                  const norm = Math.sqrt(Math.pow(_ssp2.y - _ssp1.y, 2) + Math.pow(_ssp2.x - _ssp1.x, 2))
                  const perp = {x: -(_ssp2.y - _ssp1.y) / norm, y: (_ssp2.x - _ssp1.x) / norm}
                  perp1 = {x: node.x + perp.x * maxDistToSegment, y: node.y + perp.y * maxDistToSegment}
                  perp2 = {x: node.x - perp.x * maxDistToSegment, y: node.y - perp.y * maxDistToSegment}
                  Object.assign(perp1, {isInSpace: isIn2DPolygon(spacePoints, perp1)})
                  Object.assign(perp2, {isInSpace: isIn2DPolygon(spacePoints, perp2)})
                  Object.assign(perp1, {isInStair: isIn2DPolygon(stairShapePs, perp1)})
                  Object.assign(perp2, {isInStair: isIn2DPolygon(stairShapePs, perp2)})
                  break
                }
              }
              const inPoly = isIn2DPolygon(spacePoints, node)
              if(inPoly && !isClose ||
                isClose && (!perp1.isInStair && perp1.isInSpace || !perp2.isInStair && perp2.isInSpace)){
                result.push(stair)
                break
              }
            }
          }
          stair.transGeoToFloor(stair.__floor, this.__floor);
        }
      }
      return result

      // let result = []
      // if(this.__floor.__stairs.length){
      //   result.push(...this.__floor.__building.__floors[0].__stairs.slice(0, 1))
      // } else {
      //   result.push(...[this.__floor.__building.__floors[0].__stairs[0]])
      // }
      // return result

      // let result = [], spacePoints = this.shapePoints(), relevantStairs = this.__floor.__stairs.slice(0, this.__floor.__stairs.length),
      //   floorBelow = this.getFloorBelow(this.__floor);
      // if(floorBelow !== undefined){
      //   relevantStairs.push(...floorBelow.__stairs)
      // }
      // // for(let stair of relevantStairs){
      // //   stair.transGeoToFloor(this.__floor)
      // // }
      // const stairsWithCenterInPolygon = this.elemsWithCenterInPolygon(relevantStairs, spacePoints);
      // result.push(...stairsWithCenterInPolygon);
      // // let stairsWithCenterInOtherPolygons = [];
      // // for(const space of this.__floor.__spaces){
      // //   if(space !== this){
      // //     stairsWithCenterInOtherPolygons.push(...this.elemsWithCenterInPolygon(relevantStairs, space.shapePoints()))
      // //   }
      // // }
      // // let remainingStairs = relevantStairs.filter(stair => {
      // //   for(const stairWCIP of stairsWithCenterInPolygon){
      // //     if(stair === stairWCIP){
      // //       return false
      // //     }
      // //   }
      // //   for(const stairWCIOP of stairsWithCenterInOtherPolygons){
      // //     if(stair === stairWCIOP){
      // //       return false
      // //     }
      // //   }
      // //   return true
      // // });
      // // for(const stair of remainingStairs){
      // //   if(stair.shapeTypeUndefined()){
      // //     let g = stair.getGeo();
      // //     const distThreshold = 500;
      // //     let p11, p12, p21, p22;
      // //     if(stair.isPureRectangleRepresentation()){
      // //       if(g.width < g.height){
      // //         p11 = { x: g.x, y: g.y },
      // //         p12 = { x: g.x + g.width, y: g.y },
      // //         p21 = { x: g.x, y: g.y + g.height },
      // //         p22 = { x: g.x + g.width, y: g.y +g.height}
      // //       } else {
      // //         p11 = { x: g.x, y: g.y + g.height},
      // //         p12 = { x: g.x, y: g.y },
      // //         p21 = { x: g.x + g.width, y: g.y +g.height},
      // //         p22 = { x: g.x + g.width, y: g.y }
      // //       }
      // //     } else if(stair.isPurePolylineRepresentation() && g.points.length == 5){
      // //       let horiDists = [];
      // //       for(let iii=0;iii<g.points.length - 1; ++iii){
      // //         horiDists.push(horiDist(g.points[iii], g.points[iii + 1]))
      // //       }
      // //       if(horiDists[0] < horiDists[1] && horiDists[0] < horiDists[3] &&
      // //         horiDists[2] < horiDists[1] && horiDists[2] < horiDists[3]){
      // //           p11 = { x: g.points[0].x, y: g.points[0].y },
      // //           p12 = { x: g.points[1].x, y: g.points[1].y },
      // //           p21 = { x: g.points[2].x, y: g.points[2].y },
      // //           p22 = { x: g.points[3].x, y: g.points[3].y }
      // //       } else if(horiDists[1] < horiDists[0] && horiDists[1] < horiDists[2] &&
      // //         horiDists[3] < horiDists[0] && horiDists[3] < horiDists[2]){
      // //           p11 = { x: g.points[1].x, y: g.points[1].y },
      // //           p12 = { x: g.points[2].x, y: g.points[2].y },
      // //           p21 = { x: g.points[3].x, y: g.points[3].y },
      // //           p22 = { x: g.points[4].x, y: g.points[4].y }
      // //       } else {
      // //         console.log('stair representation type not supported')
      // //       }
      // //     } else {
      // //       console.log('stair representation type unknown or not supported')
      // //     }
      // //     if(p11 && p12 && p21 && p22){
      // //       const isRect = this.isPureRectangleRepresentation();
      // //       for(let iii=0; iii<(isRect ? spacePoints.length : spacePoints.length - 1); ++iii){
      // //         const sp1 = spacePoints[iii],
      // //               sp2 = (isRect ? spacePoints[(iii + 1) % 4] : spacePoints[iii + 1]),
      // //               d11 = distToSegment2D(p11, sp1, sp2),
      // //               d12 = distToSegment2D(p12, sp1, sp2),
      // //               d21 = distToSegment2D(p21, sp1, sp2),
      // //               d22 = distToSegment2D(p22, sp1, sp2);
      // //         if(d11 <= distThreshold && d12 <= distThreshold || d21 <= distThreshold && d22 <= distThreshold){
      // //           result.push(stair);
      // //           break
      // //         }
      // //       }
      // //     } else {
      // //       console.log('stair points for heuristic calculation are undefined')
      // //     }
      // //   } else {
      // //     console.log('stair shape type ', stair.shapeType, 'not implemented')
      // //   }
      // // }
      // // for(let stair of relevantStairs){
      // //   stair.transGeoToFloor(stair.__floor, this.__floor)
      // // }
      // return result
    }

    IfcSpace.prototype.getAdjacentRamps = function () {
      let spacePoints = this.shapePoints(), result = [];
      const maxDistToSegment = 1000;
      const isRect = this.isPureRectangleRepresentation();
      for(let floor of this.__floor.__building.__floors){
        for(let ramp of floor.__ramps){
          ramp.transGeoToFloor(this.__floor);
          for(let node of ramp.getDijNodes()){
            if(node.z == this.__floor.elevation * 1000){
              let isClose = false, perp1, perp2;
              for(let iii=0; iii<(isRect ? spacePoints.length : spacePoints.length - 1); ++iii){
                const sp1 = spacePoints[iii], sp2 = (isRect ? spacePoints[(iii + 1) % 4] : spacePoints[iii + 1]);
                if(distToSegment2D(node, sp1, sp2) < maxDistToSegment){
                  isClose = true;
                  let rampShapePs = ramp.shapePoints(), _ssp1, _ssp2, minDist = Number.MAX_SAFE_INTEGER;
                  for(let jjj=0; jjj<rampShapePs.length - 1;++jjj){
                    const ssp1 = rampShapePs[jjj], ssp2 = rampShapePs[jjj + 1]
                          let rampSegDist = distToSegment2D(node, ssp1, ssp2)
                    if(rampSegDist < minDist){
                      minDist = rampSegDist;
                      _ssp1 = ssp1;
                      _ssp2 = ssp2;
                    }
                  }
                  const norm = Math.sqrt(Math.pow(_ssp2.y - _ssp1.y, 2) + Math.pow(_ssp2.x - _ssp1.x, 2))
                  const perp = {x: -(_ssp2.y - _ssp1.y) / norm, y: (_ssp2.x - _ssp1.x) / norm}
                  perp1 = {x: node.x + perp.x * maxDistToSegment, y: node.y + perp.y * maxDistToSegment}
                  perp2 = {x: node.x - perp.x * maxDistToSegment, y: node.y - perp.y * maxDistToSegment}
                  Object.assign(perp1, {isInSpace: isIn2DPolygon(spacePoints, perp1)})
                  Object.assign(perp2, {isInSpace: isIn2DPolygon(spacePoints, perp2)})
                  Object.assign(perp1, {isInRamp: isIn2DPolygon(rampShapePs, perp1)})
                  Object.assign(perp2, {isInRamp: isIn2DPolygon(rampShapePs, perp2)})
                  break
                }    
              }
              const inPoly = isIn2DPolygon(spacePoints, node)
              if(inPoly && !isClose || 
                isClose && (!perp1.isInRamp && perp1.isInSpace || !perp2.isInRamp && perp2.isInSpace)){
                result.push(ramp)
                break
              }
            }
          }
          ramp.transGeoToFloor(ramp.__floor, this.__floor);
        }
      }
      return result
    }

    function IfcDoor() {
    }

// #18364= IFCDOOR('3f78saFoD1wveHNtzlMDbN',#41,'Int_Single_Flush_Deutsch_3XN:1.010 x 2.135:404816',$,'1.010 x 2.135',#259764,#18358,'404816',2135.,1010.);

    IfcDoor.prototype.create = function (project, name) {
      this.type = 'IFCDOOR';
      this.id = getUuid();
      this.name = name;
      this.__project = project;
      this.predefinedType = 'DOOR';
      this.operationType = 'SINGLE_SWING_LEFT';
      this.overallWidth = 1000
      this.overallHeight = 2000

      project.register(this);

      return this;
    };

    IfcDoor.prototype.hasRepresentation = function () {
      return this.__representation ? true : false
    };

    IfcDoor.prototype.centerOrDirDiff = function (otherCenter, otherDir) {
      const center = this.getCenter();
      const dir = this.getDirection();
      if(otherCenter.x != center.x || otherCenter.y != center.y || dir.x != otherDir.x || dir.y != otherDir.y){
        return true;
      }
      return false;
    };

    IfcDoor.prototype.assObjChange = function (prevSpaceIds = [], prevElevatorId = undefined) {
      if(prevElevatorId && !this.__elevator || !prevElevatorId && this.__elevator || prevElevatorId && this.__elevator && prevElevatorId != this.__elevator.id){
        return true;
      }
      for(let space of this.__spaces){
        let found = false;
        for(let id of prevSpaceIds){
          if(id == space.id){
            found = true;
            break;
          }
        }
        if(!found){
          return true;
        }
      }
      return false;
    };

    IfcDoor.prototype.getDirection = function () {
      if(this.__objectPlacement && this.__objectPlacement.__relativePlacement && this.__objectPlacement.__relativePlacement.__direction){
        return this.__objectPlacement.__relativePlacement.__direction.coordinates;
      }
      return null;
    };

    IfcDoor.prototype.getCenter = function () {
      let zPos, b = this.getGeo().getBoundingBox()
      if(this.__floor.elevation !== undefined && this.__floor.elevation !== null){
        zPos = this.__floor.elevation * 1000;
      }
      if(this.__objectPlacement && this.__objectPlacement.__relativePlacement && this.__objectPlacement.__relativePlacement.__direction){
        const dir = this.__objectPlacement.__relativePlacement.__direction.coordinates
        return {x: b.x + 0.5 * dir.x * this.overallWidth, y: b.y + 0.5 * dir.y * this.overallWidth, z: zPos}
      } else {
        return {x: b.xCenter, y: b.yCenter, z: zPos};
      }
    };
    
    IfcDoor.prototype.getExclSurrSpaceIds = function () {
      const res = [];
      const exclSpaces = this.getIfcElementAttributes().exclSurrSpaceIds;
      if(exclSpaces && exclSpaces.length){
        res.push(...exclSpaces);
      }
      return res;
    };
    
    IfcDoor.prototype.getIfcElementAttributes = function () {
      if (!this.__project.__ifcElementAttributes) {
        this.__project.__ifcElementAttributes = {};
      }
      if (!this.__project.__ifcElementAttributes[this.id]) {
        this.__project.__ifcElementAttributes[this.id] = {};
      }
      return this.__project.__ifcElementAttributes[this.id];
    };
    
    IfcDoor.prototype.getExclSurrSpaceIds = function () {
      const res = [];
      const exclSpaces = this.getIfcElementAttributes().exclSurrSpaceIds;
      if(exclSpaces && exclSpaces.length){
        res.push(...exclSpaces);
      }
      return res;
    };

    IfcDoor.prototype.getAdjacentSpaces = function () {
      let res = [], mid;
      const b = this.getGeo().getBoundingBox();
      if (
        this.__objectPlacement &&
        this.__objectPlacement.__relativePlacement &&
        this.__objectPlacement.__relativePlacement.__direction
      ) {
        const dir = this.__objectPlacement.__relativePlacement.__direction.coordinates;
        mid = {x: b.x + 0.5 * dir.x * this.overallWidth, y: b.y + 0.5 * dir.y * this.overallWidth};
      } else {
        mid = {x: b.xCenter, y: b.yCenter};
      }
      if(this.__spaces && this.__spaces.length){
        res.push(...this.__spaces);
        for (const space of this.__floor.__spaces) {
          if (
            !this.__spaces.filter((__space) => {
              if (__space === space) {
                return true;
              }
              return false;
            }).length &&
            isIn2DPolygon(space.shapePoints(), mid, 500)  //500mm to avoid taking explicit door assignments into account
          ) {
            res.push(space);
          }
        }
      }
      return res.filter(e => {
        for(const exclSpaceId of this.getExclSurrSpaceIds()){
          if(exclSpaceId == e.id) return false;
        }
        return true;
      });
    };

    IfcDoor.prototype.scale = function (scaleFactor) {
      const relPlac = this.__objectPlacement.__relativePlacement
      const dirCoords = {x: relPlac.__direction.coordinates.x,
        y: relPlac.__direction.coordinates.y}
      const locCoors = {x: relPlac.__location.coordinates.x * scaleFactor,
          y: relPlac.__location.coordinates.y * scaleFactor}
      this.setPointRepresentation(locCoors.x, locCoors.y, dirCoords.x, dirCoords.y)
  };

    IfcDoor.prototype.remove = function () {
      this.__project.getRelContainedInSpatialStructure(this.__floor).removeRelatedElement(this);

      if(this.__elevator){
        this.__project.getRelReferencedInSpatialStructure(this.__elevator).removeRelatedElement(this)
      }

      for (const n in this.__spaces) {
        this.__project.getRelReferencedInSpatialStructure(this.__spaces[n]).removeRelatedElement(this);
      }

      if (this.__objectPlacement) {
        this.__objectPlacement.remove(this.__project);
      }

      if (this.__representation) {
        this.__representation.remove(this.__project);
      }

      this.__project.unregister(this);
    };

    IfcDoor.prototype.load = function (errors) {
        this.id = this.__attributes[0];
        this.name = this.__attributes[2];
        this.description = this.__attributes[3];
        this.objectType = this.__attributes[4];
        this.objectPlacement = this.__attributes[5];
        this.representation = this.__attributes[6];
        this.overallHeight = this.__attributes[8];
        this.overallWidth = this.__attributes[9];
    };

    IfcDoor.prototype.bind = function (project) {
        this.__project = project;

        if (this.objectPlacement) {
            this.__objectPlacement = project.map[this.objectPlacement];
        }

        if (this.representation) {
            this.__representation = project.map[this.representation];
        }

        this.__spaces = [];
    };

    IfcDoor.prototype.getProject = function () {
        return this.__floor.__building.__site.__project;
    };

    IfcDoor.prototype.getPlacement = function () {
        var geo = this.__representation.getGeo();

        if (this.__objectPlacement) {
            geo.applyPlacement(this.__objectPlacement.getPlacement());
        }

        return geo;
    };

    IfcDoor.prototype.get2DBoundingBox = function () {
        return this.__representation.get2DBoundingBox();
    };

    IfcDoor.prototype.getGeo = function () {
      let geo;

      if (this.__representation && this.__representation.getGeo() &&
        !(this.__representation.getGeo() instanceof GeoNullProfile)) { // TODO Skipped Mapped Representation
        geo = this.__representation.getGeo();
      } else {
        geo = new GeoPolylineProfile().initialize([{x: 0, y: 0}, {x: 1500, y: 0}]);
      }

      if (this.__objectPlacement) {
        geo.applyPlacement(this.__objectPlacement.getPlacement());
      }

      return geo;
    };

    IfcDoor.prototype.setPointRepresentation = function (x, y, xRotation, yRotation) {
      if (!this.__representation) {
        this.__representation = new IfcProductDefinitionShape().create(this.getProject(), '');
        this.representation = this.__representation.oid;
      }

      // Dummy to get a bounding box

      this.__representation.setRectangleRepresentation(this.getProject(), 0, 0, 10, 10);

      if (xRotation !== undefined && yRotation !== undefined) {
        if(this.__objectPlacement){
          this.__objectPlacement.remove(this.__project)
        }
        this.__objectPlacement = new IfcLocalPlacement().create(this.getProject(), x, y, xRotation, yRotation);
        this.objectPlacement = this.__objectPlacement.oid;
      }
    };

    IfcDoor.prototype.dump = function (stream) {
        stream.write('#' + this.oid);
        stream.write('= ');
        stream.write('IFCDOOR(');
        dumpStringAttribute(stream, this.id, true);
        dumpOidAttribute(stream, null);
        dumpStringAttribute(stream, this.name);
        stream.write(');\n');
    };

    IfcDoor.prototype.renderFloorPlanSvg = function (svg, options) {
      if (!this.__representation) {
        return;
      }
  
      const boundingBox = this.getGeo().getBoundingBox();
      let group = svg.group().attr({id: this.id}).addClass('ifcDoor');
  
      group.move(boundingBox.x, boundingBox.y);
  
      const thisRef = this;
      setTimeout(()=>{
        group.text(this.name).font({
          size: 400,
          anchor: 'middle'
        }).attr({
          stroke: 'none',
          fill: options.selectedDoor && options.selectedDoor === this ? 'red' : 'green',
          class: 'counter-rotate',
          x: boundingBox.width / 2 + 500,
          y: boundingBox.height - 1000
        }).on('click', function(){
          event.preventDefault();
          if(!(options && options.editor && options.editor && (options.editor.buttonActionStarted || options.editor.door == thisRef))){
            event.stopPropagation();
            floorplanElemSelectChange.next({obj: thisRef, floor: null});
          }
        });
      });
  
      if(this.__objectPlacement && this.__objectPlacement.__relativePlacement && this.__objectPlacement.__relativePlacement.__direction){
        group.rotate(this.__objectPlacement.__relativePlacement.__direction.getExactRotation())
      }
  
      group.polyline('0,0 ' + this.overallWidth + ',0').attr('stroke', '#EEEEEE').attr('stroke-width', 250).attr('fill', 'none');
      group.polyline('0,0 ' + this.overallWidth + ',0').attr('stroke', '#BBBBBB').attr('stroke-width', 50).attr('fill', 'none');
      group.polyline('0,0 ' + this.overallWidth + ',0').attr('stroke', '#000000').attr('stroke-width', 50).attr('fill', 'none').transform({
        rotation: 45,
        cx: 0,
        cy: 0
      });
  
      if (boundingBox.width == 0) {
        group.transform({rotation: 90, cx: group.bbox().x, cy: group.bbox().y});
      }
  
      this.doorToSpAEl(svg, group, options);
    };

    IfcDoor.prototype.doorToSpAEl = function (svg, group, options) {
      group.off();
      let dragStartPosi, dragEndPosi;
      let doorRef = this;
      group.mouseover(function (event) {
        event.preventDefault();
        this.style('cursor', 'hand');
      }).mouseout(function (event) {
        event.preventDefault();
        this.style('cursor', 'default');
      }).draggable().on('dragstart', function (event) {
        event.preventDefault();
        dragStartPosi = {x: event.detail.handler.p.x, y: event.detail.handler.p.y};
      }).on('dragend', function (event) {
        event.preventDefault();
        dragEndPosi = {x: event.detail.handler.p.x, y: event.detail.handler.p.y};
        if(dragStartPosi.x != dragEndPosi.x || dragStartPosi.y != dragEndPosi.y){
          let editor = options.editor;
          const dragendPos = editor.pageToGeoPoint({x: event.detail.handler.p.x, y: event.detail.handler.p.y});
          const matchResult = editor.floor.getElevatorOrSpaceClosestToPos(editor.geoToWorldPoint(dragendPos));
          const doorBaseLine = new Vector(matchResult.segment[1].x - matchResult.segment[0].x,
            matchResult.segment[1].y - matchResult.segment[0].y, 0).unit().multiply(doorRef.overallWidth);
          const direction = doorBaseLine.unit();
          doorRef.setPointRepresentation(matchResult.x, matchResult.y, direction.x, direction.y);
          group.remove();
          editor.floor.bindDoorsToElASp(doorRef, 500);
          doorRef.renderFloorPlanSvg(svg, options);
          editor.onDoorChange({doors: [doorRef], propagateChange: true});
        }
      }).on('click', function(){
        event.preventDefault();
        // if(dragStartPosi.x == dragEndPosi.x && dragStartPosi.y == dragEndPosi.y){
        //   group.remove();
        //   let _doorRot = -1;
        //   if(options.doorRot) _doorRot *= options.doorRot;
        //   doorRef.renderFloorPlanSvg(svg, {editor: options.editor, doorRot: _doorRot});
        // }
      });
    }

    // Methods for Dijkstra

    IfcDoor.prototype.getDijNodes = function(knownNodes = null){
      if(knownNodes !== null && knownNodes[this.id]){
        return knownNodes[this.id];
      } else {
        let center = this.getCenter();
        Object.assign(center, {id: getUuid(), door: this.id});
        return [center];
      }
    }

    function IfcWall() {
    }

    // #112002= IFCWALL('0eKybLt31EyQVOyyE09nUa',#41,'Basic Wall:Concrete_300_3XN:456398',$,'Basic Wall:Concrete_300_3XN:335904',#111899,#112000,'456398');

    IfcWall.prototype.load = function (errors) {
        this.id = this.__attributes[0];
        this.description = this.__attributes[3];
        this.objectType = this.__attributes[4];
        this.objectPlacement = this.__attributes[5];
        this.representation = this.__attributes[6];
    };

    IfcWall.prototype.scale = function (scaleFactor) {
      let shapePs = this.shapePoints()
      for(let p of shapePs){
        p.x *= scaleFactor
        p.y *= scaleFactor
      }
      this.setPolylineRepresentation(shapePs)
  };

    IfcWall.prototype.isInWallShape = function (point, shapePs = null) {
      if(isIn2DPolygon(shapePs != null ? shapePs : this.shapePoints(), point)){
        return true;
      } else {
        return false;
      }
    };

    IfcWall.prototype.intersectWith = function (p1, p2, minDistToSeg = null) {
      if(this.__floor.elevation === undefined || this.__floor.elevation === null ||
        p1.z != this.__floor.elevation * 1000 || p2.z != this.__floor.elevation * 1000){
          return false;
      }
      const shapePoints = this.shapePoints();
      let intersectCount = 0;
      for(let iii=0;iii<shapePoints.length - 1; ++iii){
        if(psIntersect(p1, p2, shapePoints[iii], shapePoints[iii + 1])){
          intersectCount++
        }
      }
      if(intersectCount == 0 && minDistToSeg !== null){
        for(let iii=0;iii<shapePoints.length; ++iii){
          if(distToSegment2D(shapePoints[iii], p1, p2) < minDistToSeg){
            intersectCount++
          }
        }
      }
      return intersectCount
    };

    IfcWall.prototype.create = function (project) {
        this.type = 'IFCWALL';
        this.id = getUuid();
        this.__project = project;
        project.register(this);
        return this;
      }

      IfcWall.prototype.remove = function () {
        this.__project.getRelContainedInSpatialStructure(this.__floor).removeRelatedElement(this);
        for(let iii=0;iii<this.__floor.__walls.length;++iii){
          if(this.__floor.__walls[iii] === this){
            this.__floor.__walls.splice(iii, 1);
            break;
          }
        }
        if (this.__representation) {
          this.__representation.remove(this.__project);
        }
        if (this.__objectPlacement) {
          this.__objectPlacement.remove(this.__project);
        }
        this.__project.unregister(this);
      }

      IfcWall.prototype.getProject = function () {
        return this.__project
      }

      IfcWall.prototype.setPolylineRepresentation = function (points) {
        if (!this.__representation) {
            this.__representation = new IfcProductDefinitionShape().create(this.getProject(), '');
            this.representation = this.__representation.oid;
          }
          this.__representation.setPolylineRepresentation(this.getProject(), points);
          return this
      }

    IfcWall.prototype.bind = function (project) {
        if (this.objectPlacement) {
            this.__objectPlacement = project.map[this.objectPlacement];
        }

        if (this.representation) {
            this.__representation = project.map[this.representation];
        }
    };

    IfcWall.prototype.getPlacement = function () {
        return this.__objectPlacement.getPlacement();
    };

    IfcWall.prototype.get2DBoundingBox = function () {
        return this.__representation.get2DBoundingBox();
    };

    IfcWall.prototype.getGeo = function () {
        var geo = this.__representation.getGeo();

        if (this.__objectPlacement) {
            geo.applyPlacement(this.__objectPlacement.getPlacement());
        }

        return geo;
    };

    IfcWall.prototype.renderFloorPlanSvg = function (svg, options) {
      if (!this.__representation) {
        return;
      }

      let color = '#333333';
      if(options && options.relWallsIds){
        for(const relWallId of options.relWallsIds){
          if(relWallId == this.id){
            color = 'red';
            break
          }
        }
      }

      const group = this.getGeo().renderSvg(svg).attr('id', this.id).addClass('ifcWall').attr('stroke', color).attr('fill', 'none');

      if (this.__properties && this.__properties.Reference) {
        const reference = this.__properties.Reference.split('_');

        if (reference.length >= 3) {
          group.attr('stroke-width', Number(reference[1]));
        } else {
          group.attr('stroke-width', 50);
        }
      } else {
        group.attr('stroke-width', 50);
      }
      return this
    };

    IfcWall.prototype.shapePoints = function (){
      let shapePoints = this.getGeo().points;
      if(this.__floor.elevation !== undefined && this.__floor.elevation !== null){
        for(let point of shapePoints){
          Object.assign(point, {z: this.__floor.elevation * 1000})
        }
      }
      return shapePoints;
    }


    function IfcWallStandardCase() {
    }

    // #112002= IfcWallStandardCase('0eKybLt31EyQVOyyE09nUa',#41,'Basic Wall:Concrete_300_3XN:456398',$,'Basic Wall:Concrete_300_3XN:335904',#111899,#112000,'456398');

    IfcWallStandardCase.prototype.load = function (errors) {
        this.id = this.__attributes[0];
        this.name = this.__attributes[2];
        this.description = this.__attributes[3];
        this.objectType = this.__attributes[4];
        this.objectPlacement = this.__attributes[5];
        this.representation = this.__attributes[6];
    };

    IfcWallStandardCase.prototype.bind = function (project) {
        if (this.objectPlacement) {
            this.__objectPlacement = project.map[this.objectPlacement];
        }

        if (this.representation) {
            this.__representation = project.map[this.representation];
        }

        var name = this.name.split(':');
    };

    IfcWallStandardCase.prototype.getPlacement = function () {
        return this.__objectPlacement.getPlacement();
    };

    IfcWallStandardCase.prototype.get2DBoundingBox = function () {
        return this.__representation.get2DBoundingBox();
    };

    IfcWallStandardCase.prototype.getGeo = function () {
        var geo = this.__representation.getGeo();

        if (this.__objectPlacement) {
            geo.applyPlacement(this.__objectPlacement.getPlacement());
        }

        return geo;
    };

    IfcWallStandardCase.prototype.renderFloorPlanSvg = function (svg) {
        if (!this.__representation) {
            return;
        }

        var group = this.getGeo().renderSvg(svg).attr('id', this.id).addClass('ifcWallStandardCase').attr('stroke', '#333333').attr('stroke-width', 220).attr('fill', 'none');
    };

    function IfcSlab() {
    }

// #5565= IFCSLAB('3uHBCVHL9AVey1uj8WPv85',#41,'Floor:Concrete_250_3XN:395690',$,'Floor:Concrete_250_3XN',#5477,#5563,'395690',.FLOOR.);

    IfcSlab.prototype.load = function (errors) {
        this.id = this.__attributes[0];
        this.name = this.__attributes[2];
        this.description = this.__attributes[3];
        //this.objectType = this.__attributes[4];
        //this.objectPlacement = this.__attributes[5];
        //this.representation = this.__attributes[6];
    };

    IfcSlab.prototype.bind = function (project) {
        //if (this.objectPlacement) {
        //    this.__objectPlacement = project.map[this.objectPlacement];
        //}
        //
        //if (this.representation) {
        //    this.__representation = project.map[this.representation];
        //}
        //
        //var name = this.name.split(':');
    };


    function IfcFlowTerminal() {
    }

// #27231= IFCFLOWTERMINAL('0grQ9iUdr0QPyleyajtTAZ',#41,'Toilet_Wall_Small_3D_3XN:Toilet_Wall_Small_3D_3XN:407908',$,'Toilet_Wall_Small_3D_3XN',#27229,#27220,'407908');

    IfcFlowTerminal.prototype.load = function (errors) {
        this.id = this.__attributes[0];
        this.name = this.__attributes[2];
        this.description = this.__attributes[3];
        //this.objectType = this.__attributes[4];
        //this.objectPlacement = this.__attributes[5];
        //this.representation = this.__attributes[6];
    };

    IfcFlowTerminal.prototype.bind = function (project) {
        //if (this.objectPlacement) {
        //    this.__objectPlacement = project.map[this.objectPlacement];
        //}
        //
        //if (this.representation) {
        //    this.__representation = project.map[this.representation];
        //}
        //
        //var name = this.name.split(':');
    };

    function IfcBuildingElementProxy() {
    }

// #34051= IFCBUILDINGELEMENTPROXY('1jMhq$_cr0PBX0Wl_PanPp',#41,'Slab Edge:Slab Edge_200x400',$,'Slab Edge:Slab Edge_200x400',#33930,#34049,'441721',$);

    IfcBuildingElementProxy.prototype.load = function (errors) {
        this.id = this.__attributes[0];
        this.name = this.__attributes[2];
        this.description = this.__attributes[3];
        //this.objectType = this.__attributes[4];
        //this.objectPlacement = this.__attributes[5];
        //this.representation = this.__attributes[6];
    };

    IfcBuildingElementProxy.prototype.bind = function (project) {
        //if (this.objectPlacement) {
        //    this.__objectPlacement = project.map[this.objectPlacement];
        //}
        //
        //if (this.representation) {
        //    this.__representation = project.map[this.representation];
        //}
        //
        //var name = this.name.split(':');
    };

    function IfcColumn() {
    }

//    #46648= IFCCOLUMN('3BQudNg$zEmfOe_pZzRHNZ',#41,'Round_3XN:Concrete_500mm:445422',$,'Concrete_500mm',#46647,#46642,'445422');

    IfcColumn.prototype.load = function (errors) {
        this.id = this.__attributes[0];
        this.name = this.__attributes[2];
        this.description = this.__attributes[3];
        //this.objectType = this.__attributes[4];
        //this.objectPlacement = this.__attributes[5];
        //this.representation = this.__attributes[6];
    };

    IfcColumn.prototype.bind = function (project) {
        //if (this.objectPlacement) {
        //    this.__objectPlacement = project.map[this.objectPlacement];
        //}
        //
        //if (this.representation) {
        //    this.__representation = project.map[this.representation];
        //}
        //
        //var name = this.name.split(':');
    };

    const IfcRampTypeEnum = {
      STRAIGHT_RUN_RAMP: 0, 
      TWO_STRAIGHT_RUN_RAMP: 1, 
      QUARTER_TURN_RAMP: 2, 
      TWO_QUARTER_TURN_RAMP: 3, 
      HALF_TURN_RAMP: 4, 
      SPIRAL_RAMP: 5, 
      USERDEFINED: 6, 
      NOTDEFINED: 7
    }

    function IfcRamp(){}

    IfcRamp.prototype.create = function (project, name) {
      this.type = 'IFCRAMP';
      this.id = getUuid();
      this.__project = project;
      this.name = name;
      this.shapeType = IfcRampTypeEnum.NOTDEFINED;
      project.register(this);
      return this;
    }
  
    IfcRamp.prototype.initialize = function () {
      return this;
    }

    IfcRamp.prototype.getProject = function () {
      return this.__project;
    }

    IfcRamp.prototype.remove = function () {
      this.__project.getRelContainedInSpatialStructure(this.__floor).removeRelatedElement(this);

      for(let iii=0;iii<this.__floor.__ramps.length;++iii){
        if(this.__floor.__ramps[iii].id === this.id){
          this.__floor.__ramps.splice(iii, 1);
          break;
        }
      }
  
      if (this.__representation) {
        this.__representation.remove(this.__project);
      }
      if (this.__objectPlacement) {
        this.__objectPlacement.remove(this.__project);
      }
      this.removeIfcElementAttributes();
      this.__project.unregister(this);
    }

    IfcRamp.prototype.hasRepresentation = function () {
      return this.representation != null;
    }

    IfcRamp.prototype.load = function (errors) {
      this.id = this.__attributes[0];
      this.name = this.__attributes[2];
      this.description = this.__attributes[3];
      this.objectPlacement = this.__attributes[5];
      this.representation = this.__attributes[6];
    }

    IfcRamp.prototype.bind = function (project) {
      this.__project = project;
      if (this.representation) {
         this.__representation = project.map[this.representation];
      }
      if (this.objectPlacement) {
        this.__objectPlacement = this.__project.map[this.objectPlacement];
      }
    }

    IfcRamp.prototype.shapeTypeUndefined = function () {
      if(this.shapeType === IfcRampTypeEnum.NOTDEFINED){
        return true;
      } else {
        return false;
      }
    }

    IfcRamp.prototype.getGeo = function () {
      if (!this.__representation) {
        return new GeoNullProfile();
      }
      const geo = this.__representation.getGeo();
      if (this.__objectPlacement) {
        geo.applyPlacement(this.__objectPlacement.getPlacement());
      }
      return geo;
    }

    IfcRamp.prototype.getShapeCenter = function () {
      return visualCenter(this.shapePoints());
    }

    IfcRamp.prototype.getDijNodes = function () {
      if(this.shapeTypeUndefined()){
        let nodes = this.getEntryExitPs();
        for(let node of nodes){
          Object.assign(node, {rampId: this.id})
        }
        return nodes
      } else {
        console.log('ramp type ', this.shapeType, 'not implemented');
        return [];
      }
    }

    IfcRamp.prototype.getDijGrid = function () {
      if(this.shapeTypeUndefined()){
        let floorAbove = this.__floor.getFloorAbove(), center0 = this.getShapeCenter();
        if(!floorAbove || !center0){
          return [];
        }
        const setCenter = (center) => {
          center.z = (this.__floor.elevation + (floorAbove.elevation - this.__floor.elevation) / 2) * 1000;
          Object.assign(center, {rampId: this.id})
        };
        setCenter(center0);
        this.transGeoToFloor(floorAbove);
        let center1 = this.getShapeCenter();
        this.transGeoToFloor(this.__floor, floorAbove);
        setCenter(center1);
        if(center1 && ( center0.x != center1.x || center0.y != center1.y )){
          return [center0, center1]
        } else {
          return [center0]
        }
      } else {
        console.log('ramp type ', this.shapeType, 'not implemented');
        return [];
      }
    }

    IfcRamp.prototype.renderFloorPlanSvg = function (svg, options, floor) {
      if (!this.__representation) {
        return;
      }
      const thisRef = this;
      const geo = this.transGeoToFloor(floor);
      const group = geo.renderSvg(svg).attr('id', this.id).addClass('ifcRamp').attr('id', this.id).attr('stroke', 'green')
        .attr('stroke-width', 50).attr('fill', 'none');
      this.transGeoToFloor(this.__floor, floor);
      const boundingBox = geo.getBoundingBox();
      group.text(this.name).font({
        size: 400,
        anchor: 'middle'
      }).attr({
        stroke: 'none',
        fill: 'green',
        class: 'counter-rotate',
        x: boundingBox.x + 0.5 * boundingBox.width,
        y: boundingBox.y + 0.5 * boundingBox.height - 400
      }).on('click', function(){
        event.preventDefault();
        if(!(options && options.editor && options.editor && (options.editor.buttonActionStarted || options.editor.ramp == thisRef))){
          event.stopPropagation();
          floorplanElemSelectChange.next({obj: thisRef, floor: null});
        }
      }).mouseover(function (event) {
        event.preventDefault();
        if(!(options && options.editor && options.editor && options.editor.buttonActionStarted)){
          this.style('cursor', 'hand');
        } else {
          this.style('cursor', 'crosshair');
        }
      }).mouseout(function (event) {
        event.preventDefault();
        this.style('cursor', 'default');
      });
    }

    IfcRamp.prototype.centerMatch = function (otherRamp, threshold = 500) {
      const thisB = this.getGeo().getBoundingBox(), otherRampB = otherRamp.getGeo().getBoundingBox();
      return (horiDist({x: thisB.xCenter, y: thisB.yCenter}, {x: otherRampB.xCenter, y: otherRampB.yCenter}) <= threshold);
    }

    IfcRamp.prototype.move = function (diffX, diffY) {
      let g = this.getGeo();
      let shapePoints = g.points;
      for(let p of shapePoints){
        p.x -= diffX;
        p.y -= diffY;
      }
      this.setPolylineRepresentation(shapePoints);
    }

    IfcRamp.prototype.setPolylineRepresentation = function (points) {
      if (!this.__representation) {
        this.__representation = new IfcProductDefinitionShape().create(this.getProject(), '');
        this.representation = this.__representation.oid;
      }
      this.__representation.setPolylineRepresentation(this.getProject(), points);
    }

    IfcRamp.prototype.adaptRepresentation = function (ramps) {
      for(const ramp of ramps){
        if(ramp.hasRepresentation()){
          ramp.transGeoToFloor(this.__floor)
          let geo = ramp.getGeo();
          ramp.transGeoToFloor(ramp.__floor, this.__floor)
          this.setPolylineRepresentation(geo.points)
          let ps = ramp.getEntryExitPs(), nPs = []
          const zDiff = (this.__floor.elevation - ramp.__floor.elevation) * 1000, thisOrigin = this.__floor.getOrigin(), 
            rampOrigin = ramp.__floor.getOrigin(), floorAboveThisFloorOrigin = this.__floor.getFloorAbove().getOrigin(), 
            floorAboveRampFloorOrigin = ramp.__floor.getFloorAbove().getOrigin()
          for(let p of ps){
            let nP = Object.assign({}, p)
            if(nP.id){
              delete nP.id
            }
            const moveP = (o1, o2) => {
              nP.x -= o2.x - o1.x
              nP.y -= o2.y - o1.y
            }
            if(nP.z > ramp.__floor.elevation * 1000){
              moveP(floorAboveThisFloorOrigin, floorAboveRampFloorOrigin)
            } else {
              moveP(thisOrigin, rampOrigin)
            }
            nP.z += zDiff
            nPs.push(nP)
          }
          this.setEntryExitPs(nPs)
          break
        }
      }
      return this
    }

    IfcRamp.prototype.getAdjacentSpaces = function () {
      let adjSpaces = [];
      for(const floor of this.__floor.__building.__floors){
        for(const space of floor.__spaces){
          const rampRefs = space.getAdjacentRamps();
           for(const rampRef of rampRefs){
            if(rampRef === this){
              adjSpaces.push(space);
            }
          }
        }
      }
      return adjSpaces;
    }

    IfcRamp.prototype.transGeoToFloor = function (toFloor, fromFloor = this.__floor) {
      let geo = this.getGeo();
      const toFloorOrigin = toFloor.getOrigin(), fromFloorOrigin = fromFloor.getOrigin();
      for(let p of geo.points){
        p.x -= fromFloorOrigin.x - toFloorOrigin.x;
        p.y -= fromFloorOrigin.y - toFloorOrigin.y
      }
      this.setPolylineRepresentation(geo.points)
      return geo
    }

    IfcRamp.prototype.shapePoints = function () {
      let g = this.getGeo();
      let shapePoints = g.points;
      if(this.__floor.elevation !== undefined && this.__floor.elevation !== null){
        for(let point of shapePoints){
          Object.assign(point, {z: this.__floor.elevation * 1000})
        }
      }
      return shapePoints;
    }

    IfcRamp.prototype.intersectWith = function (p1, p2) {
      const relFloors = [this.__floor], floorAbove = this.__floor.getFloorAbove();
      if(floorAbove){
        relFloors.push(floorAbove)
      }
      const outerRampPoints = this.shapePoints();
      for(const floor of relFloors){
        if(floor.elevation === undefined || floor.elevation === null || p1.z != floor.elevation * 1000 || p2.z != floor.elevation * 1000){
          continue
        }
        let intersectCount = 0;
        for(let iii=0;iii<outerRampPoints.length - 1; ++iii){
          if(psIntersect(p1, p2, outerRampPoints[iii], outerRampPoints[iii + 1])){
            intersectCount++
          }
        }
        if(intersectCount > 0){
          return intersectCount
        }
      }
      return false;
    }

    IfcRamp.prototype.getIfcElementAttributes = function () {
      if (!this.__project.__ifcElementAttributes) {
        this.__project.__ifcElementAttributes = {};
      }
      if (!this.__project.__ifcElementAttributes[this.id]) {
        this.__project.__ifcElementAttributes[this.id] = {};
      }
      return this.__project.__ifcElementAttributes[this.id];
    }

    IfcRamp.prototype.removeIfcElementAttributes = function () {
      if(this.__project.__ifcElementAttributes && this.__project.__ifcElementAttributes[this.id]){
        delete this.__project.__ifcElementAttributes[this.id]
      }
    }

    IfcRamp.prototype.setEntryExitPs = function (nodes) {
      for(let node of nodes){
        if(!node.id){
          Object.assign(node, {id: getUuid()})
        }
      }
      this.getIfcElementAttributes().entryExitPs = nodes
    }

    IfcRamp.prototype.getEntryExitPs = function () {
      let res = this.getIfcElementAttributes().entryExitPs;
      if(res){
        for(let node of res){
          if(!node.id){
            this.setEntryExitPs(res)
            if (this.__component) {
              this.__component.rampChange(this);
            }
            break
          }
        }
        return res
      } else {
        return []
      }
    }

    IfcRamp.prototype.moveEntryExitPs = function (diffX, diffY) {
      let entryExitPs = this.getEntryExitPs()
      for(let node of entryExitPs){
        node.x -= diffX
        node.y -= diffY
      }
      this.setEntryExitPs(entryExitPs)
    }

    const IfcTransportElementTypeEnum = {
      ELEVATOR: 0,
      ESCALATOR: 1,
      MOVINGWALKWAY: 2,
      USERDEFINED: 3,
      NOTDEFINED: 4
    }

    function IfcTransportElement(){}

    IfcTransportElement.prototype.create = function (project, name) {
        this.type = 'IFCTRANSPORTELEMENT';
        this.id = getUuid();
        this.__project = project;
        this.name = name;
        this.__refFloors = [];
        this.__doors = []
        this.OperationType = IfcTransportElementTypeEnum.NOTDEFINED;
        project.register(this);
        return this;
    }

    IfcTransportElement.prototype.asElevator = function () {
        this.OperationType = IfcTransportElementTypeEnum.ELEVATOR;
        return this;
    }

    IfcTransportElement.prototype.scale = function (scaleFactor) {
      if(this.isElevator()){
        if(this.isPurePolylineRepresentation()){
          let shapePs = this.shapePoints()
          for(let p of shapePs){
            p.x *= scaleFactor
            p.y *= scaleFactor
          }
          this.setPolylineRepresentation(shapePs)
        } else {
          let geo = this.getGeo()
          geo.x *= scaleFactor
          geo.y *= scaleFactor
          geo.width *= scaleFactor
          geo.height *= scaleFactor
          this.setRectangleRepresentation(geo.x, geo.y, geo.width, geo.height)
        }
      }
    }

    IfcTransportElement.prototype.asMovingWalkway = function () {
        this.OperationType = IfcTransportElementTypeEnum.MOVINGWALKWAY;
        return this;
    }

    IfcTransportElement.prototype.isElevator = function () {
        if(this.OperationType === IfcTransportElementTypeEnum.ELEVATOR){
            return true;
          } else {
            return false;
          }
    }

    IfcTransportElement.prototype.isEscalator = function () {
        if(this.OperationType === IfcTransportElementTypeEnum.ESCALATOR){
            return true;
          } else {
            return false;
          }
    }

    IfcTransportElement.prototype.isMovingWalkway = function () {
        if(this.OperationType === IfcTransportElementTypeEnum.MOVINGWALKWAY){
            return true;
          } else {
            return false;
          }
    }

    IfcTransportElement.prototype.hasRepresentation = function () {
        return this.representation != null;
    }

    IfcTransportElement.prototype.getAdjacentSpaces = function () {
        let adjSpaces = [];
        for(const floor of this.__floor.__building.__floors){
          for(const space of floor.__spaces){
            const TERefs = space.getAdjacentTransportElements()
             for(const TERef of TERefs){
              if(TERef === this){
                adjSpaces.push(space);
              }
            }
          }
        }
        return adjSpaces;
    }

    IfcTransportElement.prototype.load = function (errors) {
        this.id = this.__attributes[0];
        this.name = this.__attributes[2];
        this.description = this.__attributes[3];
        this.objectPlacement = this.__attributes[5];
        this.representation = this.__attributes[6];
    }

    IfcTransportElement.prototype.remove = function () {
      for(let iii=this.__doors.length - 1;iii>=0;--iii){
        if(this.isElevator()){
          this.__doors[iii].__elevator = undefined
        }
      }
      this.__project.getRelReferencedInSpatialStructure(this).remove();
      this.__project.getRelContainedInSpatialStructure(this.__floor).removeRelatedElement(this);
      for(let n in this.__floor.__transportElements){
        if(this.__floor.__transportElements[n] === this){
          this.__floor.__transportElements.splice(n, 1);
          break;
        }
      }
      for(const floor of this.__refFloors){
        this.__project.getRelReferencedInSpatialStructure(floor).removeRelatedElement(this);
        for(let n in floor.__refTransportElements){
          if(floor.__refTransportElements[n] === this){
            floor.__refTransportElements.splice(n, 1);
            break;
          }
        }
      }
      this.__refFloors = [];
      if (this.__representation) {
        this.__representation.remove(this.__project);
      }
      if (this.__objectPlacement) {
        this.__objectPlacement.remove(this.__project);
      }
      this.__project.unregister(this);
    }

    IfcTransportElement.prototype.bind = function (project) {
        this.__project = project;
        this.__refFloors = [];
        this.__doors = []
        if (this.representation) {
           this.__representation = project.map[this.representation];
        }
        if (this.objectPlacement) {
          this.__objectPlacement = this.__project.map[this.objectPlacement];
        }
    }

    IfcTransportElement.prototype.getShapeCenter = function () {
        let center
        if(this.isElevator()){
          if(this.isPureRectangleRepresentation()){
            center = centerOfMass(this.shapePoints())
          } else if(this.isPurePolylineRepresentation()){
            // //TODO
            // let points = this.shapePoints()
            // center = centerOfMass(points.slice(0, points.length - 1)) //only for testing purposes
            center = visualCenter(this.shapePoints());
          } else {
            console.log('representation type unknwon')
          }
        } else if(this.isEscalator()){
          //TODO
        } else if(this.isMovingWalkway()){
          //TODO
        } else {
          console.log('unknown transport element type')
        }
        return center
    }

    IfcTransportElement.prototype.getDijNodes = function () {
        let nodes = [];
        if(this.isElevator()){
          for(const door of this.__doors){
            let doorNodes = door.getDijNodes();
            for(let doorNode of doorNodes){
              let newNode = JSON.parse(JSON.stringify(doorNode));
              Object.assign(newNode, {id: getUuid()});
              nodes.push(newNode);
            }
          }
          // for(const floor of [this.__floor, ...this.__refFloors]){
          //   this.transGeoToFloor(floor)
          //   let TECenter = this.getShapeCenter()
          //   if(TECenter && floor.elevation !== undefined && floor.elevation !== null){
          //     nodes.push(JSON.parse(JSON.stringify(Object.assign(TECenter, {z: floor.elevation * 1000, id: getUuid()}))))
          //   }
          //   this.transGeoToFloor(this.__floor, floor)
          // }
        } else if(this.isEscalator()){
          //TODO
        } else if(this.isMovingWalkway()){
          //TODO
        } else {
          console.log('unknown transport element type')
        }
        for(let node of nodes){
          Object.assign(node, {transportElementId: this.id})
        }
        return nodes
    }

    IfcTransportElement.prototype.getDijGrid = function () {
        let grid = [];
        if(this.isElevator()){
          const center = this.getShapeCenter(), relFloors = [this.__floor, ...this.__refFloors];
          center.transportElementId = this.id;
          if (!center) {
            return [];
          }
          const setZ = (z) => {
            center.z = (z);
            grid.push(JSON.parse(JSON.stringify(center)));
          }
          for (const floor of relFloors.sort((a, b) => a.elevation - b.elevation)) {
            const floorAbove = floor.getFloorAbove(relFloors);
            if (floorAbove) {
              setZ(floor.elevation * 1000);
              // setZ((floor.elevation  + floorAbove.elevation ) * 1000 / 2);
              setZ(floorAbove.elevation * 1000);
            }
          }
        } else if(this.isEscalator()){
          //TODO
        } else if(this.isMovingWalkway()){
          //TODO
        } else {
          console.log('unknown transport element type')
        }
        return grid
    }

    IfcTransportElement.prototype.setRectangleRepresentation = function (x, y, width, height) {
        if (!this.__representation) {
            this.__representation = new IfcProductDefinitionShape().create(this.__project, '');
            this.representation = this.__representation.oid;
          }
          this.__representation.setRectangleRepresentation(this.__project, x + 0.5 * width, y + 0.5 * height, width, height);
    }

    IfcTransportElement.prototype.setPolylineRepresentation = function (points) {
        if (!this.__representation) {
            this.__representation = new IfcProductDefinitionShape().create(this.__project, '');
            this.representation = this.__representation.oid;
          }
          this.__representation.setPolylineRepresentation(this.__project, points);
    }

    IfcTransportElement.prototype.centerMatch = function (otherTransportElement, threshold = 500) {
        const thisB = this.getGeo().getBoundingBox(), otherTransportElementB = otherTransportElement.getGeo().getBoundingBox();
        return (horiDist({x: thisB.xCenter, y: thisB.yCenter}, {x: otherTransportElementB.xCenter, y: otherTransportElementB.yCenter}) <= threshold);
    }

    IfcTransportElement.prototype.renderFloorPlanSvg = function (svg, options, floor) {
      if (!this.__representation) {
        return;
      }
      const thisRef = this;
      let geo = this.transGeoToFloor(floor);
      const group = geo.renderSvg(svg).attr('id', this.id).addClass('ifcTransportElement').attr('id', this.id).attr('stroke', 'green')
        .attr('stroke-width', 50).attr('fill', 'none');
      this.transGeoToFloor(this.__floor, floor);
      const boundingBox = geo.getBoundingBox();
      group.text(this.name).font({
        size: 400,
        anchor: 'middle'
      }).attr({
        stroke: 'none',
        fill: 'green',
        class: 'counter-rotate',
        x: boundingBox.x + 0.5 * boundingBox.width,
        y: boundingBox.y + 0.5 * boundingBox.height - 400
      }).on('click', function(){
        event.preventDefault();
        if(!(options && options.editor && options.editor && (options.editor.buttonActionStarted || options.editor.transportElement == thisRef))){
          event.stopPropagation();
          floorplanElemSelectChange.next({obj: thisRef, floor: null});
        }
      }).mouseover(function (event) {
        event.preventDefault();
        if(!(options && options.editor && options.editor && options.editor.buttonActionStarted)){
          this.style('cursor', 'hand');
        } else {
          this.style('cursor', 'crosshair');
        }
      }).mouseout(function (event) {
        event.preventDefault();
        this.style('cursor', 'default');
      });
    }

    IfcTransportElement.prototype.transGeoToFloor = function(toFloor, fromFloor = this.__floor){
        let geo = this.getGeo()
        const toFloorOrigin = toFloor.getOrigin(), fromFloorOrigin = fromFloor.getOrigin()
        if(this.isPurePolylineRepresentation()){
          for(let p of geo.points){
            p.x -= fromFloorOrigin.x - toFloorOrigin.x
            p.y -= fromFloorOrigin.y - toFloorOrigin.y
          }
          this.setPolylineRepresentation(geo.points)
        } else if(this.isPureRectangleRepresentation()){
          geo.x -= fromFloorOrigin.x - toFloorOrigin.x
          geo.y -= fromFloorOrigin.y - toFloorOrigin.y
          this.setRectangleRepresentation(geo.x, geo.y, geo.width, geo.height)
        } else {
          console.log('unknown transport element representation; return unaltered geometry');
        }
        return geo
      }

    IfcTransportElement.prototype.getGeo = function () {
        if (!this.__representation) {
            return new GeoNullProfile();
          }
          const geo = this.__representation.getGeo();
          if (this.__objectPlacement) {
            geo.applyPlacement(this.__objectPlacement.getPlacement());
          }
          return geo;
    }

    IfcTransportElement.prototype.intersectWith = function (p1, p2) {
      if(this.isElevator()){
        const outerTElemPoints = this.shapePoints(), isRect = this.isPureRectangleRepresentation();
        for(const floor of [this.__floor, ...this.__refFloors]){
          if(floor.elevation !== undefined && floor.elevation !== null && (p1.z != floor.elevation * 1000 || p2.z != floor.elevation * 1000)){
            continue
          }
          let intersectCount = 0;
          for(let iii=0;iii<(isRect ? outerTElemPoints.length : outerTElemPoints.length - 1); ++iii){
            if(psIntersect(p1, p2, outerTElemPoints[iii], outerTElemPoints[isRect ? ((iii + 1) % 4) : (iii + 1)])){
              intersectCount++
            }
          }
          if(intersectCount > 0){
            return intersectCount
          }
        }
        return false
      } else if(this.isEscalator()){
        //TODO
        return false
      } else if(this.isMovingWalkway()){
        //TODO
        return false
      } else {
        console.log('unkwnon transport element');
        return false
      }
    }

    IfcTransportElement.prototype.shapePoints = function () {
        let g = this.getGeo();
        let shapePoints = [];
        if(this.isPurePolylineRepresentation()){
          shapePoints = g.points;
        } else if(this.isPureRectangleRepresentation()){
          shapePoints.push({x: g.x, y: g.y});
          shapePoints.push({x: g.x + g.width, y: g.y});
          shapePoints.push({x: g.x + g.width, y: g.y + g.height});
          shapePoints.push({x: g.x, y: g.y + g.height});
        } else {
          console.log('unknown transport element representation')
          return []
        }
        if(this.__floor.elevation !== undefined && this.__floor.elevation !== null){
          for(let point of shapePoints){
            Object.assign(point, {z: this.__floor.elevation * 1000})
          }
        }
        return shapePoints;
    };

    IfcTransportElement.prototype.isPurePolylineRepresentation = function () {
        return this.isPureXRepresentation('IFCARBITRARYCLOSEDPROFILEDEF');
    }

    IfcTransportElement.prototype.isPureRectangleRepresentation = function () {
        return this.isPureXRepresentation('IFCRECTANGLEPROFILEDEF');
    }

    IfcTransportElement.prototype.isPureXRepresentation = function (repType) {
        if(this.hasRepresentation()){
            if(this.__representation.__representations){
              for(const rep of this.__representation.__representations){
                if(rep.__items){
                  for(const item of rep.__items){
                    if(item.__sweptArea && item.__sweptArea.type !== repType){
                      return false;
                    }
                  }
                }
              }
            }
            return true;
          }
          return false;
    }

    const IfcStairTypeEnum = {
        STRAIGHT_RUN_STAIR: 0,
        TWO_STRAIGHT_RUN_STAIR: 1,
        QUARTER_WINDING_STAIR: 2,
        QUARTER_TURN_STAIR: 3,
        HALF_WINDING_STAIR: 4,
        HALF_TURN_STAIR: 5,
        TWO_QUARTER_WINDING_STAIR: 6,
        TWO_QUARTER_TURN_STAIR: 7,
        THREE_QUARTER_WINDING_STAIR: 8,
        THREE_QUARTER_TURN_STAIR: 9,
        SPIRAL_STAIR: 10,
        DOUBLE_RETURN_STAIR: 11,
        CURVED_RUN_STAIR: 12,
        TWO_CURVED_RUN_STAIR: 13,
        USERDEFINED: 14,
        NOTDEFINED: 15
    }

    function IfcStair() {
    }

//    #46648= IFCSTAIR('3BQudNg$zEmfOe_pZzRHNZ',#41,'Round_3XN:Concrete_500mm:445422',$,'Concrete_500mm',#46647,#46642,'445422');

    IfcStair.prototype.initialize = function() {
        return this;
    }

    IfcStair.prototype.getProject = function() {
        return this.__project;
    }

    IfcStair.prototype.scale = function(scaleFactor) {
      const entryExitPs = this.getEntryExitPs();
      const floorAbove = this.__floor.getFloorAbove(), origin = this.getOrigin(), floorAboveOrigin = floorAbove.getOrigin();
      const oDiff = {x: floorAboveOrigin.x - origin.x, y: floorAboveOrigin.y - origin.y};
      for(let p of entryExitPs){
        if(p.z == this.__floor.elevation * 1000){
          p.x *= scaleFactor;
          p.y *= scaleFactor;
        } else if(floorAbove && floorAbove.elevation * 1000 == p.z){
          p.x -= floorAboveOrigin.x;
          p.y -= floorAboveOrigin.y;
          p.x *= scaleFactor;
          p.y *= scaleFactor;
          p.x += floorAboveOrigin.x;
          p.y += floorAboveOrigin.y;
        } else {
          console.log(p, "can't get scaled due to missing floor above");
        }
      }
      this.setEntryExitPs(entryExitPs);
      if(this.isPurePolylineRepresentation()){
        let shapePs = this.shapePoints()
        for(let p of shapePs){
          p.x *= scaleFactor
          p.y *= scaleFactor
        }
        this.setPolylineRepresentation(shapePs)
      } else {
        let geo = this.getGeo()
        geo.x *= scaleFactor
        geo.y *= scaleFactor
        geo.width *= scaleFactor
        geo.height *= scaleFactor
        this.setRectangleRepresentation(geo.x, geo.y, geo.width, geo.height)
      }
  }

    IfcStair.prototype.adaptRepresentation = function(stairs) {
      for(const stair of stairs){
        if(stair.hasRepresentation()){
          stair.transGeoToFloor(this.__floor)
          let geo = stair.getGeo();
          stair.transGeoToFloor(stair.__floor, this.__floor)
          if(stair.isPurePolylineRepresentation()){
            this.setPolylineRepresentation(geo.points)
          } else {
            this.setRectangleRepresentation(geo.x, geo.y, geo.width, geo.height)
          }
          let ps = stair.getEntryExitPs(), nPs = []
          const zDiff = (this.__floor.elevation - stair.__floor.elevation) * 1000, thisOrigin = this.__floor.getOrigin(),
            stairOrigin = stair.__floor.getOrigin(), floorAboveThisFloorOrigin = this.__floor.getFloorAbove().getOrigin(),
            floorAboveStairFloorOrigin = stair.__floor.getFloorAbove().getOrigin()
          for(let p of ps){
            let nP = Object.assign({}, p)
            if(nP.id){
              delete nP.id
            }
            const moveP = (o1, o2) => {
              nP.x -= o2.x - o1.x
              nP.y -= o2.y - o1.y
            }
            if(nP.z > stair.__floor.elevation * 1000){
              moveP(floorAboveThisFloorOrigin, floorAboveStairFloorOrigin)
            } else {
              moveP(thisOrigin, stairOrigin)
            }
            nP.z += zDiff
            nPs.push(nP)
          }
          this.setEntryExitPs(nPs)
          break
        }
      }
      return this
    }

    IfcStair.prototype.moveEntryExitPs = function(diffX, diffY) {
      let entryExitPs = this.getEntryExitPs()
      for(let node of entryExitPs){
        node.x -= diffX
        node.y -= diffY
      }
      this.setEntryExitPs(entryExitPs)
    }

    IfcStair.prototype.remove = function() {
        this.__project.getRelContainedInSpatialStructure(this.__floor).removeRelatedElement(this);
        for(let iii=0;iii<this.__floor.__stairs.length;++iii){
          if(this.__floor.__stairs[iii].id === this.id){
            this.__floor.__stairs.splice(iii, 1);
            break;
          }
        }
        if (this.__representation) {
          this.__representation.remove(this.__project);
        }
        if (this.__objectPlacement) {
          this.__objectPlacement.remove(this.__project);
        }
        this.removeIfcElementAttributes()
        this.__project.unregister(this);
    }

    IfcStair.prototype.hasRepresentation = function() {
        return this.representation != null;
    }

    IfcStair.prototype.shapeTypeUndefined = function() {
        if(this.shapeType === IfcStairTypeEnum.NOTDEFINED || this.shapeType === undefined){
        return true;
        } else {
        return false;
        }
    }

    IfcStair.prototype.load = function (errors) {
        this.id = this.__attributes[0];
        this.name = this.__attributes[2];
        this.description = this.__attributes[3];
        this.objectPlacement = this.__attributes[5];
        this.representation = this.__attributes[6];
    };

    IfcStair.prototype.create = function (project, name) {

        this.type = 'IFCSTAIR';
        this.id = getUuid();

        this.__project = project;
        this.name = name;
        this.shapeType = IfcStairTypeEnum.NOTDEFINED;

        project.register(this);

        return this;
    };

    IfcStair.prototype.bind = function (project) {
        this.__project = project;
        if (this.representation) {
           this.__representation = project.map[this.representation];
        }
        if (this.objectPlacement) {
          this.__objectPlacement = this.__project.map[this.objectPlacement];
        }
    };

    IfcStair.prototype.getShapeCenter = function () {
        let center
        if(this.isPureRectangleRepresentation()){
          center = centerOfMass(this.shapePoints())
        } else if(this.isPurePolylineRepresentation()){
          // //TODO
          // let points = this.shapePoints()
          // center = centerOfMass(points.slice(0, points.length - 1)) //only for testing purposes
          center = visualCenter(this.shapePoints());
        } else {
          console.log('representation type unknwon')
        }
        return center
    };

    IfcStair.prototype.getDijNodes = function () {
        if(this.shapeTypeUndefined()){
            let nodes = this.getEntryExitPs()
            for(let node of nodes){
              Object.assign(node, {stairId: this.id})
            }
            return nodes
            // let floorAbove = this.__floor.getFloorAbove();
            // if(!floorAbove){
            //   return []
            // }
            // // const stairCenter = this.getShapeCenter();
            // // if(!floorAbove || !stairCenter){
            // //   return []
            // // }
            // // let nodes = [
            // //   {x: stairCenter.x, y: stairCenter.y, z: this.__floor.elevation * 1000},
            // //   {x: stairCenter.x, y: stairCenter.y, z: floorAbove.elevation * 1000}];
            // // for(let node of nodes){
            // //   Object.assign(node, {id: getUuid()});
            // // }
            // let nodes = [];
            // let g1 = this.transGeoToFloor(floorAbove), g0 = this.transGeoToFloor(this.__floor, floorAbove)
            // const setDefNodes = () => {
            //   let b0 = g0.getBoundingBox(),
            //       b1 = g1.getBoundingBox()
            //   nodes = [
            //     {x: b0.xCenter, y: b0.yCenter, z: this.__floor.elevation * 1000},
            //     {x: b1.xCenter, y: b1.yCenter, z: floorAbove.elevation * 1000}
            //   ];
            // };
            // if(this.isPureRectangleRepresentation()){
            //   if(g0.width < g0.height){
            //     nodes = [
            //       {x: g0.x + g0.width / 2, y: g0.y, z: this.__floor.elevation * 1000},
            //       {x: g1.x + g1.width / 2, y: g1.y + g1.height, z: floorAbove.elevation * 1000},
            //       {x: g1.x + g1.width / 2, y: g1.y, z: floorAbove.elevation * 1000},
            //       {x: g0.x + g0.width / 2, y: g0.y + g0.height, z: this.__floor.elevation * 1000}
            //     ];
            //   } else {
            //     nodes = [
            //       {x: g0.x, y: g0.y + g0.height / 2, z: this.__floor.elevation * 1000},
            //       {x: g1.x + g1.width, y: g1.y + g1.height / 2, z: floorAbove.elevation * 1000},
            //       {x: g1.x, y: g1.y + g1.height / 2, z: floorAbove.elevation * 1000},
            //       {x: g0.x + g0.width, y: g0.y + g0.height / 2, z: this.__floor.elevation * 1000}
            //     ];
            //   }
            // } else if(this.isPurePolylineRepresentation() && g0.points.length == 5){
            //   let horiDists = [];
            //   for(let iii=0;iii<g0.points.length - 1; ++iii){
            //     horiDists.push(horiDist(g0.points[iii], g0.points[iii + 1]));
            //   }
            //   const centerOfLine = (p1, p2) => {
            //     return {
            //       x: p1.x + (p2.x - p1.x)/2,
            //       y: p1.y + (p2.y - p1.y)/2
            //     }
            //   };
            //   if(horiDists[0] < horiDists[1] && horiDists[0] < horiDists[3] &&
            //     horiDists[2] < horiDists[1] && horiDists[2] < horiDists[3]){
            //       let g0centerp0p1 = centerOfLine(g0.points[0], g0.points[1]),
            //           g0centerp2p3 = centerOfLine(g0.points[2], g0.points[3]),
            //           g1centerp0p1 = centerOfLine(g1.points[0], g1.points[1]),
            //           g1centerp2p3 = centerOfLine(g1.points[2], g1.points[3])
            //       nodes = [
            //         {x: g0centerp0p1.x, y: g0centerp0p1.y, z: this.__floor.elevation * 1000},
            //         {x: g1centerp2p3.x, y: g1centerp2p3.y, z: floorAbove.elevation * 1000},
            //         {x: g1centerp0p1.x, y: g1centerp0p1.y, z: floorAbove.elevation * 1000},
            //         {x: g0centerp2p3.x, y: g0centerp2p3.y, z: this.__floor.elevation * 1000}
            //       ];
            //   } else if(horiDists[1] < horiDists[0] && horiDists[1] < horiDists[2] &&
            //     horiDists[3] < horiDists[0] && horiDists[3] < horiDists[2]){
            //       let g0centerp1p2 = centerOfLine(g0.points[1], g0.points[2]),
            //           g0centerp3p4 = centerOfLine(g0.points[3], g0.points[4]),
            //           g1centerp1p2 = centerOfLine(g1.points[1], g1.points[2]),
            //           g1centerp3p4 = centerOfLine(g1.points[3], g1.points[4])
            //     nodes = [
            //       {x: g0centerp1p2.x, y: g0centerp1p2.y, z: this.__floor.elevation * 1000},
            //       {x: g1centerp3p4.x, y: g1centerp3p4.y, z: floorAbove.elevation * 1000},
            //       {x: g1centerp1p2.x, y: g1centerp1p2.y, z: floorAbove.elevation * 1000},
            //       {x: g0centerp3p4.x, y: g0centerp3p4.y, z: this.__floor.elevation * 1000}
            //     ];
            //   } else {
            //     setDefNodes();
            //   }
            // } else {
            //   setDefNodes();
            // }
            // for(let node of nodes){
            //   Object.assign(node, {id: getUuid()});
            // }
            // for(let node of nodes){
            //   Object.assign(node, {stairId: this.id})
            // }
            // return nodes;
          } else {
            console.log('stair type ', this.shapeType, 'not implemented');
            return [];
          }
    }

    IfcStair.prototype.getIfcElementAttributes = function () {
        if (!this.__project.__ifcElementAttributes) {
          this.__project.__ifcElementAttributes = {};
        }
        if (!this.__project.__ifcElementAttributes[this.id]) {
          this.__project.__ifcElementAttributes[this.id] = {};
        }
        return this.__project.__ifcElementAttributes[this.id];
      }

      IfcStair.prototype.removeIfcElementAttributes = function () {
        if(this.__project.__ifcElementAttributes && this.__project.__ifcElementAttributes[this.id]){
          delete this.__project.__ifcElementAttributes[this.id]
        }
      }

      IfcStair.prototype.setEntryExitPs = function (nodes) {
        for(let node of nodes){
          if(!node.id){
            Object.assign(node, {id: getUuid()})
          }
        }
        this.getIfcElementAttributes().entryExitPs = nodes
      }

      IfcStair.prototype.getEntryExitPs = function () {
        let res = this.getIfcElementAttributes().entryExitPs;
        if(res){
          for(let node of res){
            if(!node.id){
              this.setEntryExitPs(res)
              break
            }
          }
          return res
        } else {
          return []
        }
      }

    IfcStair.prototype.getDijGrid = function () {
        if(this.shapeTypeUndefined()){
            let floorAbove = this.__floor.getFloorAbove(),
              center0 = this.getShapeCenter()
            if(!floorAbove || !center0){
              return [];
            }
            const setCenter = (center) => {
              center.z = (this.__floor.elevation + (floorAbove.elevation - this.__floor.elevation) / 2) * 1000
              Object.assign(center, {stairId: this.id})
            }
            setCenter(center0)
            this.transGeoToFloor(floorAbove)
            let center1 = this.getShapeCenter()
            this.transGeoToFloor(this.__floor, floorAbove)
            if(center1 && ( center0.x != center1.x || center0.y != center1.y || center0.z == center1.z )){
              setCenter(center1)
              return [center0, center1]
            } else {
              return [center0]
            }
          } else {
            console.log('stair type ', this.shapeType, 'not implemented');
            return [];
          }
    };

    IfcStair.prototype.renderFloorPlanSvg = function(svg, options, floor) {
      if (!this.__representation) {
        return;
      }
      const thisRef = this;
      const geo = this.transGeoToFloor(floor);
      const group = geo.renderSvg(svg).attr('id', this.id).addClass('ifcStair').attr('id', this.id).attr('stroke', 'green')
        .attr('stroke-width', 50).attr('fill', 'none');
      this.transGeoToFloor(this.__floor, floor);
      const boundingBox = geo.getBoundingBox();
      group.text(this.name).font({
        size: 400,
        anchor: 'middle'
      }).attr({
        stroke: 'none',
        fill: 'green',
        class: 'counter-rotate',
        x: boundingBox.x + 0.5 * boundingBox.width,
        y: boundingBox.y + 0.5 * boundingBox.height - 400
      }).on('click', function(){
        event.preventDefault();
        if(!(options && options.editor && options.editor && (options.editor.buttonActionStarted || options.editor.stair == thisRef))){
          event.stopPropagation();
          floorplanElemSelectChange.next({obj: thisRef, floor: null});
        }
      }).mouseover(function (event) {
        event.preventDefault();
        if(!(options && options.editor && options.editor && options.editor.buttonActionStarted)){
          this.style('cursor', 'hand');
        } else {
          this.style('cursor', 'crosshair');
        }
      }).mouseout(function (event) {
        event.preventDefault();
        this.style('cursor', 'default');
      });
    }

    IfcStair.prototype.transGeoToFloor = function(toFloor, fromFloor = this.__floor){
      const geo = this.getGeo();
      const toFloorOrigin = toFloor.getOrigin(), fromFloorOrigin = fromFloor.getOrigin();
      if(toFloorOrigin && fromFloorOrigin){
        if (geo && geo.points) {
          for (const p of geo.points) {
            p.x -= fromFloorOrigin.x - toFloorOrigin.x;
            p.y -= fromFloorOrigin.y - toFloorOrigin.y;
          }
          this.setPolylineRepresentation(geo.points);
        } else if (geo && geo.width) {
          geo.x -= fromFloorOrigin.x - toFloorOrigin.x;
          geo.y -= fromFloorOrigin.y - toFloorOrigin.y;
          this.setRectangleRepresentation(geo.x, geo.y, geo.width, geo.height);
        } else {
          // console.log('unknown stair representation; return unaltered geometry');
        }
      } else {
        // console.log('unknown floor origin; return unaltered geometry');
      }
      return geo;
    }

    IfcStair.prototype.centerMatch = function (otherStair, threshold = 500) {
        const thisB = this.getGeo().getBoundingBox(), otherStairB = otherStair.getGeo().getBoundingBox();
        return (horiDist({x: thisB.xCenter, y: thisB.yCenter}, {x: otherStairB.xCenter, y: otherStairB.yCenter}) <= threshold);
    };

    IfcStair.prototype.move = function (diffX, diffY) {
      let g = this.getGeo();
      if(this.isPurePolylineRepresentation()){
        let shapePoints = g.points;
        for(let p of shapePoints){
          p.x -= diffX;
          p.y -= diffY;
        }
        this.setPolylineRepresentation(shapePoints);
      } else if(this.isPureRectangleRepresentation()){
        this.setRectangleRepresentation(g.x - diffX, g.y - diffY, g.width, g.height)
      } else {
        console.log('unknown stair representation');
      }
  };

    IfcStair.prototype.setRectangleRepresentation = function (x, y, width, height) {
        if (!this.__representation) {
            this.__representation = new IfcProductDefinitionShape().create(this.getProject(), '');
            this.representation = this.__representation.oid;
          }
          this.__representation.setRectangleRepresentation(this.getProject(), x + 0.5 * width, y + 0.5 * height, width, height);
        }

    IfcStair.prototype.setPolylineRepresentation = function (points) {
        if (!this.__representation) {
            this.__representation = new IfcProductDefinitionShape().create(this.getProject(), '');
            this.representation = this.__representation.oid;
          }
          this.__representation.setPolylineRepresentation(this.getProject(), points);
        }

    IfcStair.prototype.getAdjacentSpaces = function () {
        let surrSpaces = [];
        for(const floor of this.__floor.__building.__floors){
          for(const space of floor.__spaces){
            const stairRefs = space.getAdjacentStairs();
             for(const stairRef of stairRefs){
              if(stairRef.id === this.id){
                surrSpaces.push(space);
              }
            }
          }
        }
        return surrSpaces;
    };

    IfcStair.prototype.intersectWith = function (p1, p2) {
      const relFloors = [this.__floor], floorAbove = this.__floor.getFloorAbove();
      if(floorAbove){
        relFloors.push(floorAbove)
      }
      const outerStairPoints = this.shapePoints(), isRect = this.isPureRectangleRepresentation();
      for(const floor of relFloors){
        if(floor.elevation === undefined || floor.elevation === null || p1.z != floor.elevation * 1000 || p2.z != floor.elevation * 1000){
          continue
        }
        let intersectCount = 0;
        for(let iii=0;iii<(isRect ? outerStairPoints.length : outerStairPoints.length - 1); ++iii){
          if(psIntersect(p1, p2, outerStairPoints[iii], outerStairPoints[isRect ? ((iii + 1) % 4) : (iii + 1)])){
            intersectCount++
          }
        }
        if(intersectCount > 0){
          return intersectCount
        }
      }
      return false;
    };

    IfcStair.prototype.getGeo = function () {
        if (!this.__representation) {
            return new GeoNullProfile();
          }
          const geo = this.__representation.getGeo();
          if (this.__objectPlacement) {
            geo.applyPlacement(this.__objectPlacement.getPlacement());
          }
          return geo;
    };

    IfcStair.prototype.shapePoints = function () {
        let g = this.getGeo();
        let shapePoints = [];
        if(this.isPurePolylineRepresentation()){
          shapePoints = g.points;
        } else if(this.isPureRectangleRepresentation()){
          shapePoints.push({x: g.x, y: g.y});
          shapePoints.push({x: g.x + g.width, y: g.y});
          shapePoints.push({x: g.x + g.width, y: g.y + g.height});
          shapePoints.push({x: g.x, y: g.y + g.height});
        } else {
          // console.log('unknown stair representation')
          return []
        }
        if(this.__floor.elevation !== undefined && this.__floor.elevation !== null){
          for(let point of shapePoints){
            Object.assign(point, {z: this.__floor.elevation * 1000})
          }
        }
        return shapePoints;
    };

    IfcStair.prototype.getShapeCenter = function () {
        let center
        if(this.isPureRectangleRepresentation()){
          center = centerOfMass(this.shapePoints())
        } else if(this.isPurePolylineRepresentation()){
          // //TODO
          // let points = this.shapePoints()
          // center = centerOfMass(points.slice(0, points.length - 1)) //only for testing purposes
          center = visualCenter(this.shapePoints());
        } else {
          console.log('representation type unknwon')
        }
        return center
    };

    IfcStair.prototype.isPurePolylineRepresentation = function () {
        return this.isPureXRepresentation('IFCARBITRARYCLOSEDPROFILEDEF');
    };

    IfcStair.prototype.isPureRectangleRepresentation = function () {
        return this.isPureXRepresentation('IFCRECTANGLEPROFILEDEF');
    };

    IfcStair.prototype.isPureXRepresentation = function (repType) {
        if(this.hasRepresentation()){
            if(this.__representation.__representations){
              for(const rep of this.__representation.__representations){
                if(rep.__items){
                  for(const item of rep.__items){
                    if(item.__sweptArea && item.__sweptArea.type !== repType){
                      return false;
                    }
                  }
                }
              }
            }
            return true;
          }
          return false;
    };

    function IfcStairFlight() {
    }

// #121473= IFCSTAIRFLIGHT('0XpHH0poP64RIbb_iYeoUQ',#41,'Cast-In-Place Stair:Stair:489566 Run 1',$,'Cast-In-Place Stair:Concrete_3XN:345430',#121124,#121471,'489925',10,9,0.615157480314899,0.918635170603675);

    IfcStairFlight.prototype.load = function (errors) {
        this.id = this.__attributes[0];
        this.name = this.__attributes[2];
        this.description = this.__attributes[3];
        //this.objectType = this.__attributes[4];
        //this.objectPlacement = this.__attributes[5];
        //this.representation = this.__attributes[6];
    };

    IfcStairFlight.prototype.bind = function (project) {
        //if (this.objectPlacement) {
        //    this.__objectPlacement = project.map[this.objectPlacement];
        //}
        //
        //if (this.representation) {
        //    this.__representation = project.map[this.representation];
        //}
        //
        //var name = this.name.split(':');
    };

    function IfcLocalPlacement() {
    }

// #201= IFCLOCALPLACEMENT(#143,#200);

    IfcLocalPlacement.prototype.create = function (project, x, y, rotationX, rotationY) {
      this.type = 'IFCLOCALPLACEMENT';
      this.id = getUuid();

      project.register(this);

      this.placementRelTo = null; // For now

      this.__relativePlacement = new IfcAxis2Placement3D().create(project, '');
      this.relativePlacement = this.__relativePlacement.oid;

      this.__relativePlacement.setLocation(project, x, y, 0);

      if (rotationX !== undefined && rotationY !== undefined) {
        this.__relativePlacement.setDirection(project, rotationX, rotationY);
      }

      return this;
    };

    IfcLocalPlacement.prototype.remove = function (project) {
        if (this.__relativePlacement) {
            this.__relativePlacement.remove(project);
        }

        project.unregister(this);
    };

    IfcLocalPlacement.prototype.load = function (errors) {
        this.placementRelTo = this.__attributes[0];

        if (this.__attributes[1]) {
            this.relativePlacement = this.__attributes[1];
        } else {
            errors.push({object: this, message: 'Mandatory attribute [ObjectPlacement] not set.'})
        }
    };

    IfcLocalPlacement.prototype.bind = function (project) {
        if (this.placementRelTo) {
            this.__placementRelTo = project.map[this.placementRelTo];
        }

        if (this.relativePlacement) {
            this.__relativePlacement = project.map[this.relativePlacement];
        }
    };

    IfcLocalPlacement.prototype.getPlacement = function (stack) {
        // Prevent infinite loop

        //if (!stack) {
        //    stack = {};
        //}
        //
        //if (stack[this.oid]) {
        //    return {translation: {x: 0, y: 0}, rotation: 0};
        //}
        //
        //stack[this.oid] = this;

        var placement;
        var rotation;
        var translation;

        if (this.__placementRelTo) {
            placement = this.__placementRelTo.getPlacement(stack);

            translation = placement.translation;
            rotation = placement.rotation;

            placement = this.__relativePlacement.getPlacement(stack);

            translation.x += placement.translation.x;
            translation.y += placement.translation.y;

            rotation += placement.rotation;
        } else {
            placement = this.__relativePlacement.getPlacement(stack);

            translation = placement.translation;
            rotation = placement.rotation;
        }

        return {
            translation: translation,
            rotation: rotation
        };
    };

    IfcLocalPlacement.prototype.placeSvgShape = function (shape, stack) {
        // Prevent infinite loop

        if (!stack) {
            stack = {};
        }

        if (stack[this.oid]) {
            return;
        }

        stack[this.oid] = this;

        var placement;

        if (this.__placementRelTo) {
            this.__placementRelTo.placeSvgShape(shape, stack);
        }

        this.__relativePlacement.placeSvgShape(shape);
    };

    function IfcAxis2Placement2D() {
    }

    IfcAxis2Placement2D.prototype.remove = function (project) {
        this.__location.remove(project);

        if (this.__direction) {
            this.__direction.remove(project);
        }

        project.unregister(this);
    };

    IfcAxis2Placement2D.prototype.load = function (errors) {
        this.location = this.__attributes[0];
        this.direction = this.__attributes[1];
    };

    IfcAxis2Placement2D.prototype.bind = function (project) {
        this.__location = project.map[this.location];

        if (this.direction) {
            this.__direction = project.map[this.direction];
        }
    };

    IfcAxis2Placement2D.prototype.getPlacement = function () {
        return {
            translation: this.__location.getCoordinates(),
            rotation: this.__direction ? this.__direction.getRotation() : 0
        };
    };

    IfcAxis2Placement2D.prototype.placeSvgShape = function (shape) {
        shape.dmove(this.__location.getCoordinates().x, this.__location.getCoordinates().y);

        if (this.__direction) {
            shape.transform({rotation: this.__direction.getRotation()});
        }
    };

    function IfcAxis2Placement3D() {
    }

// #200= IFCAXIS2PLACEMENT3D(#6,$,$);

    IfcAxis2Placement3D.prototype.create = function (project, name) {
        this.type = 'IFCAXIS2PLACEMENT3D';
        this.id = getUuid();
        this.name = name;

        project.register(this);

        return this;
    };

    IfcAxis2Placement3D.prototype.remove = function (project) {
        this.__location.remove(project);

        if (this.__direction) {
            this.__direction.remove(project);
        }

        project.unregister(this);
    };

    IfcAxis2Placement3D.prototype.load = function (errors) {
        this.location = this.__attributes[0];
        this.direction = this.__attributes[2];
    };

    IfcAxis2Placement3D.prototype.bind = function (project) {
        this.__location = project.map[this.location];

        if (this.direction) {
            this.__direction = project.map[this.direction];
        }
    };

    IfcAxis2Placement3D.prototype.getPlacement = function () {
        var coordinates = this.__location.getCoordinates();

        return {
            translation: {x: coordinates.x, y: coordinates.y},
            rotation: this.__direction ? this.__direction.getRotation() : 0
        };
    };

    IfcAxis2Placement3D.prototype.setLocation = function (project, x, y, z) {
      if(!this.__location){
        this.__location = new IfcCartesianPoint().create(project);
        this.location = this.__location.oid;
      }

      this.__location.coordinates = {
        x,
        y,
        z: 0
      };
    };

    IfcAxis2Placement3D.prototype.setDirection = function (project, x, y) {
        this.__direction = new IfcDirection().create(project);
        this.direction = this.__direction.oid;
        this.__direction.coordinates = {
            x: x,
            y: y
        };
    };

    IfcAxis2Placement3D.prototype.placeSvgShape = function (shape) {
        shape.dmove(this.__location.getCoordinates().x, this.__location.getCoordinates().y);

        if (this.__direction) {
            shape.transform({rotation: this.__direction.getRotation()});
        }
    };

    function IfcCartesianPoint() {
    }

// #140= IFCCARTESIANPOINT((0.,0.,8625.));

    IfcCartesianPoint.prototype.create = function (project) {
        this.type = 'IFCCARTESIANPOINT';

        project.register(this);

        return this;
    };

    IfcCartesianPoint.prototype.remove = function (project) {
        project.unregister(this);
    };

    IfcCartesianPoint.prototype.load = function (errors) {
        if (this.__attributes[0]) {
            this.coordinates = {
                x: Number(this.__attributes[0][0]),
                y: Number(this.__attributes[0][1]),
                z: Number(this.__attributes[0][2])
            };
        } else {
            errors.push('Coordinates for Cartesian Point are not set.');
        }
    };

    IfcCartesianPoint.prototype.bind = function (project) {
    };

    IfcCartesianPoint.prototype.getCoordinates = function () {
        return this.coordinates;
    };

    function IfcDirection() {
    }

// #23= IFCDIRECTION((1.,0.));

    IfcDirection.prototype.create = function (project) {
        this.type = 'IFCDIRECTION';

        project.register(this);

        return this;
    };

    IfcDirection.prototype.remove = function (project) {
        project.unregister(this);
    };

    IfcDirection.prototype.getExactRotation = function () {
      return Math.atan2(this.coordinates.y, this.coordinates.x) * 180 / Math.PI
    };

    IfcDirection.prototype.load = function (errors) {
        if (this.__attributes[0]) {
            this.coordinates = {
                x: Number(this.__attributes[0][0]),
                y: Number(this.__attributes[0][1])
            };
        } else {
            errors.push('No coordinates for direction provided.');
        }
    };

    IfcDirection.prototype.bind = function (project) {
    };

    IfcDirection.prototype.getRotation = function () {
        if (Math.round(this.coordinates.x) == 1 && Math.round(this.coordinates.y) == 0) {
            return 0;
        }
        else if (Math.round(this.coordinates.x) == 0 && Math.round(this.coordinates.y) == 1) {
            return 270;
        }
        else if (Math.round(this.coordinates.x) == -1 && Math.round(this.coordinates.y) == 0) {
            return 180;
        }
        else if (Math.round(this.coordinates.x) == 0 && Math.round(this.coordinates.y) == -1) {
            return 90;
        }

        return 0;
    };

// #248922= IFCRELAGGREGATES('1auk217kD2Vu2unv3uKuts',#41,$,$,#144,(#218,#248,#273,#298,#323,#348,#373,#398,#423,#448,#473,#498,#523,#548,#573,#598,#623,#648,#673,#698,#723,#748,#773,#798,#823,#848,#873,#898,#923,#948,#973,#998,#1023,#1048,#1075,#1100,#1125,#1150,#1306,#1331,#1356,#1381,#1406,#1431,#1456,#1481,#1506,#1531,#1556,#1581));

    function IfcRelAggregates() {
    }

    IfcRelAggregates.prototype.create = function (project, relatingObject, relatedObjects) {
        this.type = 'IFCRELAGGREGATES';
        this.id = getUuid();
        this.relatingObject = relatingObject.oid;
        this.relatedObjects = [];
        this.__project = project;
        this.__relatingObject = relatingObject;
        this.__relatedObjects = [];

        project.register(this);

        for (var n in relatedObjects) {
            this.relatedObjects.push(relatedObjects[n]);
            this.__relatedObjects.push(relatedObjects[n]);
        }

        return this;
    };

    IfcRelAggregates.prototype.remove = function () {
        this.__project.unregister(this);
    }

    IfcRelAggregates.prototype.load = function (errors) {
        this.id = this.__attributes[0];
        this.name = this.__attributes[2];
        this.description = this.__attributes[3];
        this.relatingObject = this.__attributes[4];
        this.relatedObjects = [];

        for (var n in this.__attributes[5]) {
            this.relatedObjects.push(this.__attributes[5][n]);
        }
    };

    IfcRelAggregates.prototype.bind = function (project) {
        this.__project = project;

        if (this.relatingObject) {
            this.__relatingObject = project.map[this.relatingObject];
        }

        if (this.relatedObjects) {
            this.__relatedObjects = [];

            for (var n in this.relatedObjects) {
                this.__relatedObjects.push(project.map[this.relatedObjects[n]]);
            }

            if (!this.__relatingObject) {
                console.log('Skipping relating object ' + this.id);
            }
            else if (this.__relatingObject.type == 'IFCPROJECT') {
                this.__relatingObject.__sites = [];
                this.__relatingObject.__building = [];

                for (var n in this.__relatedObjects) {
                    if (this.__relatedObjects[n].type == 'IFCSITE') {
                        this.__relatingObject.__sites.push(this.__relatedObjects[n]);
                    }
                    else if (this.__relatedObjects[n].type == 'IFCBUILDING') {
                        this.__relatingObject.__buildings.push(this.__relatedObjects[n]);
                    }

                    this.__relatedObjects[n].__project = this.__relatingObject;
                }
            }
            else if (this.__relatingObject.type == 'IFCSITE') {
                // TODO Needed?
                this.__relatingObject.__buildings = [];

                for (var n in this.__relatedObjects) {
                    this.__relatingObject.__buildings.push(this.__relatedObjects[n]);
                    this.__relatedObjects[n].__site = this.__relatingObject;
                }
            }
            else if (this.__relatingObject.type == 'IFCBUILDING') {
                // TODO Needed?
                this.__relatingObject.__floors = [];

                for (var n in this.__relatedObjects) {
                    this.__relatingObject.__floors.push(this.__relatedObjects[n]);
                    this.__relatedObjects[n].__building = this.__relatingObject;
                }
            }
            else if (this.__relatingObject.type == 'IFCBUILDINGSTOREY') {
                // TODO Needed?
                this.__relatingObject.__spaces = [];

                for (var n in this.__relatedObjects) {
                    if (this.__relatedObjects[n].type == 'IFCSPACE') {
                        this.__relatingObject.__spaces.push(this.__relatedObjects[n]);
                        this.__relatedObjects[n].__floor = this.__relatingObject;
                    }
                }
            }
        }
    };

    IfcRelAggregates.prototype.addRelatedObject = function (object) {
        if (this.relatedObjects.indexOf(object.oid) >= 0) {
            return;
        }

        this.relatedObjects.push(object.oid);
        this.__relatedObjects.push(object);
    };

    IfcRelAggregates.prototype.removeRelatedObject = function (object) {
        if (object.type == 'IFCSITE') {
            _.pull(this.__relatingObject.__sites, object);
        }
        else if (object.type == 'IFCBUILDING') {
            _.pull(this.__relatingObject.__buildings, object);
        }
        else if (object.type == 'IFCBUILDINGSTOREY') {
            _.pull(this.__relatingObject.__floors, object);
        }
        else if (object.type == 'IFCSPACE') {
            _.pull(this.__relatingObject.__spaces, object);
        }

        _.pull(this.relatedObjects, object.oid);
        _.pull(this.__relatedObjects, object);
    };

    IfcRelAggregates.prototype.dump = function (stream) {
        stream.write('#' + this.oid);
        stream.write('= ');
        stream.write('IFCRELAGGREGATES(');
        dumpStringAttribute(stream, this.id, true);
        dumpOidAttribute(stream, this.relatingObject);
        dumpOidListAttribute(stream, this.relatedObjects);

        stream.write(');\n');
    };

    function IfcRelReferencedInSpatialStructure() {
    }

    IfcRelReferencedInSpatialStructure.prototype.create = function (project, relatingStructure, relatedElements) {
        this.type = 'IFCRELREFERENCEDINSPATIALSTRUCTURE';
        this.id = getUuid();
        this.relatingStructure = relatingStructure.oid;
        this.relatedElements = [];
        this.__project = project;
        this.__relatingStructure = relatingStructure;
        this.__relatedElements = [];

        project.register(this);

        for (const n in relatedElements) {
          this.relatedElements.push(relatedElements[n].oid);  //oid??? -> is not done in the relContained analogon -> mistake!?
          this.__relatedElements.push(relatedElements[n]);
        }

        return this;
    }

    IfcRelReferencedInSpatialStructure.prototype.remove = function () {
        this.__project.unregister(this);
    }

    IfcRelReferencedInSpatialStructure.prototype.load = function (errors) {
        this.id = this.__attributes[0];
        this.name = this.__attributes[2];
        this.description = this.__attributes[3];
        this.relatedElements = [];

        for (const n in this.__attributes[4]) {
          this.relatedElements.push(this.__attributes[4][n]);
        }
        this.relatingStructure = this.__attributes[5];
    }

    IfcRelReferencedInSpatialStructure.prototype.bind = function (project) {
      this.__project = project;

      if (this.relatingStructure) {
        this.__relatingStructure = project.map[this.relatingStructure];
      }
      if (this.relatedElements) {
        this.__relatedElements = [];

        for (const relatedElement of this.relatedElements) {
          if (!project.map[relatedElement]) {
            console.error('Cannot bind related element ' + relatedElement);
            continue;
          }
          this.__relatedElements.push(project.map[relatedElement]);
        }
        if (!this.__relatingStructure) {
          console.log(' Skipping relating structure ' + this.id);
        } else if (this.__relatingStructure.type === 'IFCBUILDINGSTOREY') {
          for (const n in this.__relatedElements) {
            if(!this.__relatedElements[n].__refFloors){
              this.__relatedElements[n].__refFloors = [];
            }
            this.__relatedElements[n].__refFloors.push(this.__relatingStructure);
            if (this.__relatedElements[n].type === 'IFCTRANSPORTELEMENT') {
              this.__relatingStructure.__refTransportElements.push(this.__relatedElements[n]);
            }
          }
        } else if (this.__relatingStructure.type === 'IFCSPACE') {
          for (const n in this.__relatedElements) {
            if (this.__relatedElements[n].type === 'IFCDOOR') {
              this.__relatingStructure.__doors.push(this.__relatedElements[n]);
              this.__relatedElements[n].__spaces.push(this.__relatingStructure);
            }
          }
        } else if (this.__relatingStructure.type === 'IFCTRANSPORTELEMENT' && this.__relatingStructure.isElevator()) {
          for (const n in this.__relatedElements) {
            this.__relatedElements[n].__elevator = this.__relatingStructure;
            if (this.__relatedElements[n].type === 'IFCDOOR') {
              this.__relatingStructure.__doors.push(this.__relatedElements[n]);
            }
          }
        }
      }
    }

    IfcRelReferencedInSpatialStructure.prototype.addRelatedElement = function (element) {
      if (this.relatedElements.indexOf(element.oid) >= 0) {
        return;
      }
      this.relatedElements.push(element.oid);
      this.__relatedElements.push(element);
      if (this.__relatingStructure.type === 'IFCBUILDINGSTOREY') {
        if (element.type === 'IFCTRANSPORTELEMENT') {
          if (!this.__relatingStructure.__refTransportElements) {
            this.__relatingStructure.__refTransportElements = [];
          }
          this.__relatingStructure.__refTransportElements.push(element);
        }
      } else if (this.__relatingStructure.type === 'IFCSPACE') {
        if (!this.__relatingStructure.__doors) {
          this.__relatingStructure.__doors = [];
        }
        if (element.type === 'IFCDOOR') {
          this.__relatingStructure.__doors.push(element);
          if (!element.__spaces) {
            element.__spaces = [];
          }
          element.__spaces.push(this.__relatingStructure);
        }
      } else if (this.__relatingStructure.type === 'IFCTRANSPORTELEMENT' && this.__relatingStructure.isElevator()) {
        element.__elevator = this.__relatingStructure;
        if (!this.__relatingStructure.__doors) {
          this.__relatingStructure.__doors = [];
        }
        if (element.type === 'IFCDOOR') {
          this.__relatingStructure.__doors.push(element);
        }
      }
    }

    IfcRelReferencedInSpatialStructure.prototype.removeRelatedElement = function (element) {
      if (element.type === 'IFCDOOR' && this.__relatingStructure.__doors) {
        _.pull(this.__relatingStructure.__doors, element);
      }

        _.pull(this.relatedElements, element.oid);
        _.pull(this.__relatedElements, element);
    }

    IfcRelReferencedInSpatialStructure.prototype.dump = function (stream) {
        stream.write('#' + this.oid);
        stream.write('= ');
        stream.write('IFCRELREFERENCEDINSPATIALSTRUCTURE(');
        dumpStringAttribute(stream, this.id, true);
        dumpOidAttribute(stream, this.relatingStructure, false);
        dumpOidListAttribute(stream, this.relatedElements, false);

        stream.write(');\n');
    }

    function IfcRelContainedInSpatialStructure() {
    }

    IfcRelContainedInSpatialStructure.prototype.create = function (project, relatingStructure, relatedElements) {
        this.type = 'IFCRELCONTAINEDINSPATIALSTRUCTURE';
        this.id = getUuid();
        this.relatingStructure = relatingStructure.oid;
        this.relatedElements = [];
        this.__project = project;
        this.__relatingStructure = relatingStructure;
        this.__relatedElements = [];

        project.register(this);

        for (var n in relatedElements) {
            this.relatedElements.push(relatedElements[n]);
            this.__relatedElements.push(relatedElements[n]);
        }

        return this;
    };

    IfcRelContainedInSpatialStructure.prototype.remove = function () {
        for (var n in this.__relatedElements) {
            if (this.__relatingStructure.type == 'IFCSPACE') {
                _.pull(this.__relatedElements[n].__spaces, this.__relatingStructure);
            }
        }

        this.__project.unregister(this);
    };

    IfcRelContainedInSpatialStructure.prototype.load = function (errors) {
        this.id = this.__attributes[0];
        this.name = this.__attributes[2];
        this.description = this.__attributes[3];
        this.relatedElements = [];

        for (var n in this.__attributes[4]) {
            this.relatedElements.push(this.__attributes[4][n]);
        }
        this.relatingStructure = this.__attributes[5];
    };

    IfcRelContainedInSpatialStructure.prototype.bind = function (project) {
      this.__project = project;

      if (this.relatingStructure) {
        this.__relatingStructure = project.map[this.relatingStructure];
      }

      if (this.relatedElements) {
        this.__relatedElements = [];

        for (const relatedElement of this.relatedElements) {
          if (!project.map[relatedElement]) {
            console.error('Cannot bind related element ' + relatedElement);

            continue;
          }

          this.__relatedElements.push(project.map[relatedElement]);
        }

        if (!this.__relatingStructure) {
          console.log('Skipping relating structure ' + this.id);
        } else if (this.__relatingStructure.type === 'IFCBUILDINGSTOREY') {
          // tslint:disable-next-line:forin

          for (const n in this.__relatedElements) {
            this.__relatedElements[n].__floor = this.__relatingStructure;

            if (this.__relatedElements[n].type === 'IFCDOOR') {
              this.__relatingStructure.__doors.push(this.__relatedElements[n]);
            } else if (this.__relatedElements[n].type === 'IFCWINDOW') {
              this.__relatingStructure.__windows.push(this.__relatedElements[n]);
            } else if (this.__relatedElements[n].type === 'IFCWALL' || this.__relatedElements[n].type === 'IFCWALLSTANDARDCASE') {
              this.__relatingStructure.__walls.push(this.__relatedElements[n]);
            } else if (this.__relatedElements[n].type === 'IFCSTAIR') {
              this.__relatingStructure.__stairs.push(this.__relatedElements[n]);
            } else if (this.__relatedElements[n].type === 'IFCRAMP') {
              this.__relatingStructure.__ramps.push(this.__relatedElements[n]);
            } else if (this.__relatedElements[n].type === 'IFCTRANSPORTELEMENT') {
              this.__relatingStructure.__transportElements.push(this.__relatedElements[n]);
            }
          }
        } else if (this.__relatingStructure.type === 'IFCSPACE') {
          // tslint:disable-next-line:forin
          for (const n in this.__relatedElements) {
            this.__relatedElements[n].__space = this.__relatingStructure;

            // if (this.__relatedElements[n].type === 'IFCDOOR') {
            //   this.__relatingStructure.__doors.push(this.__relatedElements[n]);
            //   this.__relatedElements[n].__spaces.push(this.__relatingStructure);
            // } else
            if (this.__relatedElements[n].type === 'IFCWINDOW') {
              this.__relatingStructure.__windows.push(this.__relatedElements[n]);
            }
            // else if (this.__relatedElements[n].type === 'IFCWALL' || this.__relatedElements[n].type == 'IFCWALLSTANDARDCASE') {
            //   this.__relatingStructure.__walls.push(this.__relatedElements[n]);
            // }
          }
        }
        // else if (this.__relatingStructure.type === 'IFCTRANSPORTELEMENT' && this.__relatingStructure.isElevator()) {
        //   // tslint:disable-next-line:forin
        //   for (const n in this.__relatedElements) {
        //     this.__relatedElements[n].__elevator = this.__relatingStructure;
        //     if (this.__relatedElements[n].type === 'IFCDOOR') {
        //       this.__relatingStructure.__doors.push(this.__relatedElements[n]);
        //     }
        //   }
        // }
      }
    };

    IfcRelContainedInSpatialStructure.prototype.addRelatedElement = function (element) {
      if (this.relatedElements.indexOf(element.oid) >= 0) {
        return;
      }

      this.relatedElements.push(element.oid);
      this.__relatedElements.push(element);

      if (this.__relatingStructure.type === 'IFCBUILDINGSTOREY') {
        element.__floor = this.__relatingStructure;

        if (!this.__relatingStructure.__doors) {
          this.__relatingStructure.__doors = [];
        }

        if (!this.__relatingStructure.__windows) {
          this.__relatingStructure.__windows = [];
        }

        if (!this.__relatingStructure.__walls) {
          this.__relatingStructure.__walls = [];
        }

        if (element.type === 'IFCTRANSPORTELEMENT') {
          if (!this.__relatingStructure.__transportElements) {
            this.__relatingStructure.__transportElements = [];
          }
          this.__relatingStructure.__transportElements.push(element);
        }

        if (element.type === 'IFCDOOR') {
          this.__relatingStructure.__doors.push(element);
        }

        if (element.type === 'IFCWALL') {
          this.__relatingStructure.__walls.push(element);
        }

        if (element.type === 'IFCSTAIR') {
          if (!this.__relatingStructure.__stairs) {
            this.__relatingStructure.__stairs = [];
          }
          this.__relatingStructure.__stairs.push(element);
        }

        if (element.type === 'IFCRAMP') {
          if (!this.__relatingStructure.__stairs) {
            this.__relatingStructure.__ramps = [];
          }
          this.__relatingStructure.__ramps.push(element);
        }
      } else if (this.__relatingStructure.type === 'IFCSPACE') {
        element.__space = this.__relatingStructure;

        if (!this.__relatingStructure.__doors) {
          this.__relatingStructure.__doors = [];
        }

        if (!this.__relatingStructure.__windows) {
          this.__relatingStructure.__windows = [];
        }

        // if (!this.__relatingStructure.__walls) {
        //   this.__relatingStructure.__walls = [];
        // }

        // if (element.type === 'IFCDOOR') {

        //   this.__relatingStructure.__doors.push(element);

        //   // TODO Get rid of scattered initializations and initialize on create/load/bind

        //   if (!element.__spaces) {
        //     element.__spaces = [];
        //   }

        //   element.__spaces.push(this.__relatingStructure);
        // }
      }
      // else if (this.__relatingStructure.type === 'IFCTRANSPORTELEMENT' && this.__relatingStructure.isElevator()) {
      //   element.__elevator = this.__relatingStructure;
      //   if (!this.__relatingStructure.__doors) {
      //     this.__relatingStructure.__doors = [];
      //   }
      //   if (element.type === 'IFCDOOR') {
      //     this.__relatingStructure.__doors.push(element);
      //   }
      // }
    };

    IfcRelContainedInSpatialStructure.prototype.removeRelatedElement = function (element) {
        if (element.type == 'IFCDOOR' && this.__relatingStructure.__doors) {
            _.pull(this.__relatingStructure.__doors, element);
        }

        _.pull(this.relatedElements, element.oid);
        _.pull(this.__relatedElements, element);
    };

    IfcRelContainedInSpatialStructure.prototype.dump = function (stream) {
        stream.write('#' + this.oid);
        stream.write('= ');
        stream.write('IFCRELCONTAINEDINSPATIALSTRUCTURE(');
        dumpStringAttribute(stream, this.id, true);
        dumpOidAttribute(stream, this.relatingStructure);
        dumpOidListAttribute(stream, this.relatedElements);

        stream.write(');\n');
    };

    function IfcRelDefinesByType() {
    }

    // #253914= IFCRELDEFINESBYTYPE('3CvHW8tl9DXOVoYf62ZEmO',#41,$,$,(#47149,#47221,#47466,#47564,#48887,#73140,#73196,#91071,#91127,#91160,#91216,#92120,#93392,#93448,#93986,#94084,#94224,#94280,#95375,#97928,#98026,#98289,#104828,#104861,#104894,#104927,#105819,#106007,#112002,#112100,#112159,#129641,#129674,#129707,#129740,#129773,#129955,#129988,#130021,#130054,#130087,#130120,#130153,#130186,#130402,#130435,#130468,#130501,#130534,#153275,#153308,#153341,#154484,#154540,#154785,#154883,#156193,#157729,#157785,#158604,#158660,#158905,#159003,#160313,#161849,#161905,#162724,#162780,#163025,#163123,#164433,#165969,#166025,#166844,#166900,#167145,#167243,#168553,#170089,#170145,#170964,#171020,#171265,#171363,#172673,#174209,#174265,#175084,#175140,#175385,#175483,#176793,#178329,#178385,#179204,#179260,#179505,#179603,#180913,#182449,#182505,#183288,#183344,#183589,#183687,#184997,#186533,#186589),#47190);

    IfcRelDefinesByType.prototype.load = function (errors) {
        this.id = this.__attributes[0];
        this.name = this.__attributes[2];
        this.description = this.__attributes[3];
        this.relatedObjects = [];

        for (var n in this.__attributes[4]) {
            this.relatedObjects.push(this.__attributes[4][n]);
        }

        this.relatingType = this.__attributes[5];
    };

    IfcRelDefinesByType.prototype.bind = function (project) {
      if (this.relatingType) {
        this.__relatingType = project.map[this.relatingType];
      }

      if (this.relatedObjects) {
        this.__relatedObjects = [];

        for (const relatedObject of this.relatedObjects) {
          this.__relatedObjects.push(project.map[relatedObject]);
        }

        if (!this.__relatingType) {
          console.log('Skipping relating structure ' + this.id);
        } else if (this.__relatingType.type === 'IFCWALLTYPE') {
          for (const n in this.__relatedObjects) {
            if (this.__relatedObjects[n].type === 'IFCWALL' || this.__relatedObjects[n].type === 'IFCWALLSTANDARDCASE') {
              this.__relatedObjects[n].__wallType = this.__relatingType;
            }
          }
        }
      }
    };

    function IfcRelDefinesByProperties() {
    }

    // #112007= IFCRELDEFINESBYPROPERTIES('3lIKBlLEr8cxHhzsFcxSwu',#41,$,$,(#112002),#112005);

    IfcRelDefinesByProperties.prototype.create = function (project, propertySet, element) {
        this.type = 'IFCRELDEFINESBYPROPERTIES';
        this.id = getUuid();
        this.relatedObjects = [];
        this.__relatedObjects = [];
        this.__project = project;

        project.register(this);

        this.relatingPropertyDefinition = propertySet.oid;
        this.__relatingPropertyDefinition = propertySet;

        if (element) {
            this.relatedObjects.push(element.oid);
            this.__relatedObjects.push(element);
            element.__propertySets.push(propertySet);
        }

        return this;
    };

    IfcRelDefinesByProperties.prototype.remove = function () {
        for (var n in this.__relatedObjects) {
            _.pull(this.__relatedObjects[n].__propertySets, this.__relatingPropertyDefinition);
        }

        this.__project.unregister(this);
    };

    IfcRelDefinesByProperties.prototype.load = function (errors) {
        this.id = this.__attributes[0];
        this.name = this.__attributes[2];
        this.description = this.__attributes[3];
        this.relatedObjects = [];

        for (var n in this.__attributes[4]) {
            this.relatedObjects.push(this.__attributes[4][n]);
        }
        this.relatingPropertyDefinition = this.__attributes[5];
    };

    IfcRelDefinesByProperties.prototype.bind = function (project) {
        this.__project = project;

        if (!this.relatingPropertyDefinition) {
            return;
        }

        this.__relatingPropertyDefinition = project.map[this.relatingPropertyDefinition];
        this.__relatedObjects = [];

        for (var n in this.relatedObjects) {
            if (!project.map[this.relatedObjects[n]]) {
                console.error('No entry for related object ' + this.relatedObjects[n]);

                continue;
            }

            this.__relatedObjects.push(project.map[this.relatedObjects[n]]);
        }

        if (!this.__relatingPropertyDefinition) {
            console.log('Skipping defines by properties ' + this.__relatingPropertyDefinition);
        }
        else if (this.__relatingPropertyDefinition.type == 'IFCPROPERTYSET') {
            for (var n in this.__relatedObjects) {
                if (!this.__relatedObjects[n].__propertySets) {
                    this.__relatedObjects[n].__propertySets = [];
                }

                this.__relatedObjects[n].__propertySets.push(this.__relatingPropertyDefinition);

                // for (var m in this.__relatingPropertyDefinition.__propertyValues) {
                //     //console.log(this.__relatingPropertyDefinition.__propertyValues[m].id + ': ' + this.__relatingPropertyDefinition.__propertyValues[m].value);
                //
                //     this.__relatedObjects[n].__properties[this.__relatingPropertyDefinition.__propertyValues[m].id] = this.__relatingPropertyDefinition.__propertyValues[m].value;
                // }
            }
        }
    };

    function IfcPropertySet() {
    }

    // #6201= IFCPROPERTYSET('3uHBCVHL9AVey1wIaWPv85',#41,'Pset_SlabCommon',$,(#6198,#6199,#6200));

    IfcPropertySet.prototype.create = function (project, name) {
        this.type = 'IFCPROPERTYSET';
        this.id = getUuid();
        this.name = name;
        this.propertyValues = [];
        this.__project = project;
        this.__propertyValues = [];

        project.register(this);

        return this;
    };

    IfcPropertySet.prototype.load = function (errors) {
        this.id = this.__attributes[0];
        this.name = this.__attributes[2];
        this.description = this.__attributes[3];
        this.propertyValues = [];

        for (var n in this.__attributes[4]) {
            this.propertyValues.push(this.__attributes[4][n]);
        }
    };

    IfcPropertySet.prototype.bind = function (project) {
        this.__project = project;
        this.__propertyValues = [];

        for (var n in this.propertyValues) {
            this.__propertyValues.push(project.map[this.propertyValues[n]]);
        }
    };

    IfcPropertySet.prototype.remove = function () {
        this.__project.__relDefinesByProperties[this.oid].remove(this.__project);
        this.__project.unregister(this);
    };


    IfcPropertySet.prototype.addProperty = function (id, value) {
        var property = new IfcPropertySingleValue().create(this.__project, id, value);

        this.propertyValues.push(property.oid);
        this.__propertyValues.push(property);

        return property;
    };

    IfcPropertySet.prototype.deleteProperty = function (property) {
        this.__project.unregister(property);
        _.pull(this.propertyValues, property.id);
        _.pull(this.__propertyValues, property);
    };

    function IfcPropertySingleValue() {
    }

// #6199= IFCPROPERTYSINGLEVALUE('IsExternal',$,IFCBOOLEAN(.F.),$);

    function IfcPropertySingleValue() {
    }

    IfcPropertySingleValue.prototype.create = function (project, name, value) {
        this.type = 'IFCPROPERTYSINGLEVALUE';
        this.id = getUuid();
        this.name = name;
        this.value = value;

        project.register(this);

        return this;
    };

    IfcPropertySingleValue.prototype.load = function (errors) {
        this.name = this.__attributes[0];
        this.value = this.__attributes[2][0]; // TODO Hack! Parser Bug
    };

    IfcPropertySingleValue.prototype.bind = function (project) {
    };

    function IfcProductDefinitionShape() {
    }

// #214= IFCPRODUCTDEFINITIONSHAPE($,$,(#211));

    IfcProductDefinitionShape.prototype.create = function (project, name) {
        this.type = 'IFCPRODUCTDEFINITIONSHAPE';
        this.id = getUuid();
        this.name = name;

        project.register(this);

        return this;
    };

    IfcProductDefinitionShape.prototype.remove = function (project) {
        for (var n in this.__representations) {
            this.__representations[n].remove(project);
        }

        project.unregister(this);
    };

    IfcProductDefinitionShape.prototype.load = function (errors) {
        this.name = this.__attributes[0];
        this.description = this.__attributes[1];
        this.representations = [];

        if (this.__attributes[2]) {
            for (var n in this.__attributes[2]) {
                this.representations.push(this.__attributes[2][n]);
            }
        } else {
            // TODO Error
        }
    };

    IfcProductDefinitionShape.prototype.bind = function (project) {
        this.__representations = [];

        if (this.representations) {
            for (var n in this.representations) {
                this.__representations.push(project.map[this.representations[n]]);
            }
        }
    };

    IfcProductDefinitionShape.prototype.getGeo = function () {
        // TODO

        if (this.__representations[0]) {
            return this.__representations[0].getGeo();
        } else {
            return new GeoNullProfile();
        }
    };

    IfcProductDefinitionShape.prototype.setRectangleRepresentation = function (project, x, y, width, height) {
        if (!this.__representations || this.__representations.length == 0) {
            this.__representations = [new IfcShapeRepresentation().create(project)];
            this.representations = [this.__representations[0].oid];
        }

        this.__representations[0].setRectangleRepresentation(project, x, y, width, height);
    };

    IfcProductDefinitionShape.prototype.setPolylineRepresentation = function (project, points) {
        if (!this.__representations || this.__representations.length == 0) {
            this.__representations = [new IfcShapeRepresentation().create(project)];
            this.representations = [this.__representations[0].oid];
        }

        this.__representations[0].setPolylineRepresentation(project, points);
    };

    // #211= IFCSHAPEREPRESENTATION(#98,'Body','SweptSolid',(#210));

    function IfcShapeRepresentation() {
    }

    IfcShapeRepresentation.prototype.create = function (project, name) {
        this.type = 'IFCSHAPEREPRESENTATION';
        this.id = getUuid();
        this.name = name;

        project.register(this);

        return this;
    };

    IfcShapeRepresentation.prototype.remove = function (project) {
        for (var n in this.__items) {
          if(this.__items[n]){
            this.__items[n].remove(project);
          }
        }

        project.unregister(this);
    };

    IfcShapeRepresentation.prototype.load = function (errors) {
        this.name = this.__attributes[0];
        this.description = this.__attributes[1];
        this.items = [];

        if (this.__attributes[3]) {
            for (var n in this.__attributes[3]) {
                this.items.push(this.__attributes[3][n]);

            }
        } else {
            // TODO Error
        }
    };

    IfcShapeRepresentation.prototype.bind = function (project) {
        this.__items = [];

        for (var n in this.items) {
            this.__items.push(project.map[this.items[n]]);
        }
    };

    IfcShapeRepresentation.prototype.getGeo = function () {
        // TODO

        if (this.__items[0] && this.__items[0].getGeo) {
            return this.__items[0].getGeo();
        } else {
            return new GeoNullProfile();
        }
    };

    IfcShapeRepresentation.prototype.setRectangleRepresentation = function (project, x, y, width, height) {
      if (!this.__items || this.__items.length === 0) {
        this.__items = [new IfcExtrudedAreaSolid().create(project, '')];
        this.items = [this.__items[0].oid];
      }

      this.__items[0].setRectangleRepresentation(project, x, y, width, height);
    };

    IfcShapeRepresentation.prototype.setPolylineRepresentation = function (project, points) {
      if (!this.__items || this.__items.length === 0) {
        this.__items = [new IfcExtrudedAreaSolid().create(project, '')];
        this.items = [this.__items[0].oid];
      }

      this.__items[0].setPolylineRepresentation(project, points);
    };

    function IfcExtrudedAreaSolid() {
    }

// #210= IFCEXTRUDEDAREASOLID(#206,#209,#19,3250.);

    IfcExtrudedAreaSolid.prototype.create = function (project, name) {
        this.type = 'IFCEXTRUDEDAREASOLID';
        this.id = getUuid();
        this.name = name;

        project.register(this);

        return this;
    };

    IfcExtrudedAreaSolid.prototype.remove = function (project) {
        if (this.__sweptArea) {
            this.__sweptArea.remove(project);
        }

        if (this.__position) {
            this.__position.remove(project);
        }

        project.unregister(this);
    };

    IfcExtrudedAreaSolid.prototype.load = function () {
        if (this.__attributes[0]) {
            this.sweptArea = this.__attributes[0];
        } else {
            // TODO Error
        }

        if (this.__attributes[1]) {
            this.position = this.__attributes[1];
        } else {
            // TODO Error
        }

        //this.extrudedDirection

        this.depth = this.__attributes[3];
    };

    IfcExtrudedAreaSolid.prototype.bind = function (project) {
        if (this.sweptArea) {
            this.__sweptArea = project.map[this.sweptArea];
        }

        if (this.position) {
            this.__position = project.map[this.position];
        }
    };

    IfcExtrudedAreaSolid.prototype.getGeo = function () {
        var geo = this.__sweptArea.getGeo();

        if (this.__position) {
            geo.applyPlacement(this.__position.getPlacement());
        }

        return geo;
    };

    IfcExtrudedAreaSolid.prototype.setRectangleRepresentation = function (project, x, y, width, height) {
      if(this.__sweptArea && this.__sweptArea.type == 'IFCARBITRARYCLOSEDPROFILEDEF'){
        this.__sweptArea.remove(project);
        this.__sweptArea = null
      }
      if (!this.__sweptArea) {
        this.__sweptArea = new IfcRectangleProfileDef().create(project, '');
        this.sweptArea = this.__sweptArea.oid;
      }

      this.__sweptArea.setRectangleRepresentation(project, x, y, width, height);
    };

    IfcExtrudedAreaSolid.prototype.setPolylineRepresentation = function (project, points) {
      if(this.__sweptArea && this.__sweptArea.type == 'IFCRECTANGLEPROFILEDEF'){
        this.__sweptArea.remove(project);
        this.__sweptArea = null
      }
      if (!this.__sweptArea) {
        this.__sweptArea = new IfcArbitraryClosedProfileDef().create(project, '');
        this.sweptArea = this.__sweptArea.oid;
      }

      this.__sweptArea.setOuterCurve(project, points);
    };

    function IfcRectangleProfileDef() {
    }

// #206= IFCRECTANGLEPROFILEDEF(.AREA.,$,#205,2160.,2405.);

    IfcRectangleProfileDef.prototype.create = function (project, name) {
        this.type = 'IFCRECTANGLEPROFILEDEF';
        this.id = getUuid();
        this.name = name;

        project.register(this);

        return this;
    };

    IfcRectangleProfileDef.prototype.remove = function (project) {
        if (this.__position) {
            this.__position.remove(project);
        }

        project.unregister(this);
    };

    IfcRectangleProfileDef.prototype.load = function (errors) {
        this.profileType = this.__attributes[0];
        this.profileName = this.__attributes[1];

        if (this.__attributes[2]) {
            this.position = this.__attributes[2];
        } else {
            // TODO Error
        }

        this.xDim = this.__attributes[3];
        this.yDim = this.__attributes[4];
    };

    IfcRectangleProfileDef.prototype.bind = function (project) {
        if (this.position) {
            this.__position = project.map[this.position];
        }
    };

    IfcRectangleProfileDef.prototype.getGeo = function () {
        var geo = new GeoRectangleProfile().initialize(this.xDim, this.yDim, -0.5 * this.xDim, -0.5 * this.yDim);

        if (this.__position) {
            geo.applyPlacement(this.__position.getPlacement());
        }

        return geo;
    };

    IfcRectangleProfileDef.prototype.setRectangleRepresentation = function (project, x, y, width, height) {
        this.xDim = width;
        this.yDim = height;

        if (!this.__position) {
            this.__position = new IfcAxis2Placement3D().create(project);
            this.position = this.__position.oid;
        }

        this.__position.setLocation(project, x, y);
    };

    function IfcArbitraryClosedProfileDef() {
    }

//#1760= IFCARBITRARYCLOSEDPROFILEDEF(.AREA.,$,#1758);

    IfcArbitraryClosedProfileDef.prototype.create = function (project, name) {
        this.type = 'IFCARBITRARYCLOSEDPROFILEDEF';
        this.id = getUuid();
        this.name = name;

        project.register(this);

        return this;
    };

    IfcArbitraryClosedProfileDef.prototype.remove = function (project) {
        if (this.__outerCurve) {
            this.__outerCurve.remove(project);
        }

        project.unregister(this);
    };

    IfcArbitraryClosedProfileDef.prototype.load = function (errors) {
        this.profileType = this.__attributes[0];
        this.profileName = this.__attributes[1];

        if (this.__attributes[2]) {
            this.outerCurve = this.__attributes[2];
        } else {
            // TODO Error
        }
    };

    IfcArbitraryClosedProfileDef.prototype.bind = function (project) {
        if (this.outerCurve) {
            this.__outerCurve = project.map[this.outerCurve];
        }
    };

    IfcArbitraryClosedProfileDef.prototype.getGeo = function () {
        var geo = this.__outerCurve.getGeo();

        return geo;
    };

    IfcArbitraryClosedProfileDef.prototype.setOuterCurve = function (project, points) {
      // const offset = points[0];
      //
      // for (const n = 0; n < points.length) {
      //     points[n].x = points[n].x - points[0].x;
      //     points[n].y = points[n].y - points[0].y;
      // }

      if(this.__outerCurve){
        this.__outerCurve.remove(project)
      }

      this.__outerCurve = new IfcPolyline().create(project, points);
      this.outerCurve = this.__outerCurve.oid;

      // if (!this.__position) {
      //     this.__position = new IfcAxis2Placement3D().create(project);
      //     this.position = this.__position.oid;
      // }
      //
      // this.__position.setLocation(project, offset.x, offset.y);
    };

    function IfcArbitraryProfileDefWithVoids() {
    }

//#1760= IFCARBITRARYCLOSEDPROFILEDEF(.AREA.,$,#1758);

    IfcArbitraryProfileDefWithVoids.prototype.load = function (errors) {
        this.profileType = this.__attributes[0];
        this.profileName = this.__attributes[1];
        this.outerCurve = this.__attributes[2];
        this.innerCurves = [];

        if (this.__attributes[3]) {
            for (var n in this.__attributes[3]) {
                this.innerCurves.push(this.__attributes[3][n]);
            }
        } else {
            // TODO Error
        }
    };

    IfcArbitraryClosedProfileDef.prototype.remove = function (project) {
        if (this.__outerCurve) {
            this.__outerCurve.remove(project);
        }

        for (var n in this.__innerCurves) {
            this.__innerCurves[n].remove(project);
        }

        project.unregister(this);
    };

    IfcArbitraryProfileDefWithVoids.prototype.bind = function (project) {
        if (this.outerCurve) {
            this.__outerCurve = project.map[this.outerCurve];
        }

        this.__innerCurves = [];

        for (var n in this.innerCurves) {
            this.__innerCurves.push(project.map[this.innerCurves[n]]);
        }
    };

    IfcArbitraryProfileDefWithVoids.prototype.getGeo = function () {
        // TODO

        return this.__outerCurve.getGeo();
    };

    function IfcPolyline() {
    }

// #1758= IFCPOLYLINE((#1720,#1722,#1724,#1726,#1728,#1730,#1732,#1734,#1736,#1738,#1740,#1742,#1744,#1746,#1748,#1750,#1752,#1754,#1756,#1720));

    IfcPolyline.prototype.create = function (project, points) {
        this.type = 'IFCPOLYLINE';
        this.id = getUuid();
        this.points = [];
        this.__points = [];

        for (var n in points) {
            var placement = new IfcAxis2Placement3D().create(project);

            placement.setLocation(project, points[n].x, points[n].y, points[n].z);

            this.__points.push(placement);
            this.points.push(placement.oid);
        }

        project.register(this);

        return this;
    };

    IfcPolyline.prototype.remove = function (project) {
        for (var n in this.__points) {
            this.__points[n].remove(project);
        }

        project.unregister(this);
    };

    IfcPolyline.prototype.load = function (error) {
        this.points = this.__attributes[0];
    };

    IfcPolyline.prototype.bind = function (project) {
        this.__points = [];

        for (var n in this.points) {
            this.__points.push(project.map[this.points[n]]);
        }
    };

    IfcPolyline.prototype.getGeo = function () {
        var points = [];

        for (var n in this.__points) {
            // Safeguard TODO Why?
            if (this.__points[n].__location) {
                points.push({
                    x: this.__points[n].__location.coordinates.x,
                    y: this.__points[n].__location.coordinates.y
                });
            } else {
                points.push({x: this.__points[n].coordinates.x, y: this.__points[n].coordinates.y});
            }
        }

        return new GeoPolylineProfile().initialize(points);
    };

    function IfcCompositeCurve() {
    }

    IfcCompositeCurve.prototype.remove = function (project) {
        for (var n in this.__points) {
            this.__points[n].remove(project);
        }

        project.unregister(this);
    };

    IfcCompositeCurve.prototype.load = function (error) {
        if (this.__attributes[0]) {
            //  this.points = this.__attributes[0];
        } else {
            // TODO Error
        }
    };

    IfcCompositeCurve.prototype.bind = function (project) {
        this.__points = [];

        if (this.points) {
            for (var n in this.points) {
                this.__points.push(project.map[this.points[n]]);
            }
        }
    };

    IfcCompositeCurve.prototype.getGeo = function (viewBox) {
        return new GeoNullProfile();
    };

    function IfcWallType() {
    }

// #47190= IFCWALLTYPE('2thhQ$hg1AEP2QILLlrNZl',#41,'Basic Wall:Concrete_300_3XN',$,$,$,$,'335904',$,.STANDARD.);

    IfcWallType.prototype.load = function (error) {
        this.id = this.__attributes[0];
        this.name = this.__attributes[2];
    };

    IfcWallType.prototype.bind = function (project) {
    };

    // Geometrical classes

    function calcPolygonArea(vertices) {
      var total = 0.0;
      for (var i = 0, l = vertices.length; i < l; i++) {
        var addX = vertices[i].x;
        var addY = vertices[i == vertices.length - 1 ? 0 : i + 1].y;
        var subX = vertices[i == vertices.length - 1 ? 0 : i + 1].x;
        var subY = vertices[i].y;
        total += (addX * addY * 0.5);
        total -= (subX * subY * 0.5);
      }
      return Math.abs(total);
    }

    function matrixFromRotation(rotation) {
        if (rotation == 90) {
            return [[0, 1], [-1, 0]];
        }
        else if (rotation == 180) {
            return [[-1, 0], [0, -1]];
        }
        else if (rotation == 270) {
            return [[0, -1], [1, 0]];
        }

        return [[1, 0], [0, 1]];
    }

    function GeoNullProfile() {
        this.style = {};
    };

    GeoNullProfile.prototype.renderSvg = function (style) {
        var svg = '';

        return svg;
    };

    GeoNullProfile.prototype.getBoundingBox = function () {
        return {x: 0, y: 0, width: 0, height: 0};
    };

    GeoNullProfile.prototype.translate = function (vector) {
        return this;
    };

    GeoNullProfile.prototype.rotate = function (vector) {
        return this;
    };

    GeoNullProfile.prototype.applyPlacement = function (placement) {
        this.translate(placement.translation).rotate(placement.rotation);

        return this;
    };

    GeoNullProfile.prototype.renderSvg = function (svg) {
        var group = svg.group();

        return group;
    };

    GeoNullProfile.prototype.getPoints = function () {
        return [];
    }

    function GeoRectangleProfile() {
    };

    GeoRectangleProfile.prototype.initialize = function (width, height, x, y) {
        this.width = width;
        this.height = height;
        this.x = x;
        this.y = y;

        return this;
    };

    GeoRectangleProfile.prototype.getBoundingBox = function () {
        return new BoundingBox().initializeFromRectangle(this.x, this.y, this.width, this.height);
    };

    GeoRectangleProfile.prototype.translate = function (vector) {
        this.x += vector.x;
        this.y += vector.y;

        return this;
    };

    GeoRectangleProfile.prototype.rotate = function (rotation) {
        var matrix = matrixFromRotation(rotation);

        //var x = matrix[0][0] * this.x + matrix[0][1] * this.y;
        //var y = matrix[1][0] * this.x + matrix[1][1] * this.y;
        //
        //this.x = x;
        //this.y = y;

        var width = matrix[0][0] * this.width + matrix[0][1] * this.height;
        var height = matrix[1][0] * this.width + matrix[1][1] * this.height;

        this.width = /*x +*/ Math.abs(width);
        this.height = Math.abs(height);

        return this;
    };

    GeoRectangleProfile.prototype.applyPlacement = function (placement) {
        this.translate(placement.translation).rotate(placement.rotation);

        return this;
    };

    GeoRectangleProfile.prototype.renderSvg = function (svg) {
        var group = svg.group();
        var rect = group.rect(this.width, this.height);

        rect.dmove(this.x, this.y);

        return group;
    };

    GeoRectangleProfile.prototype.getPoints = function () {
        return [{x: this.x, y: this.y},
            {x: this.x + this.width, y: this.y},
            {x: this.x + this.width, y: this.y + this.height},
            {x: this.x, y: this.y + this.height}];
    };

    function GeoPolylineProfile() {
    };

    GeoPolylineProfile.prototype.initialize = function (points) {
        this.points = points;
        // this.rotation = 0;
        //
        // var xOffset = this.points[0].x;
        // var yOffset = this.points[0].y;
        //
        // this.translation = {x: xOffset, y: yOffset};
        //
        // for (var n in this.points) {
        //     this.points[n].x -= xOffset;
        //     this.points[n].y -= yOffset;
        // }

        return this;
    };

    GeoPolylineProfile.prototype.getBoundingBox = function () {
        return new BoundingBox().initializeFromPoints(this.points);
    };

    GeoPolylineProfile.prototype.translate = function (vector) {
        for (var n in this.points) {
            this.points[n].x += vector.x;
            this.points[n].y += vector.y;
        }

        return this;
    };

    GeoPolylineProfile.prototype.rotate = function (rotation) {
        var matrix = matrixFromRotation(rotation);

        var xRef = this.points[0].x;
        var yRef = this.points[0].y;

        for (var n in this.points) {
            var x = matrix[0][0] * (this.points[n].x - xRef) + matrix[0][1] * (this.points[n].y - yRef);
            var y = matrix[1][0] * (this.points[n].x - xRef) + matrix[1][1] * (this.points[n].y - yRef);

            this.points[n].x = x + xRef;
            this.points[n].y = y + yRef;
        }

        return this;
    };

    GeoPolylineProfile.prototype.applyPlacement = function (placement) {
        this.translate(placement.translation).rotate(placement.rotation);

        return this;
    };

    GeoPolylineProfile.prototype.renderSvg = function (svg) {
        var group = svg.group();
        var path = '';

        for (var n = 0; n < this.points.length; ++n) {
            if (n > 0) {
                path += ' ';
            }

            path += Math.round(this.points[n].x);
            path += ',';
            path += Math.round(this.points[n].y);
        }

        group.polyline(path);

        return group;
    };

    GeoPolylineProfile.prototype.getPoints = function () {
        return this.points;
    }

    function BoundingBox() {
    };

    BoundingBox.prototype.initializeFromRectangle = function (x, y, width, height) {
        this.x = x;
        this.y = y;
        this.width = width;
        this.height = height;
        this.left = x;
        this.top = y;
        this.right = x + width;
        this.bottom = y + height;
        this.xCenter = x + 0.5 * width;
        this.yCenter = y + 0.5 * height;

        return this;
    };

    BoundingBox.prototype.initializeFromPoints = function (points) {
        this.x = Number.MAX_SAFE_INTEGER;
        this.y = Number.MAX_SAFE_INTEGER;
        this.width = 0;
        this.height = 0;

        for (const point of points) {
          this.x = Math.min(this.x, point.x);
          this.y = Math.min(this.y, point.y);
          // this.width = Math.max(this.width, point.x - this.x);
          // this.height = Math.max(this.height, point.y - this.y);
        }

        for (const point of points) {
          this.width = Math.max(this.width, point.x - this.x);
          this.height = Math.max(this.height, point.y - this.y);
        }

        this.left = this.x;
        this.top = this.y;
        this.right = this.x + this.width;
        this.bottom = this.y + this.height;
        this.xCenter = this.x + 0.5 * this.width;
        this.yCenter = this.y + 0.5 * this.height;

        return this;
    };

    /**
     * @link http://jcward.com/UUID.js
     */
    function getUuid() {
        var lut = [];

        for (var i = 0; i < 256; i++) {
            lut[i] = (i < 16 ? '0' : '') + (i).toString(16);
        }

        var d0 = Math.random() * 0xffffffff | 0;
        var d1 = Math.random() * 0xffffffff | 0;
        var d2 = Math.random() * 0xffffffff | 0;
        var d3 = Math.random() * 0xffffffff | 0;

        return lut[d0 & 0xff] + lut[d0 >> 8 & 0xff] + lut[d0 >> 16 & 0xff] + lut[d0 >> 24 & 0xff] + '-' +
            lut[d1 & 0xff] + lut[d1 >> 8 & 0xff] + '-' + lut[d1 >> 16 & 0x0f | 0x40] + lut[d1 >> 24 & 0xff] + '-' +
            lut[d2 & 0x3f | 0x80] + lut[d2 >> 8 & 0xff] + '-' + lut[d2 >> 16 & 0xff] + lut[d2 >> 24 & 0xff] +
            lut[d3 & 0xff] + lut[d3 >> 8 & 0xff] + lut[d3 >> 16 & 0xff] + lut[d3 >> 24 & 0xff];
    }

    function dumpStringAttribute(stream, string, first) {
        if (!first) {
            stream.write(', ');
        }

        if (string) {
            stream.write('\'');
            stream.write(string);
            stream.write('\'');
        } else {
            stream.write('$');
        }
    }

    function dumpEnumAttribute(stream, string, first) {
        if (!first) {
            stream.write(', ');
        }

        if (string) {
            stream.write(string);
        } else {
            stream.write('$');
        }
    }

    function dumpNumberAttribute(stream, number, first) {
        if (!first) {
            stream.write(', ');
        }

        if (number != null) {
            stream.write(number);
        } else {
            stream.write('$');
        }
    }

    function dumpOidAttribute(stream, oid, first) {
        if (!first) {
            stream.write(', ');
        }

        if (oid != null) {
            stream.write('#' + oid);
        } else {
            stream.write('$');
        }
    }

    function dumpOidListAttribute(stream, list, first) {
        if (!first) {
            stream.write(', ');
        }

        stream.write('(');

        for (var n = 0; n < list.length; ++n) {
            if (n > 0) {
                stream.write(', ');
            }

            stream.write('#' + list[n]);
        }

        stream.write(')');
    }

    /**
     *
     */
    function cloneFiltered(object, exp) {
        var objects = [];

        return deepOmit(object, objects, function (key) {
            if (_.isString(key)) {
                if (!exp) {
                    if (key.indexOf("__") == 0 || key.indexOf("$$") == 0) {
                        return true;
                    }
                } else {
                    return exp.test(key);
                }
            }

            return false;
        });
    };

    /**
     *
     * @param sourceObj
     * @param callback
     * @param thisArg
     * @returns {*}
     */
    function deepOmit(sourceObj, objects, callback, thisArg) {
        var destObj, i, shouldOmit, newValue;

        if (_.isUndefined(sourceObj)) {
            return undefined;
        }

        callback = thisArg ? _.bind(callback, thisArg) : callback;

        if (_.isFunction(sourceObj)) {
            return undefined;
        } else if (_.isArray(sourceObj)) {
            destObj = [];

            for (i = 0; i < sourceObj.length; i++) {
                if (!callback(i)) {
                    destObj.push(deepOmit(sourceObj[i], objects, callback));
                }
            }
        }
        else if (_.isObject(sourceObj)) {
            if (objects.indexOf(sourceObj) >= 0) {
                console.log("Circular", sourceObj);

                return undefined;
            }

            objects.push(sourceObj);

            destObj = {};

            _.forOwn(sourceObj, function (value, key) {
                if (!callback(key)) {
                    destObj[key] = deepOmit(value, objects, callback);
                }
            });
        } else {
            return sourceObj;
        }

        return destObj;
    }

    function inheritMethods(target, source) {
        for (var member in source) {
            if (source[member] instanceof Function) {
                target[member] = source[member];
            }
        }
    }

    function isRelationship(object) {
        return object.type === 'IFCRELAGGREGATES' ||
        object.type === 'IFCRELDEFINESBYPROPERTIES' ||
        object.type === 'IFCRELCONTAINEDINSPATIALSTRUCTURE' ||
        object.type === 'IFCRELREFERENCEDINSPATIALSTRUCTURE';
    }


    // SNIP

    /**
     *
     * @param point
     * @param x
     * @param y
     * @param width
     * @param height
     * @returns {{x, y, i, fTo, fFrom}|{x: *, y: *, i: *, fTo: *, fFrom: *}}
     */
    function getClosestPointOnRectangle(point, x, y, width, height) {
        var vertices = [{x: x, y: y},
            {x: x + width, y: y},
            {x: x + width, y: y + height},
            {x: x, y: y + height},
            {x: x, y: y}];

        return getClosestPointOnLines(point, vertices);
    }

    /**
     * Find point on lines nearest test point test point point with properties .x and .y
     * lines defined by array vertices with elements having properties .x and .y
     * return is object with .x and .y properties and property i indicating nearest vertex in vertices
     * and property fFrom the fractional distance of the returned point from segments[i-1]
     * and property fTo the fractional distance of the returned point from segments[i]
     *
     * @param point
     * @param vertices
     * @returns {{x: *, y: *, i: *, fTo: *, fFrom: *}}
     */
    function getClosestPointOnLines(point, vertices) {
        var shortestDistance = Number.MAX_SAFE_INTEGER;
        var result;

        if (vertices.length > 1) {
            result = {};

            for (var n = 1; n < vertices.length; n++) {
                var start = new Vector(vertices[n - 1].x, vertices[n - 1].y);
                var end = new Vector(vertices[n].x, vertices[n].y);
                var line = end.subtract(start);
                var point = new Vector(point.x, point.y);
                var to = point.subtract(start);
                var pointProduct = line.dot(to);
                var normLine = line.length();
                var normTo = to.length();
                var projection = pointProduct / normLine;
                var distance;
                var pointOnLine;

                if (projection <= 0) {
                    distance = start.subtract(point).length();
                    pointOnLine = start;
                } else if (projection >= normLine) {
                    distance = end.subtract(point).length();
                    pointOnLine = end;
                } else {
                    pointOnLine = start.add(line.clone().unit().multiply(projection));
                    distance = pointOnLine.subtract(point).length();
                }

                if (distance < shortestDistance) {
                    result.x = pointOnLine.x;
                    result.y = pointOnLine.y;
                    result.distance = distance;
                    result.i = n;
                    result.segment = [{x: vertices[n - 1].x, y: vertices[n - 1].y}, {
                        x: vertices[n].x,
                        y: vertices[n].y
                    }];
                    shortestDistance = distance;
                }
            }
        }

        return result;
    }

    function distanceBetweenFeatureLines(startPt1, endPt1, startPt2, endPt2) {
      var u, v, w, a, b, c, d, e, D, sc, sN, sD, tc, tN, tD;
          
      u = new THREE.Vector3();
      u.subVectors(endPt1, startPt1);
      v = new THREE.Vector3();
      v.subVectors(endPt2, startPt2);
      w = new THREE.Vector3();
      w.subVectors(startPt1, startPt2);
      
      a = u.dot(u);           // always >= 0
      b = u.dot(v);
      c = v.dot(v);           // always >= 0
      d = u.dot(w);
      e = v.dot(w);
      
      D = (a * c) - (b * b);  // always >= 0
      sD = D;                 // default sD = D >= 0
      tD = D;                 // default tD = D >= 0
    
      const EPSILON = 0.0001;
      
      // compute the line parameters of the two closest points
      if (D < EPSILON) {  
          // the lines are almost parallel
          sN = 0.0;         // force using point P0 on segment S1
          sD = 1.0;         // to prevent possible division by 0.0 later
          tN = e;
          tD = c;
      }
      else 
      {
          // get the closest points on the infinite lines
          sN = (b*e - c*d);
          tN = (a*e - b*d);
          if (sN < 0.0) {     
              // sc < 0 => the s=0 edge is visible
              sN = 0.0;
              tN = e;
              tD = c;
          }
          else if (sN > sD) 
          {  
              // sc > 1  => the s=1 edge is visible
              sN = sD;
              tN = e + b;
              tD = c;
          }
      }
      
      if (tN < 0.0) {            
          // tc < 0 => the t=0 edge is visible
          tN = 0.0;
          // recompute sc for this edge
          if (-d < 0.0) 
          {
              sN = 0.0;
          }
          else if (-d > a) 
          {
              sN = sD;
          }
          else 
          {
              sN = -d;
              sD = a;
          }
      }
      else if (tN > tD) 
      {      
          // tc > 1  => the t=1 edge is visible
          tN = tD;
          // recompute sc for this edge
          if ((-d + b) < 0.0) 
          {
              sN = 0;
          }
          else if ((-d + b) > a) 
          {
              sN = sD;
          }
          else 
          {
              sN = (-d +  b);
              sD = a;
          }
      }
      
      // finally do the division to get sc and tc
      sc = (Math.abs(sN) < EPSILON ? 0.0 : sN / sD);
      tc = (Math.abs(tN) < EPSILON ? 0.0 : tN / tD);
    
      // get the difference of the two closest points
      var sc_mult_u = new THREE.Vector3();
      sc_mult_u.copy(u);
      sc_mult_u.multiplyScalar(sc);
      var tc_mult_v = new THREE.Vector3();
      tc_mult_v.copy(v);
      tc_mult_v.multiplyScalar(tc);
      var dP = new THREE.Vector3();
      dP.copy(w);
      dP.add(sc_mult_u);
      dP.sub(tc_mult_v);
    
      return dP.length();   // return the closest distance
    }

    function isIn2DPolygon(points, point, innerDist = null, edgesToExFromSpacing = [], customInnerDist = null){
      let res = false;
      for (let iii=0, jjj=points.length-1; iii<points.length; jjj=iii++) {
        if (((points[iii].y>point.y) != (points[jjj].y>point.y)) &&
          (point.x < (points[jjj].x-points[iii].x) * (point.y-points[iii].y) /
          (points[jjj].y-points[iii].y) + points[iii].x) ){
          res = !res;
        }
      }
      if(res){
        const closestOnLineRes = getClosestPointOnLines(point, points);
        if(edgesToExFromSpacing.length && customInnerDist !== null){
          let toExclude = false;
          for(let edge of edgesToExFromSpacing){
            if(closestOnLineRes.segment[0].x == edge[0].x && closestOnLineRes.segment[0].y == edge[0].y &&
              closestOnLineRes.segment[1].x == edge[1].x && closestOnLineRes.segment[1].y == edge[1].y ||
              closestOnLineRes.segment[1].x == edge[0].x && closestOnLineRes.segment[1].y == edge[0].y &&
              closestOnLineRes.segment[0].x == edge[1].x && closestOnLineRes.segment[0].y == edge[1].y){
                toExclude = true;
                break;
            }
          }
          if(toExclude){
            if(closestOnLineRes.distance < customInnerDist){
              return false
            } else {
              return true
            }
          }
        }
        if(innerDist !== null && closestOnLineRes.distance < innerDist){
          return false
        }
      }
      return res;
      }

    /**
     *
     * @param segment
     */
    function unitVector(segment) {
        var length = norm(segment);

        return {x: (segment[1].x - segment[0].x) / length, y: (segment[1].y - segment[0].y) / length};
    }

    /**
     *
     * @param vector
     * @returns {*|number}
     */
    function normVector(vector) {
        return Math.sqrt(Math.pow(vector.x, 2) + Math.pow(vector.y, 2));
    }

    /**
     *
     * @param segment
     * @returns {*|number}
     */
    function norm(segment) {
        return Math.sqrt(Math.pow(segment[0].x - segment[1].x, 2) + Math.pow(segment[0].y - segment[1].y, 2));
    }

    /**
     *
     * @constructor
     */
    function Navigator() {
    };

    /**
     *
     * @param speaker
     * @returns {Navigator}
     */
    Navigator.prototype.initialize = function () {
        this.listeners = [];
        this.step = -1;

        return this;
    };

    /**
     *
     * @param listener
     */
    Navigator.prototype.addListener = function (listener) {
        this.listeners.push(listener);
    };

    /**
     *
     * @param listener
     */
    Navigator.prototype.removeListener = function (listener) {
        Utils.removeItemFromArray(this.listeners, listener);
    };

    /**
     * Just in case that we need the algorithm.
     *
     * @param startSpace
     */
    Navigator.prototype.breadthFirstSearch = function (startSpace) {
        var queue = [{vertex: startSpace, count: 0}];
        var visited = {};
        var tail = 0;

        visited[startSpace.id] = true;

        while (tail < queue.length) {
            var u = queue[tail].vertex; // Pop a vertex off the queue.
            var count = queue[tail++].count;

            if (u.type == 'IFCSPACE') {
                u.__doors.forEach(function (v) {
                    if (!visited[v.id]) {
                        visited[v.id] = true;
                        queue.push({vertex: v, count: count + 1});
                    }
                });
            } else {
                u.__spaces.forEach(function (v) {
                    if (!visited[v.id]) {
                        visited[v.id] = true;
                        queue.push({vertex: v, count: count + 1});
                    }
                });
            }
        }
    };

    /**
     *
     * @param source
     * @param target
     * @returns {*|{then, catch, finally}}
     */
    Navigator.prototype.calculate = function (source, target) {
        this.startSpace = source;
        this.destinationSpace = target;

        for (var n in this.listeners) {
            try {
                this.listeners[n].onNavigationEvent('routingStarted', this);
            } catch (x) {
                console.error(x);
            }
        }

        this.paths = [];

        if (source.id == target.id) {
            window.alert('Start and Destination are equal.');
            return;
        }

        var queue = [source];
        var visited = {};
        var predecessor = {};
        var tail = 0;

        visited[source.id] = true;

        while (tail < queue.length) {
            var u = queue[tail++];
            var neighbors;

            if (u.type == 'IFCSPACE') {
                neighbors = u.__doors;
            } else if (u.type == 'IFCDOOR') {
                neighbors = u.__spaces;
            }

            for (var i = 0; i < neighbors.length; ++i) {
                var v = neighbors[i];

                if (visited[v.id]) {
                    continue;
                }

                visited[v.id] = true;

                if (v.id === target.id) {
                    var path = [v];   // If so, backtrack through the path.

                    while (u.id !== source.id) {
                        path.push(u);

                        u = predecessor[u.id];
                    }

                    path.push(u);
                    path.reverse();

                    this.paths.push({path: path});
                } else {
                    predecessor[v.id] = u;
                    queue.push(v);
                }
            }
        }

        // Calculate path lengths and determine shortest path

        try {
            this.shortestRoute = null;
            this.routes = [];

            for (var n in this.paths) {
                var wayPoints = this.paths[n].path;
                var lastWayPoint = wayPoints[0];
                var route = {length: 0, steps: []};

                this.routes.push(route);

                this.speed = 1389; // mm/s

                for (var m = 1; m < wayPoints.length; ++m) {
                    var currentWayPoint = wayPoints[m];
                    var lastGeo = lastWayPoint.getGeo();
                    var lastBox = lastGeo.getBoundingBox();
                    var currentGeo = currentWayPoint.getGeo();
                    var currentBox = currentGeo.getBoundingBox();
                    var step = {start: lastWayPoint, end: currentWayPoint};

                    route.steps.push(step);

                    step.length = Math.sqrt(Math.pow(currentBox.xCenter - lastBox.xCenter, 2) +
                        Math.pow(currentBox.yCenter - lastBox.yCenter, 2));

                    if (!((lastWayPoint.type == 'IFCSPACE' && currentWayPoint.type == 'IFCDOOR' && m != 0) ||
                        (lastWayPoint.type == 'IFCDOOR' && currentWayPoint.type == 'IFCSPACE' && m != wayPoints.length - 1))) {
                        step.length = 0.5 * step.length;
                    }

                    route.length += step.length;

                    lastWayPoint = wayPoints[m];
                }

                if (this.shortestRoute == null || route.length < this.shortestRoute.steps.length) {
                    this.shortestRoute = route;
                }
            }

            if (this.shortestRoute != null) {
                for (var n in this.listeners) {
                    this.listeners[n].onNavigationEvent('pathFound', this);
                }
            } else {
                for (var n in this.listeners) {
                    this.listeners[n].onNavigationEvent('noPathFound', this);
                }
            }
        } catch (x) {
            console.error(x);
        }
    };

    /**
     *
     * @returns {boolean}
     */
    Navigator.prototype.isActive = function () {
        return this.step >= 0;
    };

    /**
     *
     * @returns {boolean}
     */
    Navigator.prototype.isDestinationReached = function () {
        return this.step == this.shortestRoute.steps.length - 1;
    };

    /**
     *
     * @returns {*|{then, catch, finally}}
     */
    Navigator.prototype.next = function (start) {
        var deferred = jQuery.Deferred();

        if (start) {
            this.step = 0;
            this.departureTime = moment();
            this.remainingDistance = this.shortestRoute.length;

            for (var n in this.listeners) {
                try {
                    this.listeners[n].onNavigationEvent('navigationStarted', this);
                } catch (error) {
                    console.log(error);
                }
            }
        }

        if (this.shortestRoute.steps[this.step].end.type == 'IFCSPACE') {
            for (var n in this.listeners) {
                try {
                    this.listeners[n].onNavigationEvent('spaceStep', this);
                } catch (error) {
                    console.log(error);
                }
            }
        }
        else if (this.shortestRoute.steps[this.step].end.type == 'IFCDOOR') {
            for (var n in this.listeners) {
                try {
                    this.listeners[n].onNavigationEvent('doorStep', this);
                } catch (error) {
                    console.log(error);
                }
            }
        }

        if (this.step == this.shortestRoute.steps.length - 1) {
            this.remainingDistance = 0;
            this.remainingTime = 0;
            this.estimatedArrivalTime = moment();

            for (var n in this.listeners) {
                this.listeners[n].onNavigationEvent('destinationReached', this);
            }

            deferred.resolve();
        } else {
            this.remainingDistance -= this.shortestRoute.steps[this.step].length;
            this.remainingTime = this.remainingDistance / 1389;
            this.estimatedArrivalTime = moment().add(this.remainingTime, 's');

            this.step++;

            this.confirmation = jQuery.Deferred();

            this.confirmation.promise().done(function () {
                this.next().done(function () {
                    deferred.resolve();
                }.bind(this));
            }.bind(this));
        }

        return deferred.promise();
    };

    /**
     *
     */
    Navigator.prototype.confirmStep = function () {
        console.log('Promise', this.confirmation);
        this.confirmation.resolve();
    };

    /**
     *
     */
    Navigator.prototype.endNavigation = function () {
        this.remainingDistance = null;
        this.remainingTime = null;
        this.estimatedArrivalTime = null;

        this.step = -1;
    };

    /**
     *
     */
    Navigator.prototype.clear = function () {
        this.startSpace = null;
        this.destinationSpace = null;
        this.shortestRoute = null;
        this.remainingDistance = null;
        this.remainingTime = null;
        this.estimatedArrivalTime = null;
        this.step = -1;
    };

    return {
        addProject: function (name) {
            return new IfcProject().create(name);
        },
        bind: function (project) {
            var start = new Date().getTime();

            if (!project.type) {
                project.type = 'IFCPROJECT';
            }

            // Retrofit from TIP2

            var ifcElementAttributes = {};

            if (project.ifcElementAttributes) {
                Object.keys(project.ifcElementAttributes).forEach(function (key, index) {
                    ifcElementAttributes[key] = project.ifcElementAttributes[key];
                });
            }

            project.__ifcElementAttributes = ifcElementAttributes;

            if (!project.map) {
                project.map = {};
            }

            project.map[project.oid] = project;
            project.__relAggregates = {};
            project.__relContainedInSpatialStructure = {};
            project.__relReferencedInSpatialStructure = {};
            project.__relDefinesByProperties = {};

            for (var n in project.map) {
                if (!ifcTypes[project.map[n].type]) {
                    console.error('Skipping ' + project.map[n].type + ' - no type');

                    continue;
                }
                else if (isRelationship(project.map[n])) {
                    continue;
                }

                inheritMethods(project.map[n], ifcTypes[project.map[n].type]);
                project.map[n].bind(project);
            }

            for (var n in project.map) {
                if (!ifcTypes[project.map[n].type]) {
                    console.error('Skipping ' + project.map[n].type + ' - no type');

                    continue;
                }
                else if (!isRelationship(project.map[n])) {
                    continue;
                }

                inheritMethods(project.map[n], ifcTypes[project.map[n].type]);
                project.map[n].bind(project);

                // Register relationship objects

                if (project.map[n].type == 'IFCRELAGGREGATES') {
                    project.__relAggregates[project.map[n].relatingObject] = project.map[n];
                } else if (project.map[n].type == 'IFCRELCONTAINEDINSPATIALSTRUCTURE') {
                    project.__relContainedInSpatialStructure[project.map[n].relatingStructure] = project.map[n];
                } else if (project.map[n].type === 'IFCRELREFERENCEDINSPATIALSTRUCTURE') {
                    project.__relReferencedInSpatialStructure[project.map[n].relatingStructure] = project.map[n];
                } else if (project.map[n].type == 'IFCRELDEFINESBYPROPERTIES') {
                    project.__relDefinesByProperties[project.map[n].relatingPropertyDefinition] = project.map[n];
                }
            }

            var end = new Date().getTime();

            console.log('>>> Binding >>> ' + (end - start));

            project.__project = project;

            return project;
        },
        getPrototype: function (type) {
            return ifcTypes[type]
        },
        createNavigator: function () {
            return new Navigator().initialize();
        }
    };
})
();

// TODO Move to separate file and build with IFC

function Vector(x, y, z) {
    this.x = x || 0;
    this.y = y || 0;
    this.z = z || 0;
}

Vector.prototype = {
    negative: function () {
        return new Vector(-this.x, -this.y, -this.z);
    },
    add: function (v) {
        if (v instanceof Vector) return new Vector(this.x + v.x, this.y + v.y, this.z + v.z);
        else return new Vector(this.x + v, this.y + v, this.z + v);
    },
    subtract: function (v) {
        if (v instanceof Vector) return new Vector(this.x - v.x, this.y - v.y, this.z - v.z);
        else return new Vector(this.x - v, this.y - v, this.z - v);
    },
    multiply: function (v) {
        if (v instanceof Vector) return new Vector(this.x * v.x, this.y * v.y, this.z * v.z);
        else return new Vector(this.x * v, this.y * v, this.z * v);
    },
    divide: function (v) {
        if (v instanceof Vector) return new Vector(this.x / v.x, this.y / v.y, this.z / v.z);
        else return new Vector(this.x / v, this.y / v, this.z / v);
    },
    equals: function (v) {
        return this.x == v.x && this.y == v.y && this.z == v.z;
    },
    dot: function (v) {
        return this.x * v.x + this.y * v.y + this.z * v.z;
    },
    cross: function (v) {
        return new Vector(
            this.y * v.z - this.z * v.y,
            this.z * v.x - this.x * v.z,
            this.x * v.y - this.y * v.x
        );
    },
    length: function () {
        return Math.sqrt(this.dot(this));
    },
    unit: function () {
        return this.divide(this.length());
    },
    min: function () {
        return Math.min(Math.min(this.x, this.y), this.z);
    },
    max: function () {
        return Math.max(Math.max(this.x, this.y), this.z);
    },
    toAngles: function () {
        return {
            theta: Math.atan2(this.z, this.x),
            phi: Math.asin(this.y / this.length())
        };
    },
    angleTo: function (a) {
        return Math.acos(this.dot(a) / (this.length() * a.length()));
    },
    toArray: function (n) {
        return [this.x, this.y, this.z].slice(0, n || 3);
    },
    clone: function () {
        return new Vector(this.x, this.y, this.z);
    },
    init: function (x, y, z) {
        this.x = x;
        this.y = y;
        this.z = z;
        return this;
    }
};

Vector.negative = function (a, b) {
    b.x = -a.x;
    b.y = -a.y;
    b.z = -a.z;
    return b;
};
Vector.add = function (a, b, c) {
    if (b instanceof Vector) {
        c.x = a.x + b.x;
        c.y = a.y + b.y;
        c.z = a.z + b.z;
    }
    else {
        c.x = a.x + b;
        c.y = a.y + b;
        c.z = a.z + b;
    }
    return c;
};
Vector.subtract = function (a, b, c) {
    if (b instanceof Vector) {
        c.x = a.x - b.x;
        c.y = a.y - b.y;
        c.z = a.z - b.z;
    }
    else {
        c.x = a.x - b;
        c.y = a.y - b;
        c.z = a.z - b;
    }
    return c;
};
Vector.multiply = function (a, b, c) {
    if (b instanceof Vector) {
        c.x = a.x * b.x;
        c.y = a.y * b.y;
        c.z = a.z * b.z;
    }
    else {
        c.x = a.x * b;
        c.y = a.y * b;
        c.z = a.z * b;
    }
    return c;
};
Vector.divide = function (a, b, c) {
    if (b instanceof Vector) {
        c.x = a.x / b.x;
        c.y = a.y / b.y;
        c.z = a.z / b.z;
    }
    else {
        c.x = a.x / b;
        c.y = a.y / b;
        c.z = a.z / b;
    }
    return c;
};
Vector.cross = function (a, b, c) {
    c.x = a.y * b.z - a.z * b.y;
    c.y = a.z * b.x - a.x * b.z;
    c.z = a.x * b.y - a.y * b.x;
    return c;
};
Vector.unit = function (a, b) {
    var length = a.length();
    b.x = a.x / length;
    b.y = a.y / length;
    b.z = a.z / length;
    return b;
};
Vector.fromAngles = function (theta, phi) {
    return new Vector(Math.cos(theta) * Math.cos(phi), Math.sin(phi), Math.sin(theta) * Math.cos(phi));
};
Vector.randomDirection = function () {
    return Vector.fromAngles(Math.random() * Math.PI * 2, Math.asin(Math.random() * 2 - 1));
};
Vector.min = function (a, b) {
    return new Vector(Math.min(a.x, b.x), Math.min(a.y, b.y), Math.min(a.z, b.z));
};
Vector.max = function (a, b) {
    return new Vector(Math.max(a.x, b.x), Math.max(a.y, b.y), Math.max(a.z, b.z));
};
Vector.lerp = function (a, b, fraction) {
    return b.subtract(a).multiply(fraction).add(a);
};
Vector.fromArray = function (a) {
    return new Vector(a[0], a[1], a[2]);
};
Vector.angleBetween = function (a, b) {
    return a.angleTo(b);
};

function pToPDist(p1, p2){
  return Math.sqrt(Math.pow(p2.x-p1.x, 2) + Math.pow(p2.y-p1.y, 2) + Math.pow(p2.z-p1.z, 2));
}

function pPCenter2D(p1, p2){
  return {x: p1.x + 0.5 * (p2.x - p1.x), y: p1.y + 0.5 * (p2.y - p1.y)}
}

function dist2D(p1, p2){
  return Math.sqrt(Math.pow(p2.x - p1.x, 2) + Math.pow(p2.y - p1.y, 2))
}

function distToSegment2D(p, v, w) {
    const distToSegment2DSquared = (p, v, w) => {
      const dist2 = (v, w) => {
          const sqr = (x) => {
            return x * x
          }
          return sqr(v.x - w.x) + sqr(v.y - w.y)
        }
      var l2 = dist2(v, w);
      if (l2 == 0) return dist2(p, v);
      var t = ((p.x - v.x) * (w.x - v.x) + (p.y - v.y) * (w.y - v.y)) / l2;
      if (t < 0) return dist2(p, v);
      if (t > 1) return dist2(p, w);
      return dist2(p, { x: v.x + t * (w.x - v.x), y: v.y + t * (w.y - v.y) });
      }
    return Math.sqrt(distToSegment2DSquared(p, v, w));
  }

  function horiDist(p1, p2){
    return Math.hypot(p2.x-p1.x, p2.y-p1.y);
  }

  function psIntersect(p1, p2, p3, p4){
    let x1 = p1.x, y1 = p1.y, x2 = p2.x, y2 = p2.y, x3 = p3.x, y3 = p3.y, x4 = p4.x, y4 = p4.y;
    let a1, a2, b1, b2, c1, c2;
    let r1, r2 , r3, r4;
    let denom;
    const sameSign = (a, b) => (a * b) > 0;
    a1 = y2 - y1;
    b1 = x1 - x2;
    c1 = (x2 * y1) - (x1 * y2);
    r3 = ((a1 * x3) + (b1 * y3) + c1);
    r4 = ((a1 * x4) + (b1 * y4) + c1);
    if ((r3 !== 0) && (r4 !== 0) && sameSign(r3, r4)){
      return 0;
    }
    a2 = y4 - y3;
    b2 = x3 - x4;
    c2 = (x4 * y3) - (x3 * y4);
    r1 = (a2 * x1) + (b2 * y1) + c2;
    r2 = (a2 * x2) + (b2 * y2) + c2;
    if ((r1 !== 0) && (r2 !== 0) && (sameSign(r1, r2))){
      return 0;
    }
    denom = (a1 * b2) - (a2 * b1);
    if (denom === 0) {
      return 1;
    }
    return 1;
  }

  function centerOfMass(points){
    let doZ = true, cOM = {x: 0, y: 0, z: 0}
    for(const point of points){
      cOM.x += point.x
      cOM.y += point.y
      if(point.z === undefined || point.z === null){
        doZ = false
        continue
      }
      cOM.z += point.z
    }
    if(!doZ){
      cOM.z = undefined
    }
    Object.keys(cOM).forEach(key => {
      if(key){
        if(cOM[key] !== undefined){
          cOM[key] /= points.length
        }
      }
    })
    return cOM
  }

  function shapePointsToGeoJSONObj(shapePoints){
    let res = [];
    for(let p of shapePoints){
      res.push([p.x, p.y]);
    }
    if(res.length >= 2 && (res[0][0] != res[res.length - 1][0] || res[0][1] != res[res.length - 1][1])){
      res.push(res[0]);
    }
    return [res];
  }
  
  function visualCenter(shapePoints){
    let arr = polylabel.default(shapePointsToGeoJSONObj(shapePoints));
    if(arr && arr.length == 2){
      return {x: arr[0], y: arr[1]};
    } else {
      return {x: 0, y: 0};
    }  
  }